@@ -1,6 +1,6 @@
 /* src/native/jni.h - JNI types and data structures
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,14 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Reinhard Grafl
-            Roman Obermaisser
-
-   Changes: Christian Thalinger
-
-   $Id: jni.h 5805 2006-10-19 09:32:29Z twisti $
+   $Id: jni.h 7246 2007-01-29 18:49:05Z twisti $
 
 */
 
@@ -64,7 +57,8 @@ typedef struct localref_table localref_table;
 #include "vm/types.h"
 
 #include "vm/global.h"
-#include "vm/method.h"
+
+#include "vmcore/method.h"
 
 
 /* _Jv_JNIEnv *****************************************************************/
@@ -86,10 +80,12 @@ struct _Jv_JavaVM {
 	/* JVM instance-specific variables */
 
 	s8 starttime;                       /* VM startup time                    */
-	s8 total_started_thread_count;
 
 	s4 Java_gnu_java_lang_management_VMClassLoadingMXBeanImpl_verbose;
 	s4 Java_gnu_java_lang_management_VMMemoryMXBeanImpl_verbose;
+	s4 java_lang_management_ThreadMXBean_PeakThreadCount;
+	s4 java_lang_management_ThreadMXBean_ThreadCount;
+	s8 java_lang_management_ThreadMXBean_TotalStartedThreadCount;
 	s4 Java_java_lang_VMClassLoader_defaultAssertionStatus;
 };
 
@@ -144,6 +140,7 @@ struct hashtable_global_ref_entry {
 
 /* initialize JNI subsystem */
 bool jni_init(void);
+bool jni_init_localref_table(void);
 
 java_objectheader *_Jv_jni_invokeNative(methodinfo *m, java_objectheader *o,
 										java_objectarray *params);
