@@ -1,6 +1,6 @@
 /* src/toolbox/avl.c - AVL tree implementation
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: avl.c 5123 2006-07-12 21:45:34Z twisti $
+   $Id: avl.c 7862 2007-05-03 14:53:39Z twisti $
 
 */
 
@@ -40,14 +34,14 @@
 #include "vm/types.h"
 
 #include "mm/memory.h"
-#include "toolbox/avl.h"
 
-#if defined(ENABLE_THREADS)
-# include "threads/native/lock.h"
-# include "threads/native/threads.h"
-#else
-# include "threads/none/lock.h"
-#endif
+#include "threads/lock-common.h"
+
+#include "toolbox/avl.h"
+#include "toolbox/logging.h"
+
+#include "vm/global.h"
+#include "vm/vm.h"
 
 
 /* avl_create ******************************************************************
@@ -56,14 +50,14 @@
 
 *******************************************************************************/
 
-avl_tree *avl_create(avl_comparator *compar)
+avl_tree_t *avl_create(avl_comparator *comparator)
 {
-	avl_tree *t;
+	avl_tree_t *t;
 
-	t = NEW(avl_tree);
+	t = NEW(avl_tree_t);
 
 	t->root       = NULL;
-	t->comparator = compar;
+	t->comparator = comparator;
 	t->entries    = 0;
 
 #if defined(ENABLE_THREADS)
@@ -71,7 +65,7 @@ avl_tree *avl_create(avl_comparator *compar)
 
 	t->lock       = NEW(java_objectheader);
 
-	lock_init_object_lock(t->lock);
+	LOCK_INIT_OBJECT_LOCK(t->lock);
 #endif
 
 	return t;
@@ -84,11 +78,11 @@ avl_tree *avl_create(avl_comparator *compar)
 
 *******************************************************************************/
 
-static avl_node *avl_newnode(void *data)
+static avl_node_t *avl_newnode(void *data)
 {
-	avl_node *n;
+	avl_node_t *n;
 
-	n = NEW(avl_node);
+	n = NEW(avl_node_t);
 
 	n->data      = data;
 
@@ -114,10 +108,10 @@ static avl_node *avl_newnode(void *data)
 
 *******************************************************************************/
 
-static void avl_rotate_left(avl_node **node)
+static void avl_rotate_left(avl_node_t **node)
 {
-	avl_node *tmp;
-	avl_node *tmpnode;
+	avl_node_t *tmp;
+	avl_node_t *tmpnode;
 
 	/* rotate the node */
 
@@ -144,10 +138,10 @@ static void avl_rotate_left(avl_node **node)
 
 *******************************************************************************/
 
-static void avl_rotate_right(avl_node **node)
+static void avl_rotate_right(avl_node_t **node)
 {
-	avl_node *tmp;
-	avl_node *tmpnode;
+	avl_node_t *tmp;
+	avl_node_t *tmpnode;
 
 	/* rotate the node */
 
@@ -168,10 +162,10 @@ static void avl_rotate_right(avl_node **node)
 
 *******************************************************************************/
 
-static void avl_adjust_balance(avl_node *node)
+static void avl_adjust_balance(avl_node_t *node)
 {
-	avl_node *left;
-	avl_node *right;
+	avl_node_t *left;
+	avl_node_t *right;
 
 	left  = node->childs[AVL_LEFT];
 	right = node->childs[AVL_RIGHT];
@@ -203,13 +197,13 @@ static void avl_adjust_balance(avl_node *node)
 
 *******************************************************************************/
 
-static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
+static s4 avl_insert_intern(avl_tree_t *tree, avl_node_t **node, void *data)
 {
-	avl_node *tmpnode;
-	s4        res;
-	s4        direction;
-	s4        insert;
-	s4        balance;
+	avl_node_t *tmpnode;
+	s4          res;
+	s4          direction;
+	s4          insert;
+	s4          balance;
 
 	/* set temporary variable */
 
@@ -221,12 +215,12 @@ static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
 
 	/* compare the current node */
 
-	res = tree->comparator(data, tmpnode->data);
+	res = tree->comparator(tmpnode->data, data);
 
 	/* is this node already in the tree? */
 
 	if (res == 0)
-		assert(0);
+		vm_abort("avl_insert_intern: node already in the tree");
 
 	/* goto left or right child */
 
@@ -236,9 +230,9 @@ static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
 
 	if (tmpnode->childs[direction]) {
 		balance = avl_insert_intern(tree, &tmpnode->childs[direction], data);
-
-	} else {
-		avl_node *newnode;
+	}
+	else {
+		avl_node_t *newnode;
 
 		/* no child, create this node */
 
@@ -270,14 +264,14 @@ static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
 
 				tmpnode->balance = 0;
 				tmpnode->childs[AVL_RIGHT]->balance = 0;
-
-			} else {
+			}
+			else {
 				avl_rotate_left(&tmpnode->childs[AVL_LEFT]);
 				avl_rotate_right(&tmpnode);
 				avl_adjust_balance(tmpnode);
 			}
-
-		} else if (tmpnode->balance > 1) {
+		}
+		else if (tmpnode->balance > 1) {
 			/* right subtree too tall: left rotation needed */
 
 			if (tmpnode->childs[AVL_RIGHT]->balance > 0) {
@@ -287,14 +281,14 @@ static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
 
 				tmpnode->balance = 0;
 				tmpnode->childs[AVL_LEFT]->balance = 0;
-
-			} else {
+			}
+			else {
 				avl_rotate_right(&tmpnode->childs[AVL_RIGHT]);
 				avl_rotate_left(&tmpnode);
 				avl_adjust_balance(tmpnode);
 			}
-
-		} else {
+		}
+		else {
 			insert = 1;
 		}
 	}
@@ -315,7 +309,7 @@ static s4 avl_insert_intern(avl_tree *tree, avl_node **node, void *data)
 
 *******************************************************************************/
 
-bool avl_insert(avl_tree *tree, void *data)
+bool avl_insert(avl_tree_t *tree, void *data)
 {
 	assert(tree);
 	assert(data);
@@ -324,25 +318,15 @@ bool avl_insert(avl_tree *tree, void *data)
 
 	/* if we don't have a root node, create one */
 
-	if (tree->root == NULL) {
+	if (tree->root == NULL)
 		tree->root = avl_newnode(data);
-
-	} else {
+	else
 		avl_insert_intern(tree, &tree->root, data);
-	}
 
 	/* increase entries count */
 
 	tree->entries++;
 
-#if 0
-	printf("tree=%p entries=%d\n", tree, tree->entries);
-
-	printf("-------------------\n");
-	avl_dump(tree->root, 0);
-	printf("-------------------\n");
-#endif
-
 	LOCK_MONITOR_EXIT(tree->lock);
 
 	/* insertion was ok */
@@ -358,10 +342,10 @@ bool avl_insert(avl_tree *tree, void *data)
 
 *******************************************************************************/
 
-void *avl_find(avl_tree *tree, void *data)
+void *avl_find(avl_tree_t *tree, void *data)
 {
-	avl_node *node;
-	s4        res;
+	avl_node_t *node;
+	s4          res;
 
 	assert(tree);
 	assert(data);
@@ -373,7 +357,7 @@ void *avl_find(avl_tree *tree, void *data)
 	for (node = tree->root; node != NULL; ) {
 		/* compare the current node */
 
-		res = tree->comparator(data, node->data);
+		res = tree->comparator(node->data, data);
 
 		/* was the entry found? return it */
 
@@ -403,7 +387,7 @@ void *avl_find(avl_tree *tree, void *data)
 *******************************************************************************/
 
 #if !defined(NDEBUG)
-void avl_dump(avl_node* node, s4 indent)
+void avl_dump(avl_node_t* node, s4 indent)
 {
 	s4 tmp;
 
@@ -415,10 +399,13 @@ void avl_dump(avl_node* node, s4 indent)
 	if (node->childs[AVL_RIGHT])
 		avl_dump(node->childs[AVL_RIGHT], tmp + 1);
 
-	while(indent--)
-		printf("   ");
+	log_start();
 
-	printf("%p (%d)\n", node->data, node->balance);
+	while(indent--)
+		log_print("   ");
+
+	log_print("%p (%d)", node->data, node->balance);
+	log_finish();
 
 	if (node->childs[AVL_LEFT])
 		avl_dump(node->childs[AVL_LEFT], tmp + 1);
