@@ -1,6 +1,6 @@
 /* src/toolbox/logging.c - contains logging functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,14 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Reinhard Grafl
-
-   Changes: Christian Thalinger
-   			Edwin Steiner
-
-   $Id: logging.c 4967 2006-05-26 16:24:58Z edwin $
+   $Id: logging.c 7685 2007-04-11 08:21:06Z twisti $
 
 */
 
@@ -46,10 +39,9 @@
 #include "toolbox/logging.h"
 #include "toolbox/util.h"
 #include "vm/global.h"
-#include "vm/statistics.h"
 
-#if defined(ENABLE_THREADS)
-# include "threads/native/threads.h"
+#if defined(ENABLE_STATISTICS)
+# include "vmcore/statistics.h"
 #endif
 
 
@@ -77,16 +69,41 @@ void log_init(const char *fname)
 
 *******************************************************************************/
 
+/* ATTENTION: Don't include threads-common.h, because we can't
+   bootstrap this file in that case (missing java_lang_Thread.h).
+   Instead we declare threads_get_current_threadobject differently: */
+
+/* #if defined(ENABLE_THREADS) */
+/* # include "threads/threads-common.h" */
+/* #endif */
+
+extern ptrint threads_get_current_tid(void);
+
+
 void log_start(void)
 {
-	if (logfile) {
 #if defined(ENABLE_THREADS)
-		fprintf(logfile, "[%p] ", (void *) THREADOBJECT);
+	ptrint tid;
+
+	tid = threads_get_current_tid();
 #endif
 
-	} else {
+	if (logfile) {
 #if defined(ENABLE_THREADS)
-		fprintf(stdout, "LOG: [%p] ", (void *) THREADOBJECT);
+# if SIZEOF_VOID_P == 8
+		fprintf(logfile, "[0x%016lx] ", tid );
+# else
+		fprintf(logfile, "[0x%08x] ", tid);
+# endif
+#endif
+	}
+	else {
+#if defined(ENABLE_THREADS)
+# if SIZEOF_VOID_P == 8
+		fprintf(stdout, "LOG: [0x%016lx] ", tid);
+# else
+		fprintf(stdout, "LOG: [0x%08x] ", tid);
+# endif
 #else
 		fputs("LOG: ", stdout);
 #endif
@@ -155,11 +172,11 @@ void log_println(const char *text, ...)
 void log_finish(void)
 {
 	if (logfile) {
+		fputs("\n", logfile);
 		fflush(logfile);
-
-	} else {
+	}
+	else {
 		fputs("\n", stdout);
-
 		fflush(stdout);
 	}
 }
