@@ -1,6 +1,6 @@
 /* src/vm/builtin.c - functions for unsupported operations
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,22 +22,13 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Reinhard Grafl
-            Andreas Krall
-            Mark Probst
-
-   Changes: Christian Thalinger
-            Edwin Steiner
-
    Contains C functions for JavaVM Instructions that cannot be
    translated to machine language directly. Consequently, the
    generated machine code for these instructions contains function
    calls instead of machine instructions, using the C calling
    convention.
 
-   $Id: builtin.c 5862 2006-10-30 00:45:30Z edwin $
+   $Id: builtin.c 7986 2007-05-30 20:45:43Z twisti $
 
 */
 
@@ -46,7 +37,6 @@
 
 #include <assert.h>
 #include <errno.h>
-#include <math.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
@@ -61,32 +51,40 @@
 # undef Bias
 #endif
 
-#include "mm/boehm.h"
+#include "mm/gc-common.h"
 #include "mm/memory.h"
-#include "native/native.h"
-#include "native/include/java_lang_Cloneable.h"
-#include "native/include/java_lang_Object.h"          /* required by VMObject */
-#include "native/include/java_lang_VMObject.h"
 
-#if defined(ENABLE_THREADS)
-# include "threads/native/threads.h"
-#endif
+#include "native/jni.h"
+#include "native/include/java_lang_String.h"
+#include "native/include/java_lang_Throwable.h"
+
+#include "threads/lock-common.h"
 
 #include "toolbox/logging.h"
 #include "toolbox/util.h"
+
 #include "vm/builtin.h"
-#include "vm/class.h"
+#include "vm/cycles-stats.h"
 #include "vm/exceptions.h"
 #include "vm/global.h"
 #include "vm/initialize.h"
-#include "vm/loader.h"
-#include "vm/options.h"
 #include "vm/stringlocal.h"
+
 #include "vm/jit/asmpart.h"
 #include "vm/jit/patcher.h"
-#include "vm/rt-timing.h"
-#include "vm/cycles-stats.h"
 
+#include "vmcore/class.h"
+#include "vmcore/loader.h"
+#include "vmcore/options.h"
+#include "vmcore/rt-timing.h"
+
+#if defined(ENABLE_VMLOG)
+#include <vmlog_cacao.h>
+#endif
+
+#if defined(ENABLE_DEBUG_FILTER)
+#	include "vm/jit/show.h"
+#endif
 
 /* include builtin tables *****************************************************/
 
@@ -320,10 +318,14 @@ builtintable_entry *builtintable_get_automatic(s4 opcode)
 
 *******************************************************************************/
 
-bool builtintable_replace_function(instruction *iptr)
+#if defined(ENABLE_JIT)
+bool builtintable_replace_function(void *iptr_)
 {
 	constant_FMIref    *mr;
 	builtintable_entry *bte;
+	instruction        *iptr;
+
+	iptr = (instruction *) iptr_; /* twisti will kill me ;) */
 
 	/* get name and descriptor of the function */
 
@@ -365,112 +367,47 @@ bool builtintable_replace_function(instruction *iptr)
 
 	return false;
 }
+#endif /* defined(ENABLE_JIT) */
 
 
 /*****************************************************************************
 								TYPE CHECKS
 *****************************************************************************/
 
+/* builtin_instanceof **********************************************************
 
+   Checks if an object is an instance of some given class (or subclass
+   of that class). If class is an interface, checks if the interface
+   is implemented.
 
-/*************** internal function: builtin_isanysubclass *********************
-
-	Checks a subclass relation between two classes. Implemented interfaces
-	are interpreted as super classes.
-	Return value:  1 ... sub is subclass of super
-				   0 ... otherwise
-
-******************************************************************************/
-
-s4 builtin_isanysubclass(classinfo *sub, classinfo *super)
-{
-	s4 res;
-	castinfo classvalues;
-
-	if (sub == super)
-		return 1;
-
-	if (super->flags & ACC_INTERFACE) {
-		res = (sub->vftbl->interfacetablelength > super->index) &&
-			(sub->vftbl->interfacetable[-super->index] != NULL);
-
-	} else {
-		/* java.lang.Object is the only super_class_ of any interface */
-		if (sub->flags & ACC_INTERFACE)
-			return (super == class_java_lang_Object);
-
-		ASM_GETCLASSVALUES_ATOMIC(super->vftbl, sub->vftbl, &classvalues);
-
-		res = (u4) (classvalues.sub_baseval - classvalues.super_baseval) <=
-			(u4) classvalues.super_diffval;
-	}
-
-	return res;
-}
-
-
-s4 builtin_isanysubclass_vftbl(vftbl_t *sub, vftbl_t *super)
-{
-	s4 res;
-	s4 base;
-	castinfo classvalues;
-
-	if (sub == super)
-		return 1;
-
-	ASM_GETCLASSVALUES_ATOMIC(super, sub, &classvalues);
-
-	if ((base = classvalues.super_baseval) <= 0) {
-		/* super is an interface */
-		res = (sub->interfacetablelength > -base) &&
-			(sub->interfacetable[base] != NULL);
-	} 
-	else {
-		/* java.lang.Object is the only super_class_ of any interface */
-		if (classvalues.sub_baseval <= 0)
-			return classvalues.super_baseval == 1;
-
-	    res = (u4) (classvalues.sub_baseval - classvalues.super_baseval)
-			<= (u4) classvalues.super_diffval;
-	}
-
-	return res;
-}
-
-
-/****************** function: builtin_instanceof *****************************
-
-	Checks if an object is an instance of some given class (or subclass of
-	that class). If class is an interface, checks if the interface is
-	implemented.
-	Return value:  1 ... obj is an instance of class or implements the interface
-				   0 ... otherwise or if obj == NULL
+   Return value: 1 ... o is an instance of class or implements the interface
+                 0 ... otherwise or if o == NULL
 			 
-*****************************************************************************/
+*******************************************************************************/
 
-s4 builtin_instanceof(java_objectheader *obj, classinfo *class)
+s4 builtin_instanceof(java_objectheader *o, classinfo *class)
 {
-	if (!obj)
+	if (o == NULL)
 		return 0;
 
-	return builtin_isanysubclass(obj->vftbl->class, class);
+	return class_isanysubclass(o->vftbl->class, class);
 }
 
 
 
-/**************** function: builtin_checkcast *******************************
+/* builtin_checkcast ***********************************************************
 
-	The same as builtin_instanceof except that 1 is returned when
-	obj == NULL
+   The same as builtin_instanceof except that 1 is returned when o ==
+   NULL.
 			  
-****************************************************************************/
+*******************************************************************************/
 
-s4 builtin_checkcast(java_objectheader *obj, classinfo *class)
+s4 builtin_checkcast(java_objectheader *o, classinfo *class)
 {
-	if (obj == NULL)
+	if (o == NULL)
 		return 1;
 
-	if (builtin_isanysubclass(obj->vftbl->class, class))
+	if (class_isanysubclass(o->vftbl->class, class))
 		return 1;
 
 	return 0;
@@ -508,8 +445,8 @@ static s4 builtin_descriptorscompatible(arraydescriptor *desc,
 			(target->elementvftbl->baseval == 1))
 			return 1;
 
-		return builtin_isanysubclass_vftbl(desc->elementvftbl,
-										   target->elementvftbl);
+		return class_isanysubclass(desc->elementvftbl->class,
+								   target->elementvftbl->class);
 	}
 
 	if (desc->dimension < target->dimension)
@@ -517,8 +454,8 @@ static s4 builtin_descriptorscompatible(arraydescriptor *desc,
 
 	/* {desc has higher dimension than target} */
 
-	return builtin_isanysubclass_vftbl(pseudo_class_Arraystub->vftbl,
-									   target->elementvftbl);
+	return class_isanysubclass(pseudo_class_Arraystub,
+							   target->elementvftbl->class);
 }
 
 
@@ -539,10 +476,12 @@ s4 builtin_arraycheckcast(java_objectheader *o, classinfo *targetclass)
 {
 	arraydescriptor *desc;
 
-	if (!o)
+	if (o == NULL)
 		return 1;
 
-	if ((desc = o->vftbl->arraydesc) == NULL)
+	desc = o->vftbl->arraydesc;
+
+	if (desc == NULL)
 		return 0;
  
 	return builtin_descriptorscompatible(desc, targetclass->vftbl->arraydesc);
@@ -551,7 +490,7 @@ s4 builtin_arraycheckcast(java_objectheader *o, classinfo *targetclass)
 
 s4 builtin_arrayinstanceof(java_objectheader *o, classinfo *targetclass)
 {
-	if (!o)
+	if (o == NULL)
 		return 0;
 
 	return builtin_arraycheckcast(o, targetclass);
@@ -628,7 +567,7 @@ void *builtin_throw_exception(java_objectheader *xptr)
 
 	/* actually set the exception */
 
-	*exceptionptr = xptr;
+	exceptions_set_exception(xptr);
 
 	/* Return a NULL pointer.  This is required for vm_call_method to
 	   check for an exception.  This is for convenience. */
@@ -642,7 +581,7 @@ void *builtin_throw_exception(java_objectheader *xptr)
    Checks, if an object can be stored in an array.
 
    Return value: 1 ... possible
-                 0 ... otherwise
+                 0 ... otherwise (throws an ArrayStoreException)
 
 *******************************************************************************/
 
@@ -654,8 +593,9 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 	vftbl_t         *valuevftbl;
 	s4               base;
 	castinfo         classvalues;
+	s4               result;
 
-	if (!o)
+	if (o == NULL)
 		return 1;
 
 	/* The following is guaranteed (by verifier checks):
@@ -668,10 +608,9 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 	desc           = oa->header.objheader.vftbl->arraydesc;
 	componentvftbl = desc->componentvftbl;
 	valuevftbl     = o->vftbl;
+	valuedesc      = valuevftbl->arraydesc;
 
 	if ((desc->dimension - 1) == 0) {
-		s4 res;
-
 		/* {oa is a one-dimensional array} */
 		/* {oa is an array of references} */
 		
@@ -680,28 +619,41 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 
 		ASM_GETCLASSVALUES_ATOMIC(componentvftbl, valuevftbl, &classvalues);
 
-		if ((base = classvalues.super_baseval) <= 0)
-			/* an array of interface references */
-			return (valuevftbl->interfacetablelength > -base &&
-					valuevftbl->interfacetable[base] != NULL);
-		
-		res = ((unsigned) (classvalues.sub_baseval - classvalues.super_baseval)
-			   <= (unsigned) classvalues.super_diffval);
+		base = classvalues.super_baseval;
 
-		return res;
+		if (base <= 0) {
+			/* an array of interface references */
+
+			result = ((valuevftbl->interfacetablelength > -base) &&
+					(valuevftbl->interfacetable[base] != NULL));
+		}
+		else {
+			result = ((unsigned) (classvalues.sub_baseval - classvalues.super_baseval)
+				   <= (unsigned) classvalues.super_diffval);
+		}
+	}
+	else if (valuedesc == NULL) {
+		/* {oa has dimension > 1} */
+		/* {componentvftbl->arraydesc != NULL} */
+
+		/* check if o is an array */
+
+		return 0;
+	}
+	else {
+		/* {o is an array} */
+
+		result = builtin_descriptorscompatible(valuedesc, componentvftbl->arraydesc);
 	}
 
-	/* {oa has dimension > 1} */
-	/* {componentvftbl->arraydesc != NULL} */
+	/* if not possible, throw an exception */
 
-	/* check if o is an array */
+	if (result == 0)
+		exceptions_throw_arraystoreexception();
 
-	if ((valuedesc = valuevftbl->arraydesc) == NULL)
-		return 0;
+	/* return result */
 
-	/* {o is an array} */
-
-	return builtin_descriptorscompatible(valuedesc, componentvftbl->arraydesc);
+	return result;
 }
 
 
@@ -814,9 +766,7 @@ java_objectheader *builtin_new(classinfo *c)
 	/* check if we can instantiate this class */
 
 	if (c->flags & ACC_ABSTRACT) {
-		*exceptionptr =
-			new_exception_utfmessage(string_java_lang_InstantiationError,
-									 c->name);
+		exceptions_throw_instantiationerror(c);
 		return NULL;
 	}
 
@@ -891,14 +841,16 @@ java_arrayheader *builtin_newarray(s4 size, classinfo *arrayclass)
 
 	actualsize = dataoffset + size * componentsize;
 
-	if (((u4) actualsize) < ((u4) size)) { /* overflow */
-		*exceptionptr = new_exception(string_java_lang_OutOfMemoryError);
+	/* check for overflow */
+
+	if (((u4) actualsize) < ((u4) size)) {
+		exceptions_throw_outofmemoryerror();
 		return NULL;
 	}
 
 	a = heap_allocate(actualsize, (desc->arraytype == ARRAYTYPE_OBJECT), NULL);
 
-	if (!a)
+	if (a == NULL)
 		return NULL;
 
 	a->objheader.vftbl = arrayclass->vftbl;
@@ -1201,6 +1153,14 @@ java_objectheader *builtin_trace_exception(java_objectheader *xptr,
 	s4    dumpsize;
 	codeinfo *code;
 
+#if defined(ENABLE_DEBUG_FILTER)
+	if (! show_filters_test_verbosecall_exit(m)) return xptr;
+#endif
+
+#if defined(ENABLE_VMLOG)
+	return xptr;
+#endif
+
 	if (opt_verbosecall && indent)
 		methodindent--;
 
@@ -1337,7 +1297,6 @@ static char *builtin_print_argument(char *logtext, s4 *logtextlen,
 {
 	imm_union          imu;
 	java_objectheader *o;
-	java_lang_String  *s;
 	classinfo         *c;
 	utf               *u;
 	u4                 len;
@@ -1389,9 +1348,7 @@ static char *builtin_print_argument(char *logtext, s4 *logtextlen,
 				/* get java.lang.String object and the length of the
 				   string */
 
-				s = (java_lang_String *) o;
-
-				u = javastring_toutf(s, false);
+				u = javastring_toutf(o, false);
 
 				len = strlen(" (String = \"") + utf_bytes(u) + strlen("\")");
 
@@ -1443,8 +1400,7 @@ static char *builtin_print_argument(char *logtext, s4 *logtextlen,
 }
 #endif /* !defined(NDEBUG) */
 
-
-/* builtin_trace_args **********************************************************
+/* builtin_verbosecall_enter ***************************************************
 
    Print method call with arguments for -verbose:call.
 
@@ -1453,17 +1409,17 @@ static char *builtin_print_argument(char *logtext, s4 *logtextlen,
 #if !defined(NDEBUG)
 
 #ifdef TRACE_ARGS_NUM
-void builtin_trace_args(s8 a0, s8 a1,
-#if TRACE_ARGS_NUM >= 4
-						s8 a2, s8 a3,
-#endif /* TRACE_ARGS_NUM >= 4 */
-#if TRACE_ARGS_NUM >= 6
-						s8 a4, s8 a5,
-#endif /* TRACE_ARGS_NUM >= 6 */
-#if TRACE_ARGS_NUM == 8
-						s8 a6, s8 a7,
-#endif /* TRACE_ARGS_NUM == 8 */
-						methodinfo *m)
+void builtin_verbosecall_enter(s8 a0, s8 a1,
+# if TRACE_ARGS_NUM >= 4
+							   s8 a2, s8 a3,
+# endif
+# if TRACE_ARGS_NUM >= 6
+							   s8 a4, s8 a5,
+# endif
+# if TRACE_ARGS_NUM == 8
+							   s8 a6, s8 a7,
+# endif
+							   methodinfo *m)
 {
 	methoddesc *md;
 	char       *logtext;
@@ -1472,6 +1428,15 @@ void builtin_trace_args(s8 a0, s8 a1,
 	s4          i;
 	s4          pos;
 
+#if defined(ENABLE_DEBUG_FILTER)
+	if (! show_filters_test_verbosecall_enter(m)) return;
+#endif
+
+#if defined(ENABLE_VMLOG)
+	vmlog_cacao_enter_method(m);
+	return;
+#endif
+
 	md = m->parseddesc;
 
 	/* calculate message length */
@@ -1484,13 +1449,12 @@ void builtin_trace_args(s8 a0, s8 a1,
 		utf_bytes(m->class->name) +
 		strlen(".") +
 		utf_bytes(m->name) +
-		utf_bytes(m->descriptor) +
-		strlen("(") + strlen(")");
+		utf_bytes(m->descriptor);
 
 	/* Actually it's not possible to have all flags printed, but:
 	   safety first! */
 
-	logtextlen =
+	logtextlen +=
 		strlen(" PUBLIC") +
 		strlen(" PRIVATE") +
 		strlen(" PROTECTED") +
@@ -1506,8 +1470,10 @@ void builtin_trace_args(s8 a0, s8 a1,
 	/* add maximal argument length */
 
 	logtextlen +=
+		strlen("(") +
 		strlen("-9223372036854775808 (0x123456789abcdef0), ") * TRACE_ARGS_NUM +
-		strlen("...(255)");
+		strlen("...(255)") +
+		strlen(")");
 
 	/* allocate memory */
 
@@ -1620,19 +1586,20 @@ void builtin_trace_args(s8 a0, s8 a1,
 	dump_release(dumpsize);
 
 	methodindent++;
+
 }
 #endif
 #endif /* !defined(NDEBUG) */
 
 
-/* builtin_displaymethodstop ***************************************************
+/* builtin_verbosecall_exit ****************************************************
 
    Print method exit for -verbose:call.
 
 *******************************************************************************/
 
 #if !defined(NDEBUG)
-void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
+void builtin_verbosecall_exit(s8 l, double d, float f, methodinfo *m)
 {
 	methoddesc *md;
 	char       *logtext;
@@ -1642,6 +1609,15 @@ void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 	s4          pos;
 	imm_union   val;
 
+#if defined(ENABLE_DEBUG_FILTER)
+	if (! show_filters_test_verbosecall_exit(m)) return;
+#endif
+
+#if defined(ENABLE_VMLOG)
+	vmlog_cacao_leave_method(m);
+	return;
+#endif
+
 	md = m->parseddesc;
 
 	/* calculate message length */
@@ -1718,6 +1694,7 @@ void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 	/* release memory */
 
 	dump_release(dumpsize);
+
 }
 #endif /* !defined(NDEBUG) */
 
@@ -1748,7 +1725,7 @@ void builtin_print_cycles_stats(FILE *file)
 
 ******************************************************************************/
 
-#if !SUPPORT_DIVISION
+#if !SUPPORT_DIVISION || defined(DISABLE_GC)
 s4 builtin_idiv(s4 a, s4 b)
 {
 	s4 c;
@@ -1766,7 +1743,7 @@ s4 builtin_irem(s4 a, s4 b)
 
 	return c;
 }
-#endif /* !SUPPORT_DIVISION */
+#endif /* !SUPPORT_DIVISION || defined(DISABLE_GC) */
 
 
 /* functions for long arithmetics **********************************************
@@ -1834,7 +1811,7 @@ s8 builtin_lmul(s8 a, s8 b)
 #endif /* !(SUPPORT_LONG && SUPPORT_LONG_MUL) */
 
 
-#if !(SUPPORT_DIVISION && SUPPORT_LONG && SUPPORT_LONG_DIV)
+#if !(SUPPORT_DIVISION && SUPPORT_LONG && SUPPORT_LONG_DIV) || defined (DISABLE_GC)
 s8 builtin_ldiv(s8 a, s8 b)
 {
 	s8 c;
@@ -2114,17 +2091,26 @@ float builtin_fneg(float a)
 #endif /* !SUPPORT_FLOAT */
 
 
-#if !SUPPORT_FLOAT || defined(ENABLE_INTRP)
+#if !SUPPORT_FLOAT || !SUPPORT_FLOAT_CMP || defined(ENABLE_INTRP)
 s4 builtin_fcmpl(float a, float b)
 {
-	if (isnanf(a)) return -1;
-	if (isnanf(b)) return -1;
+	if (isnanf(a))
+		return -1;
+
+	if (isnanf(b))
+		return -1;
+
 	if (!finitef(a) || !finitef(b)) {
 		a = finitef(a) ? 0 : copysignf(1.0,	a);
 		b = finitef(b) ? 0 : copysignf(1.0, b);
 	}
-	if (a > b) return 1;
-	if (a == b) return 0;
+
+	if (a > b)
+		return 1;
+
+	if (a == b)
+		return 0;
+
 	return -1;
 }
 
@@ -2141,7 +2127,7 @@ s4 builtin_fcmpg(float a, float b)
 	if (a == b) return 0;
 	return -1;
 }
-#endif /* !SUPPORT_FLOAT || defined(ENABLE_INTRP) */
+#endif /* !SUPPORT_FLOAT || !SUPPORT_FLOAT_CMP || defined(ENABLE_INTRP) */
 
 
 float builtin_frem(float a, float b)
@@ -2293,34 +2279,52 @@ double builtin_dneg(double a)
 #endif /* !SUPPORT_DOUBLE */
 
 
-#if !SUPPORT_DOUBLE || defined(ENABLE_INTRP)
+#if !SUPPORT_DOUBLE || !SUPPORT_DOUBLE_CMP || defined(ENABLE_INTRP)
 s4 builtin_dcmpl(double a, double b)
 {
-	if (isnan(a)) return -1;
-	if (isnan(b)) return -1;
+	if (isnan(a))
+		return -1;
+
+	if (isnan(b))
+		return -1;
+
 	if (!finite(a) || !finite(b)) {
 		a = finite(a) ? 0 : copysign(1.0, a);
 		b = finite(b) ? 0 : copysign(1.0, b);
 	}
-	if (a > b) return 1;
-	if (a == b) return 0;
+
+	if (a > b)
+		return 1;
+
+	if (a == b)
+		return 0;
+
 	return -1;
 }
 
 
 s4 builtin_dcmpg(double a, double b)
 {
-	if (isnan(a)) return 1;
-	if (isnan(b)) return 1;
+	if (isnan(a))
+		return 1;
+
+	if (isnan(b))
+		return 1;
+
 	if (!finite(a) || !finite(b)) {
 		a = finite(a) ? 0 : copysign(1.0, a);
 		b = finite(b) ? 0 : copysign(1.0, b);
 	}
-	if (a > b) return 1;
-	if (a == b) return 0;
+
+	if (a > b)
+		return 1;
+
+	if (a == b)
+		return 0;
+
 	return -1;
 }
-#endif /* !SUPPORT_DOUBLE || defined(ENABLE_INTRP) */
+#endif /* !SUPPORT_DOUBLE || !SUPPORT_DOUBLE_CMP || defined(ENABLE_INTRP) */
 
 
 double builtin_drem(double a, double b)
@@ -2399,7 +2403,7 @@ double builtin_l2d(s8 a)
 #endif /* !(SUPPORT_LONG && SUPPORT_DOUBLE && SUPPORT_L2D) */
 
 
-#if !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP)
+#if !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP) || defined(DISABLE_GC)
 s4 builtin_f2i(float a) 
 {
 	s4 i;
@@ -2424,7 +2428,7 @@ s4 builtin_f2i(float a)
 		return 2147483647;
 		return (-2147483648); */
 }
-#endif /* !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP) */
+#endif /* !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP) || defined(DISABLE_GC) */
 
 
 #if !(SUPPORT_FLOAT && SUPPORT_LONG && SUPPORT_F2L)
@@ -2455,7 +2459,7 @@ s8 builtin_f2l(float a)
 #endif /* !(SUPPORT_FLOAT && SUPPORT_LONG && SUPPORT_F2L) */
 
 
-#if !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP)
+#if !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP) || defined(DISABLE_GC)
 s4 builtin_d2i(double a) 
 { 
 	double d;
@@ -2474,7 +2478,7 @@ s4 builtin_d2i(double a)
 		return 2147483647;
 	return (-2147483647-1);
 }
-#endif /* !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP) */
+#endif /* !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP) || defined(DISABLE_GC) */
 
 
 #if !(SUPPORT_DOUBLE && SUPPORT_LONG && SUPPORT_D2L)
@@ -2583,10 +2587,8 @@ bool builtin_arraycopy(java_arrayheader *src, s4 srcStart,
 			for (i = 0; i < len; i++) {
 				java_objectheader *o = oas->data[srcStart + i];
 
-				if (!builtin_canstore(oad, o)) {
-					exceptions_throw_arraystoreexception();
+				if (!builtin_canstore(oad, o))
 					return false;
-				}
 
 				oad->data[destStart + i] = o;
 			}
@@ -2601,10 +2603,8 @@ bool builtin_arraycopy(java_arrayheader *src, s4 srcStart,
 			for (i = len - 1; i >= 0; i--) {
 				java_objectheader *o = oas->data[srcStart + i];
 
-				if (!builtin_canstore(oad, o)) {
-					exceptions_throw_arraystoreexception();
+				if (!builtin_canstore(oad, o))
 					return false;
-				}
 
 				oad->data[destStart + i] = o;
 			}
@@ -2637,24 +2637,85 @@ s8 builtin_currenttimemillis(void)
 }
 
 
-/* builtin_clone_array *********************************************************
+/* builtin_clone ***************************************************************
 
-   Wrapper function for cloning arrays.
+   Function for cloning objects or arrays.
 
 *******************************************************************************/
 
-java_arrayheader *builtin_clone_array(void *env, java_arrayheader *o)
+java_objectheader *builtin_clone(void *env, java_objectheader *o)
 {
-	java_arrayheader    *ah;
-	java_lang_Cloneable *c;
+	arraydescriptor   *ad;
+	java_arrayheader  *ah;
+	u4                 size;
+	classinfo         *c;
+	java_objectheader *co;              /* cloned object header               */
 
-	c = (java_lang_Cloneable *) o;
+	/* get the array descriptor */
 
-	ah = (java_arrayheader *) Java_java_lang_VMObject_clone(0, 0, c);
+	ad = o->vftbl->arraydesc;
 
-	return ah;
+	/* we are cloning an array */
+
+	if (ad != NULL) {
+		ah = (java_arrayheader *) o;
+
+		size = ad->dataoffset + ad->componentsize * ah->size;
+        
+		co = heap_allocate(size, (ad->arraytype == ARRAYTYPE_OBJECT), NULL);
+
+		if (co == NULL)
+			return NULL;
+
+		MCOPY(co, o, u1, size);
+
+#if defined(ENABLE_GC_CACAO)
+		heap_init_objectheader(co, size);
+#endif
+
+#if defined(ENABLE_THREADS)
+		lock_init_object_lock(co);
+#endif
+
+		return co;
+	}
+    
+    /* we are cloning a non-array */
+
+    if (!builtin_instanceof(o, class_java_lang_Cloneable)) {
+        exceptions_throw_clonenotsupportedexception();
+        return NULL;
+    }
+
+	/* get the class of the object */
+
+    c = o->vftbl->class;
+
+	/* create new object */
+
+    co = builtin_new(c);
+
+    if (co == NULL)
+        return NULL;
+
+    MCOPY(co, o, u1, c->instancesize);
+
+#if defined(ENABLE_GC_CACAO)
+	heap_init_objectheader(co, c->instancesize);
+#endif
+
+#if defined(ENABLE_THREADS)
+	lock_init_object_lock(co);
+#endif
+
+    return co;
 }
 
+#if defined(ENABLE_VMLOG)
+#define NDEBUG
+#include <vmlog_cacao.c>
+#endif
+
 
 /*
  * These are local overrides for various environment variables in Emacs.
