@@ -1,6 +1,6 @@
 /* src/vm/jit/abi.h - common ABI defines
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: abi.h 5079 2006-07-06 11:36:01Z twisti $
+   $Id: abi.h 7938 2007-05-23 09:37:37Z tbfg $
 
 */
 
@@ -40,7 +34,9 @@
 #include "vm/types.h"
 
 #include "vm/jit/abi-asm.h"
+#include "vm/jit/jit.h"
 
+#include "arch.h"
 
 /* ABI externs ****************************************************************/
 
@@ -48,9 +44,32 @@ extern s4 nregdescint[];
 extern char *regs[];
 extern s4 nregdescfloat[];
 
+#if defined(HAS_ADDRESS_REGISTER_FILE)
+extern s4 nregdescadr[];
+
+extern const char *abi_registers_address_name[];
+extern const s4    abi_registers_address_argument[];
+extern const s4    abi_registers_address_saved[];
+extern const s4    abi_registers_address_temporary[];
+#endif
+
+extern const char *abi_registers_integer_name[];
+extern const s4    abi_registers_integer_argument[];
+extern const s4    abi_registers_integer_saved[];
+extern const s4    abi_registers_integer_temporary[];
+
+extern const char *abi_registers_float_name[];
+extern const s4    abi_registers_float_argument[];
+extern const s4    abi_registers_float_saved[];
+extern const s4    abi_registers_float_temporary[];
+
 
 /* function prototypes ********************************************************/
 
+/* machine dependent descriptor function */
+void md_param_alloc(methoddesc *md);
+void md_param_alloc_native(methoddesc *md);
+
 /* machine dependent return value handling function */
 void md_return_alloc(jitdata *jd, stackptr stackslot);
 
