@@ -1,6 +1,6 @@
 /* src/vm/jit/allocator/simplereg.c - register allocator
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
+   Copyright (C) 1996-2005, 2007 R. Grafl, A. Krall, C. Kruegel, C. Oates,
    R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
    C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
    Institut f. Computersprachen - TU Wien
@@ -22,39 +22,34 @@
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
-
-   Authors: Andreas Krall
-
-   Changes: Stefan Ring
-            Christian Thalinger
-            Christian Ullrich
-            Michael Starzinger
-            Edwin Steiner
-
-   $Id: simplereg.c 5789 2006-10-16 09:02:40Z twisti $
+   $Id: simplereg.c 7743 2007-04-17 20:53:41Z edwin $
 
 */
 
 
 #include "config.h"
-#include "vm/types.h"
 
 #include <assert.h>
 
+#include "vm/types.h"
+
 #include "arch.h"
 #include "md-abi.h"
 
+#include "mm/memory.h"
+
 #include "vm/builtin.h"
 #include "vm/exceptions.h"
-#include "mm/memory.h"
-#include "vm/method.h"
-#include "vm/options.h"
 #include "vm/resolve.h"
 #include "vm/stringlocal.h"
+
+#include "vm/jit/abi.h"
 #include "vm/jit/reg.h"
-#include "vm/jit/allocator/simplereg.h"
 #include "vm/jit/show.h"
+#include "vm/jit/allocator/simplereg.h"
+
+#include "vmcore/method.h"
+#include "vmcore/options.h"
 
 
 #if 0
@@ -165,7 +160,7 @@ static void simplereg_allocate_temporaries(jitdata *jd);
 #define AVAIL_FREE_TMP_INT  AVAIL_BACK_INT(rd->freetmpinttop)
 #define AVAIL_FREE_SAV_INT  AVAIL_BACK_INT(rd->freesavinttop)
 
-#define TAKE_ARG_FLT(r)  POP_FRONT(rd->argfltregs, rd->argfltreguse, r)
+#define TAKE_ARG_FLT(r)  POP_FRONT(abi_registers_float_argument, rd->argfltreguse, r)
 #define TAKE_TMP_FLT(r)  POP_BACK(rd->tmpfltregs, rd->tmpfltreguse, r)
 #define TAKE_SAV_FLT(r)  POP_BACK(rd->savfltregs, rd->savfltreguse, r)
 
@@ -173,7 +168,7 @@ static void simplereg_allocate_temporaries(jitdata *jd);
 #define TAKE_TMP_ADR(r)  POP_BACK(rd->tmpadrregs, rd->tmpadrreguse, r)
 #define TAKE_SAV_ADR(r)  POP_BACK(rd->savadrregs, rd->savadrreguse, r)
 
-#define TAKE_ARG_INT(r)  POP_FRONT_INT(rd->argintregs, rd->argintreguse, r)
+#define TAKE_ARG_INT(r)  POP_FRONT_INT(abi_registers_integer_argument, rd->argintreguse, r)
 #define TAKE_TMP_INT(r)  POP_BACK_INT(rd->tmpintregs, rd->tmpintreguse, r)
 #define TAKE_SAV_INT(r)  POP_BACK_INT(rd->savintregs, rd->savintreguse, r)
 
@@ -359,7 +354,7 @@ static void simplereg_allocate_interfaces(jitdata *jd)
 #endif
 	}
 
-	for (s = 0; s < cd->maxstack; s++) {
+	for (s = 0; s < jd->maxinterfaces; s++) {
 		intalloc = -1; fltalloc = -1;
 
 		/* check if the interface at this stack depth must be a SAVEDVAR */
@@ -569,7 +564,7 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 	registerdata *rd;
 	methoddesc *md;
 
-	int     p, s, t, tt, lm;
+	int     p, s, t, tt, varindex;
 	int     intalloc, fltalloc;
 	varinfo *v;
 	int     intregsneeded = 0;
@@ -593,15 +588,15 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 #ifdef HAS_ADDRESS_REGISTER_FILE
 	aargcnt = rd->argadrreguse;
 #endif
-	for (p = 0, s = 0; s < cd->maxlocals; s++, p++) {
+	for (p = 0, s = 0; s < jd->maxlocals; s++, p++) {
 		intalloc = -1; fltalloc = -1;
 		for (tt = 0; tt <= 4; tt++) {
 			t = typeloop[tt];
-			lm = jd->local_map[s * 5 + t];
-			if (lm == UNUSED)
+			varindex = jd->local_map[s * 5 + t];
+			if (varindex == UNUSED)
 				continue;
 
-			v = VAR(lm);
+			v = VAR(varindex);
 
 #if defined(SUPPORT_COMBINE_INTEGER_REGISTERS)
 			intregsneeded = (IS_2_WORD_TYPE(t)) ? 1 : 0;
@@ -663,11 +658,9 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 #if !defined(SUPPORT_PASS_FLOATARGS_IN_INTREGS)
 					/* We can only use float arguments as local variables,
 					 * if we do not pass them in integer registers. */
-					else if ((p < md->paramcount) &&
-							 !md->params[p].inmemory) 
-					{
+					else if ((p < md->paramcount) && !md->params[p].inmemory) {
 						v->flags = 0;
-						v->vv.regoff = rd->argfltregs[md->params[p].regoff];
+						v->vv.regoff = md->params[p].regoff;
 					}
 #endif
 					else if (AVAIL_TMP_FLT) {
@@ -675,11 +668,10 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 						TAKE_TMP_FLT(v->vv.regoff);
 					}
 					/* use unused argument registers as local registers */
-					else if ((p >= md->paramcount) &&
-							 (fargcnt < FLT_ARG_CNT)) 
-					{
+					else if ((p >= md->paramcount) && (fargcnt < FLT_ARG_CNT)) {
 						v->flags = 0;
-						POP_FRONT(rd->argfltregs, fargcnt, v->vv.regoff);
+						POP_FRONT(abi_registers_float_argument,
+								  fargcnt, v->vv.regoff);
 					}
 					else if (AVAIL_SAV_FLT) {
 						v->flags = 0;
@@ -721,13 +713,12 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 							v->flags = 0;
 #if defined(SUPPORT_COMBINE_INTEGER_REGISTERS)
 							if (IS_2_WORD_TYPE(t))
-								v->vv.regoff = PACK_REGS(
-						rd->argintregs[GET_LOW_REG(md->params[p].regoff)],
-						rd->argintregs[GET_HIGH_REG(md->params[p].regoff)]);
+								v->vv.regoff =
+									PACK_REGS(GET_LOW_REG(md->params[p].regoff),
+											  GET_HIGH_REG(md->params[p].regoff));
 								else
 #endif
-									v->vv.regoff =
-									   rd->argintregs[md->params[p].regoff];
+									v->vv.regoff = md->params[p].regoff;
 						}
 						else if (AVAIL_TMP_INT) {
 							v->flags = 0;
@@ -740,7 +731,8 @@ static void simplereg_allocate_locals_leafmethod(jitdata *jd)
 								 (iargcnt + intregsneeded < INT_ARG_CNT)) 
 						{
 							v->flags = 0;
-							POP_FRONT_INT(rd->argintregs, iargcnt, v->vv.regoff);
+							POP_FRONT_INT(abi_registers_integer_argument,
+										  iargcnt, v->vv.regoff);
 						}
 						else if (AVAIL_SAV_INT) {
 							v->flags = 0;
@@ -778,7 +770,7 @@ static void simplereg_allocate_locals(jitdata *jd)
 	codegendata  *cd;
 	registerdata *rd;
 
-	int     s, t, tt, lm;
+	int     s, t, tt, varindex;
 	int     intalloc, fltalloc;
 	varinfo *v;
 	int     memneeded = 0;
@@ -797,16 +789,16 @@ static void simplereg_allocate_locals(jitdata *jd)
 		return;
 	}
 
-	for (s = 0; s < cd->maxlocals; s++) {
+	for (s = 0; s < jd->maxlocals; s++) {
 		intalloc = -1; fltalloc = -1;
 		for (tt=0; tt<=4; tt++) {
 			t = typeloop[tt];
 
-			lm = jd->local_map[s * 5 + t];
-			if (lm == UNUSED)
+			varindex = jd->local_map[s * 5 + t];
+			if (varindex == UNUSED)
 				continue;
 
-			v = VAR(lm);
+			v = VAR(varindex);
 
 #ifdef SUPPORT_COMBINE_INTEGER_REGISTERS
 				intregsneeded = (IS_2_WORD_TYPE(t)) ? 1 : 0;
@@ -928,14 +920,14 @@ static void simplereg_init(jitdata *jd, registerdata *rd)
 	/* record the interface registers as used */
 
 	for (i=0; i<rd->argintreguse; ++i)
-		rd->intusedinout[rd->argintregs[i]] = 1;
+		rd->intusedinout[abi_registers_integer_argument[i]] = 1;
 	for (i=rd->tmpintreguse; i<INT_TMP_CNT; ++i)
 		rd->intusedinout[rd->tmpintregs[i]] = 1;
 	for (i=rd->savintreguse; i<INT_SAV_CNT; ++i)
 		rd->intusedinout[rd->savintregs[i]] = 1;
 
 	for (i=0; i<rd->argfltreguse; ++i)
-		rd->fltusedinout[rd->argfltregs[i]] = 1;
+		rd->fltusedinout[abi_registers_float_argument[i]] = 1;
 	for (i=rd->tmpfltreguse; i<FLT_TMP_CNT; ++i)
 		rd->fltusedinout[rd->tmpfltregs[i]] = 1;
 	for (i=rd->savfltreguse; i<FLT_SAV_CNT; ++i)
@@ -1017,6 +1009,10 @@ static void simplereg_new_temp(jitdata *jd, s4 index)
 	rd = jd->rd;
 	v = VAR(index);
 
+	/* assert that constants are not allocated */
+
+	assert(v->type != TYPE_RET);
+
 	/* Try to allocate a saved register if there is no temporary one          */
 	/* available. This is what happens during the second run.                 */
 	tryagain = (v->flags & SAVEDVAR) ? 1 : 2;
@@ -1165,6 +1161,10 @@ static void simplereg_new_temp(jitdata *jd, s4 index)
 
 static void simplereg_free(registerdata *rd, s4 flags, s4 regoff, s4 type)
 {
+	/* assert that constants are not freed */
+
+	assert(type != TYPE_RET);
+
 	/* if this is a copy of another variable, just decrement the copy counter */
 
 	/* XXX split reg/mem variables on arm may need special handling here */
@@ -1350,6 +1350,8 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 			for (i=0; i<bptr->indepth; ++i) 
 			{
 				v = VAR(bptr->invars[i]);
+				if (v->type == TYPE_RET)
+					continue;
 
 				v->vv.regoff = jd->interface_map[5*i + v->type].regoff;
 				v->flags  = jd->interface_map[5*i + v->type].flags;
@@ -1363,6 +1365,8 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 			for (i=0; i<bptr->outdepth; ++i) 
 			{
 				v = VAR(bptr->outvars[i]);
+				if (v->type == TYPE_RET)
+					continue;
 
 				v->vv.regoff = jd->interface_map[5*i + v->type].regoff;
 				v->flags  = jd->interface_map[5*i + v->type].flags;
@@ -1376,7 +1380,7 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 
 			/* free interface registers not used in this block */
 
-			for (i=0; i < 5 * m->maxstack; ++i) {
+			for (i=0; i < 5 * jd->maxinterfaces; ++i) {
 				type = i%5;
 				regoff = jd->interface_map[i].regoff;
 				flags = jd->interface_map[i].flags;
@@ -1409,10 +1413,6 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 					/* pop 0 push 0 */
 
 				case ICMD_JSR:
-#if !defined(NDEBUG)
-					/* avoid problems with show_allocation */
-					VAROP(iptr->dst)->vv.regoff = 0;
-#endif
 				case ICMD_NOP:
 				case ICMD_CHECKNULL:
 				case ICMD_IINC:
@@ -1422,7 +1422,7 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 				case ICMD_PUTSTATICCONST:
 				case ICMD_INLINE_START:
 				case ICMD_INLINE_END:
-				case ICMD_INLINE_GOTO:
+				case ICMD_INLINE_BODY:
 					break;
 
 					/* pop 0 push 1 const */
@@ -1791,9 +1791,8 @@ static void simplereg_allocate_temporaries(jitdata *jd)
 					break;
 
 				default:
-					*exceptionptr =
-						new_internalerror("Unknown ICMD %d during register allocation",
-										  iptr->opc);
+					exceptions_throw_internalerror("Unknown ICMD %d during register allocation",
+												   iptr->opc);
 					return;
 				} /* switch */
 				iptr++;
@@ -1810,11 +1809,13 @@ void simplereg_make_statistics(jitdata *jd)
 	methodinfo   *m;
 	codegendata  *cd;
 	registerdata *rd;
-	int i,type;
+	int i;
 	s4 len;
+#if 0
 	stackptr    src, src_old;
 	stackptr    dst;
 	instruction *iptr;
+#endif
 	basicblock  *bptr;
 	int size_interface; /* == maximum size of in/out stack at basic block boundaries */
 	bool in_register;
