@@ -1,6 +1,6 @@
 /* src/vm/jit/code.c - codeinfo struct for representing compiled code
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,27 +22,49 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Edwin Steiner
-
-   Changes: Christian Thalinger
-
    $Id$
 
 */
 
 
 #include "config.h"
-#include "vm/types.h"
 
 #include <assert.h>
 
-#include "vm/jit/code.h"
-#include "mm/memory.h"
-#include "vm/options.h"
+#include "vm/types.h"
+
 #include "arch.h"
 
+#include "mm/memory.h"
+
+#if defined(ENABLE_THREADS)
+# include "threads/native/lock.h"
+#endif
+
+#include "vm/jit/code.h"
+#include "vm/jit/codegen-common.h"
+#include "vm/jit/methodheader.h"
+
+#include "vmcore/options.h"
+
+
+/* code_init *******************************************************************
+
+   Initialize the code-subsystem.
+
+*******************************************************************************/
+
+bool code_init(void)
+{
+	/* check for offset of code->m == 0 (see comment in code.h) */
+
+	assert(OFFSET(codeinfo, m) == 0);
+
+	/* everything's ok */
+
+	return true;
+}
+
 
 /* code_codeinfo_new ***********************************************************
 
@@ -79,6 +101,52 @@ codeinfo *code_codeinfo_new(methodinfo *m)
 }
 
 
+/* code_find_codeinfo_for_pc ***************************************************
+
+   Return the codeinfo for the compilation unit that contains the
+   given PC.
+
+   IN:
+       pc...............machine code position
+
+   RETURN VALUE:
+       the codeinfo * for the given PC
+
+*******************************************************************************/
+
+codeinfo *code_find_codeinfo_for_pc(u1 *pc)
+{
+	u1 *pv;
+
+	pv = codegen_get_pv_from_pc(pc);
+	assert(pv);
+
+	return *(codeinfo **)(pv + CodeinfoPointer);
+}
+
+
+/* code_get_methodinfo_for_pv **************************************************
+
+   Return the methodinfo for the given PV.
+
+   IN:
+       pv...............PV
+
+   RETURN VALUE:
+       the methodinfo *
+
+*******************************************************************************/
+
+methodinfo *code_get_methodinfo_for_pv(u1 *pv)
+{
+	codeinfo *code;
+
+	code = *((codeinfo **) (pv + CodeinfoPointer));
+
+	return code->m;
+}
+
+
 /* code_get_sync_slot_count ****************************************************
 
    Return the number of stack slots used for storing the synchronized object
@@ -93,6 +161,7 @@ codeinfo *code_codeinfo_new(methodinfo *m)
   
 *******************************************************************************/
 
+#if defined(ENABLE_REPLACEMENT)
 int code_get_sync_slot_count(codeinfo *code)
 {
 #ifdef ENABLE_THREADS
@@ -127,6 +196,7 @@ int code_get_sync_slot_count(codeinfo *code)
 
 #endif /* ENABLE_THREADS */
 }
+#endif /* defined(ENABLE_REPLACEMENT) */
 
 
 /* code_get_stack_frame_size ***************************************************
@@ -150,6 +220,7 @@ int code_get_sync_slot_count(codeinfo *code)
   
 *******************************************************************************/
 
+#if defined(ENABLE_REPLACEMENT)
 int code_get_stack_frame_size(codeinfo *code)
 {
 #if 0
@@ -193,8 +264,9 @@ int code_get_stack_frame_size(codeinfo *code)
 	return count;
 #endif
 
-	return 0;
+	return code->stackframesize;
 }
+#endif /* defined(ENABLE_REPLACEMENT) */
 
 
 /* code_codeinfo_free **********************************************************
@@ -214,7 +286,9 @@ void code_codeinfo_free(codeinfo *code)
 	if (code->mcode != NULL)
 		CFREE((void *) (ptrint) code->mcode, code->mcodelength);
 
+#if defined(ENABLE_REPLACEMENT)
 	replace_free_replacement_points(code);
+#endif
 
 	FREE(code, codeinfo);
 
