@@ -1,6 +1,6 @@
 /* src/vm/jit/disass.h - disassembler header
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: disass.h 4709 2006-03-30 10:14:22Z twisti $
+   $Id: disass.h 7596 2007-03-28 21:05:53Z twisti $
 
 */
 
@@ -94,9 +88,7 @@ extern disassemble_info info;
 extern bool disass_initialized;
 #endif
 
-extern char *regs[];
-
-#if defined(__I386__) || defined(__X86_64__)
+#if defined(__I386__) || defined(__X86_64__) || defined(__S390__) || defined(__M68K__)
 extern char disass_buf[512];
 extern s4   disass_len;
 #endif
