@@ -1,6 +1,6 @@
 /* src/vm/jit/dseg.c - data segment handling stuff
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -26,12 +26,11 @@
 
    Authors: Reinhard Grafl
             Andreas  Krall
-
-   Changes: Christian Thalinger
+            Christian Thalinger
             Joseph Wenninger
-			Edwin Steiner
+            Edwin Steiner
 
-   $Id: dseg.c 5868 2006-10-30 11:21:36Z edwin $
+   $Id: dseg.c 7596 2007-03-28 21:05:53Z twisti $
 
 */
 
@@ -40,12 +39,12 @@
 
 #include <assert.h>
 
-#include "vm/options.h"
 #include "vm/types.h"
 
 #include "mm/memory.h"
 #include "vm/jit/codegen-common.h"
-
+#include "vm/jit/methodheader.h"
+#include "vmcore/options.h"
 
 
 /* dseg_finish *****************************************************************
@@ -645,7 +644,7 @@ void dseg_addlinenumber(codegendata *cd, u2 linenumber)
 
    IN:
       cd.............current codegen data
-      iptr...........the ICMD_INLINE_START instruction
+      iptr...........the ICMD_INLINE_BODY instruction
       mcodeptr.......start mcodeptr of inlined body
 
 *******************************************************************************/
@@ -743,13 +742,119 @@ void dseg_createlinenumbertable(codegendata *cd)
 }
 
 
+/* dseg_get_linenumber_from_pc_intern ******************************************
+
+   This function search the line number table for the line
+   corresponding to a given pc. The function recurses for inlined
+   methods.
+
+*******************************************************************************/
+
+static s4 dseg_get_linenumber_from_pc_intern(methodinfo **pm, linenumbertable_entry *lntentry, s4 lntsize, u1 *pc)
+{
+	linenumbertable_entry *lntinline;     /* special entry for inlined method */
+
+	for (; lntsize > 0; lntsize--, lntentry--) {
+		/* Note: In case of inlining this may actually compare the pc
+		   against a methodinfo *, yielding a non-sensical
+		   result. This is no problem, however, as we ignore such
+		   entries in the switch below. This way we optimize for the
+		   common case (ie. a real pc in lntentry->pc). */
+
+		if (pc >= lntentry->pc) {
+			/* did we reach the current line? */
+
+			if ((s4) lntentry->line >= 0)
+				return (s4) lntentry->line;
+
+			/* we found a special inline entry (see
+			   doc/inlining_stacktrace.txt for details */
+
+			switch (lntentry->line) {
+			case -1: 
+				/* begin of inlined method (ie. INLINE_END
+				   instruction) */
+
+				lntinline = --lntentry;/* get entry with methodinfo * */
+				lntentry--;            /* skip the special entry      */
+				lntsize -= 2;
+
+				/* search inside the inlined method */
+
+				if (dseg_get_linenumber_from_pc_intern(pm, lntentry, lntsize,
+													   pc))
+				{
+					/* the inlined method contained the pc */
+
+					*pm = (methodinfo *) lntinline->pc;
+
+					assert(lntinline->line <= -3);
+
+					return (-3) - lntinline->line;
+				}
+
+				/* pc was not in inlined method, continue search.
+				   Entries inside the inlined method will be skipped
+				   because their lntentry->pc is higher than pc.  */
+				break;
+
+			case -2: 
+				/* end of inlined method */
+
+				return 0;
+
+				/* default: is only reached for an -3-line entry after
+				   a skipped -2 entry. We can safely ignore it and
+				   continue searching.  */
+			}
+		}
+	}
+
+	/* not found, return 0 */
+
+	return 0;
+}
+
+
+/* dseg_get_linenumber_from_pc *************************************************
+
+   A wrapper for dseg_get_linenumber_from_pc_intern, so we don't have
+   to evaluate the method header on every call.
+
+*******************************************************************************/
+
+s4 dseg_get_linenumber_from_pc(methodinfo **pm, u1 *pv, u1 *pc)
+{
+	ptrint                 lntsize;     /* size of line number table          */
+	u1                    *lntstart;    /* start of line number table         */
+	linenumbertable_entry *lntentry;    /* points to last entry in the table  */
+	s4                     linenumber;
+
+	/* get size of line number table */
+
+	lntsize  = *((ptrint *) (pv + LineNumberTableSize));
+	lntstart = *((u1 **)    (pv + LineNumberTableStart));
+
+	/* Subtract the size of the line number entry of the structure,
+	   since the line number table start points to the pc. */
+
+	lntentry = (linenumbertable_entry *) (lntstart - SIZEOF_VOID_P);
+
+	/* get the line number */
+
+	linenumber = dseg_get_linenumber_from_pc_intern(pm, lntentry, lntsize, pc);
+
+	return linenumber;
+}
+
+
 /* dseg_adddata ****************************************************************
 
    Adds a data segment reference to the codegendata.
 
 *******************************************************************************/
 
-#if defined(__I386__) || defined(__X86_64__) || defined(__XDSPCORE__) || defined(ENABLE_INTRP)
+#if defined(__I386__) || defined(__X86_64__) || defined(__S390__) || defined(__XDSPCORE__) || defined(__M68K__) || defined(ENABLE_INTRP)
 void dseg_adddata(codegendata *cd)
 {
 	dataref *dr;
@@ -770,7 +875,7 @@ void dseg_adddata(codegendata *cd)
 
 *******************************************************************************/
 
-#if defined(__I386__) || defined(__X86_64__) || defined(__XDSPCORE__) || defined(ENABLE_INTRP)
+#if defined(__I386__) || defined(__X86_64__) || defined(__XDSPCORE__) || defined(__M68K__) || defined(ENABLE_INTRP)
 void dseg_resolve_datareferences(jitdata *jd)
 {
 	codeinfo    *code;
