@@ -1,6 +1,6 @@
 /* src/vm/jit/dseg.c - data segment handling stuff
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,15 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Reinhard Grafl
-            Andreas  Krall
-
-   Changes: Christian Thalinger
-            Joseph Wenninger
-
-   $Id: dseg.h 5785 2006-10-15 22:25:54Z edwin $
+   $Id: dseg.h 7903 2007-05-14 11:15:33Z tbfg $
 
 */
 
@@ -40,31 +32,20 @@
 
 /* forward typedefs ***********************************************************/
 
-typedef struct dsegentry dsegentry;
-typedef struct jumpref jumpref;
-typedef struct dataref dataref;
-typedef struct exceptionref exceptionref;
-typedef struct patchref patchref;
-typedef struct linenumberref linenumberref;
-typedef struct dseg_exception_entry dseg_exception_entry;
+typedef struct dsegentry             dsegentry;
+typedef struct linenumbertable_entry linenumbertable_entry;
+typedef struct dseg_exception_entry  dseg_exception_entry;
 
 
 #include "config.h"
-
 #include "vm/types.h"
 
+#include "toolbox/list.h"
+
 #include "vm/jit/jit.h"
 #include "vm/jit/codegen-common.h"
 
-
-/* XXX don't-break-trunk macros ***********************************************/
-
-#define dseg_adds4(cd,value)         dseg_add_unique_s4((cd), (value))
-#define dseg_adds8(cd,value)         dseg_add_unique_s8((cd), (value))
-#define dseg_addfloat(cd,value)      dseg_add_unique_float((cd), (value))
-#define dseg_adddouble(cd,value)     dseg_add_unique_double((cd), (value))
-#define dseg_addaddress(cd,value)    dseg_add_unique_address((cd), (void *) (ptrint) (value))
-#define dseg_addtarget(cd,value)     dseg_add_target((cd), (value))
+#include "vmcore/references.h"
 
 
 /* convenience macros *********************************************************/
@@ -87,59 +68,14 @@ struct dsegentry {
 };
 
 
-/* jumpref ********************************************************************/
+/* linenumbertable_entry ******************************************************/
 
-struct jumpref {
-	s4          tablepos;       /* patching position in data segment          */
-	basicblock *target;         /* target basic block                         */
-	jumpref    *next;           /* next element in jumpref list               */
-};
+/* Keep the type of line the same as the pointer type, otherwise we
+   run into alignment troubles (like on MIPS64). */
 
-
-/* dataref ********************************************************************/
-
-struct dataref {
-	s4       datapos;           /* patching position in generated code        */
-	dataref *next;              /* next element in dataref list               */
-};
-
-
-/* exceptionref ***************************************************************/
-
-struct exceptionref {
-	s4            branchpos;    /* patching position in code segment          */
-	s4            reg;          /* used for ArrayIndexOutOfBounds index reg   */
-	functionptr   function;     /* function pointer to generate exception     */
-	exceptionref *next;         /* next element in exceptionref list          */
-};
-
-
-/* patchref *******************************************************************/
-
-struct patchref {
-	s4           branchpos;
-	functionptr  patcher;
-	voidptr      ref;
-	patchref    *next;
-	s4           disp;
-};
-
-
-/* linenumberref **************************************************************/
-
-struct linenumberref {
-	s4             tablepos;    /* patching position in data segment          */
-	s4             linenumber;  /* line number, used for inserting into the   */
-	                            /* table and for validity checking            */
-	                            /* -1......start of inlined body              */
-	                            /* -2......end of inlined body                */
-	                            /* <= -3...special entry with methodinfo *    */
-								/* (see doc/inlining_stacktrace.txt)          */
-	ptrint         targetmpc;   /* machine code program counter of first      */
-	                            /* instruction for given line                 */
-								/* NOTE: for linenumber <= -3 this is a the   */
-	                            /* (methodinfo *) of the inlined method       */
-	linenumberref *next;        /* next element in linenumberref list         */
+struct linenumbertable_entry {
+	ptrint  line;               /* NOTE: see doc/inlining_stacktrace.txt for  */
+	u1     *pc;                 /*       special meanings of line and pc.     */
 };
 
 
@@ -184,7 +120,9 @@ void dseg_addlinenumber_inline_end(codegendata *cd, instruction *iptr);
 
 void dseg_createlinenumbertable(codegendata *cd);
 
-#if defined(__I386__) || defined(__X86_64__) || defined(__XDSPCORE__) || defined(ENABLE_INTRP)
+s4 dseg_get_linenumber_from_pc(methodinfo **pm, u1 *pv, u1 *pc);
+
+#if defined(__I386__) || defined(__X86_64__) || defined(__XDSPCORE__) || defined(__M68K__) || defined(ENABLE_INTRP)
 void dseg_adddata(codegendata *cd);
 void dseg_resolve_datareferences(jitdata *jd);
 #endif
