@@ -1,6 +1,6 @@
 /* src/vm/jit/i386/emit.c - i386 code emitter functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: emit.c 5562 2006-09-28 20:13:20Z edwin $
+   $Id: emit.c 7818 2007-04-25 19:47:50Z twisti $
 
 */
 
@@ -43,18 +37,22 @@
 #include "vm/jit/i386/emit.h"
 #include "vm/jit/i386/md-abi.h"
 
-#if defined(ENABLE_THREADS)
-# include "threads/native/lock.h"
-#endif
+#include "mm/memory.h"
+
+#include "threads/lock-common.h"
 
 #include "vm/builtin.h"
-#include "vm/statistics.h"
+#include "vm/exceptions.h"
+
 #include "vm/jit/asmpart.h"
 #include "vm/jit/dseg.h"
 #include "vm/jit/emit-common.h"
 #include "vm/jit/jit.h"
 #include "vm/jit/replace.h"
 
+#include "vmcore/options.h"
+#include "vmcore/statistics.h"
+
 
 /* emit_load ******************************************************************
 
@@ -77,17 +75,22 @@ inline s4 emit_load(jitdata *jd, instruction *iptr, varinfo *src, s4 tempreg)
 
 		disp = src->vv.regoff * 4;
 
-		if (IS_FLT_DBL_TYPE(src->type)) {
-			if (IS_2_WORD_TYPE(src->type))
-				M_DLD(tempreg, REG_SP, disp);
-			else
-				M_FLD(tempreg, REG_SP, disp);
-		}
-		else {
-			if (IS_2_WORD_TYPE(src->type))
-				M_LLD(tempreg, REG_SP, disp);
-			else
-				M_ILD(tempreg, REG_SP, disp);
+		switch (src->type) {
+		case TYPE_INT:
+		case TYPE_ADR:
+			M_ILD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_LNG:
+			M_LLD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_FLT:
+			M_FLD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_DBL:
+			M_DLD(tempreg, REG_SP, disp);
+			break;
+		default:
+			vm_abort("emit_load: unknown type %d", src->type);
 		}
 
 		reg = tempreg;
@@ -177,6 +180,7 @@ inline s4 emit_load_high(jitdata *jd, instruction *iptr,varinfo *src,s4 tempreg)
 inline void emit_store(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 {
 	codegendata  *cd;
+	s4            disp;
 
 	/* get required compiler data */
 
@@ -185,17 +189,24 @@ inline void emit_store(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 	if (IS_INMEMORY(dst->flags)) {
 		COUNT_SPILLS;
 
-		if (IS_FLT_DBL_TYPE(dst->type)) {
-			if (IS_2_WORD_TYPE(dst->type))
-				M_DST(d, REG_SP, dst->vv.regoff * 4);
-			else
-				M_FST(d, REG_SP, dst->vv.regoff * 4);
-		}
-		else {
-			if (IS_2_WORD_TYPE(dst->type))
-				M_LST(d, REG_SP, dst->vv.regoff * 4);
-			else
-				M_IST(d, REG_SP, dst->vv.regoff * 4);
+		disp = dst->vv.regoff * 4;
+
+		switch (dst->type) {
+		case TYPE_INT:
+		case TYPE_ADR:
+			M_IST(d, REG_SP, disp);
+			break;
+		case TYPE_LNG:
+			M_LST(d, REG_SP, disp);
+			break;
+		case TYPE_FLT:
+			M_FST(d, REG_SP, disp);
+			break;
+		case TYPE_DBL:
+			M_DST(d, REG_SP, disp);
+			break;
+		default:
+			vm_abort("emit_store: unknown type %d", dst->type);
 		}
 	}
 }
@@ -255,18 +266,30 @@ inline void emit_store_high(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 
 *******************************************************************************/
 
-void emit_copy(jitdata *jd, instruction *iptr, varinfo *src, varinfo *dst)
+void emit_copy(jitdata *jd, instruction *iptr)
 {
-	codegendata  *cd;
-	s4            s1, d;
+	codegendata *cd;
+	varinfo     *src;
+	varinfo     *dst;
+	s4           s1, d;
 
 	/* get required compiler data */
 
 	cd = jd->cd;
 
+	/* get source and destination variables */
+
+	src = VAROP(iptr->s1);
+	dst = VAROP(iptr->dst);
+
 	if ((src->vv.regoff != dst->vv.regoff) ||
 		((src->flags ^ dst->flags) & INMEMORY)) {
 
+		if ((src->type == TYPE_RET) || (dst->type == TYPE_RET)) {
+			/* emit nothing, as the value won't be used anyway */
+			return;
+		}
+
 		/* If one of the variables resides in memory, we can eliminate
 		   the register move from/to the temporary register with the
 		   order of getting the destination register and the load. */
@@ -289,13 +312,20 @@ void emit_copy(jitdata *jd, instruction *iptr, varinfo *src, varinfo *dst)
 		}
 
 		if (s1 != d) {
-			if (IS_FLT_DBL_TYPE(src->type)) {
+			switch (src->type) {
+			case TYPE_INT:
+			case TYPE_ADR:
+				M_MOV(s1, d);
+				break;
+			case TYPE_LNG:
+				M_LNGMOVE(s1, d);
+				break;
+			case TYPE_FLT:
+			case TYPE_DBL:
 /* 				M_FMOV(s1, d); */
-			} else {
-				if (IS_2_WORD_TYPE(src->type))
-					M_LNGMOVE(s1, d);
-				else
-                    M_MOV(s1, d);
+				break;
+			default:
+				vm_abort("emit_copy: unknown type %d", src->type);
 			}
 		}
 
@@ -304,82 +334,159 @@ void emit_copy(jitdata *jd, instruction *iptr, varinfo *src, varinfo *dst)
 }
 
 
-/* emit_exception_stubs ********************************************************
+/* emit_branch *****************************************************************
 
-   Generates the code for the exception stubs.
+   Emits the code for conditional and unconditional branchs.
 
 *******************************************************************************/
 
-void emit_exception_stubs(jitdata *jd)
+void emit_branch(codegendata *cd, s4 disp, s4 condition, s4 reg, u4 options)
 {
-	codegendata  *cd;
-	registerdata *rd;
-	exceptionref *eref;
-	s4            targetdisp;
+	s4 branchdisp;
 
-	/* get required compiler data */
+	/* ATTENTION: a displacement overflow cannot happen */
 
-	cd = jd->cd;
-	rd = jd->rd;
+	/* check which branch to generate */
 
-	/* generate exception stubs */
+	if (condition == BRANCH_UNCONDITIONAL) {
 
-	targetdisp = 0;
+		/* calculate the different displacements */
 
-	for (eref = cd->exceptionrefs; eref != NULL; eref = eref->next) {
-		gen_resolvebranch(cd->mcodebase + eref->branchpos,
-						  eref->branchpos,
-						  cd->mcodeptr - cd->mcodebase);
+		branchdisp = disp - BRANCH_UNCONDITIONAL_SIZE;
 
-		MCODECHECK(512);
+		M_JMP_IMM(branchdisp);
+	}
+	else {
+		/* calculate the different displacements */
 
-		/* Check if the exception is an
-		   ArrayIndexOutOfBoundsException.  If so, move index register
-		   into REG_ITMP1. */
+		branchdisp = disp - BRANCH_CONDITIONAL_SIZE;
 
-		if (eref->reg != -1)
-			M_INTMOVE(eref->reg, REG_ITMP1);
-
-		/* calcuate exception address */
-
-		M_MOV_IMM(0, REG_ITMP2_XPC);
-		dseg_adddata(cd);
-		M_AADD_IMM32(eref->branchpos - 6, REG_ITMP2_XPC);
-
-		/* move function to call into REG_ITMP3 */
-
-		M_MOV_IMM(eref->function, REG_ITMP3);
-
-		if (targetdisp == 0) {
-			targetdisp = cd->mcodeptr - cd->mcodebase;
-
-			M_ASUB_IMM(5 * 4, REG_SP);
-
-			/* first store REG_ITMP1 so we can use it */
-
-			M_AST(REG_ITMP1, REG_SP, 4 * 4);                    /* for AIOOBE */
-
-			M_AST_IMM(0, REG_SP, 0 * 4);
-			dseg_adddata(cd);
-			M_MOV(REG_SP, REG_ITMP1);
-			M_AADD_IMM(5 * 4, REG_ITMP1);
-			M_AST(REG_ITMP1, REG_SP, 1 * 4);
-			M_ALD(REG_ITMP1, REG_SP, (5 + cd->stackframesize) * 4);
-			M_AST(REG_ITMP1, REG_SP, 2 * 4);
-			M_AST(REG_ITMP2_XPC, REG_SP, 3 * 4);
-
-			M_CALL(REG_ITMP3);
-
-			M_ALD(REG_ITMP2_XPC, REG_SP, 3 * 4);
-			M_AADD_IMM(5 * 4, REG_SP);
-
-			M_MOV_IMM(asm_handle_exception, REG_ITMP3);
-			M_JMP(REG_ITMP3);
+		switch (condition) {
+		case BRANCH_EQ:
+			M_BEQ(branchdisp);
+			break;
+		case BRANCH_NE:
+			M_BNE(branchdisp);
+			break;
+		case BRANCH_LT:
+			M_BLT(branchdisp);
+			break;
+		case BRANCH_GE:
+			M_BGE(branchdisp);
+			break;
+		case BRANCH_GT:
+			M_BGT(branchdisp);
+			break;
+		case BRANCH_LE:
+			M_BLE(branchdisp);
+			break;
+		case BRANCH_ULT:
+			M_BB(branchdisp);
+			break;
+		case BRANCH_ULE:
+			M_BBE(branchdisp);
+			break;
+		case BRANCH_UGE:
+			M_BAE(branchdisp);
+			break;
+		case BRANCH_UGT:
+			M_BA(branchdisp);
+			break;
+		default:
+			vm_abort("emit_branch: unknown condition %d", condition);
 		}
-		else {
-			M_JMP_IMM((cd->mcodebase + targetdisp) -
-					  (cd->mcodeptr + PATCHER_CALL_SIZE));
+	}
+}
+
+
+/* emit_arithmetic_check *******************************************************
+
+   Emit an ArithmeticException check.
+
+*******************************************************************************/
+
+void emit_arithmetic_check(codegendata *cd, instruction *iptr, s4 reg)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TEST(reg);
+		M_BNE(6);
+		M_ALD_MEM(reg, EXCEPTION_HARDWARE_ARITHMETIC);
+	}
+}
+
+
+/* emit_arrayindexoutofbounds_check ********************************************
+
+   Emit a ArrayIndexOutOfBoundsException check.
+
+*******************************************************************************/
+
+void emit_arrayindexoutofbounds_check(codegendata *cd, instruction *iptr, s4 s1, s4 s2)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+        M_ILD(REG_ITMP3, s1, OFFSET(java_arrayheader, size));
+        M_CMP(REG_ITMP3, s2);
+        M_BB(6);
+		M_ALD_MEM(s2, EXCEPTION_HARDWARE_ARRAYINDEXOUTOFBOUNDS);
+	}
+}
+
+
+/* emit_classcast_check ********************************************************
+
+   Emit a ClassCastException check.
+
+*******************************************************************************/
+
+void emit_classcast_check(codegendata *cd, instruction *iptr, s4 condition, s4 reg, s4 s1)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		switch (condition) {
+		case BRANCH_LE:
+			M_BGT(6);
+			break;
+		case BRANCH_EQ:
+			M_BNE(6);
+			break;
+		case BRANCH_ULE:
+			M_BBE(6);
+			break;
+		default:
+			vm_abort("emit_classcast_check: unknown condition %d", condition);
 		}
+		M_ALD_MEM(s1, EXCEPTION_HARDWARE_CLASSCAST);
+	}
+}
+
+
+/* emit_nullpointer_check ******************************************************
+
+   Emit a NullPointerException check.
+
+*******************************************************************************/
+
+void emit_nullpointer_check(codegendata *cd, instruction *iptr, s4 reg)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TEST(reg);
+		M_BNE(6);
+		M_ALD_MEM(reg, EXCEPTION_HARDWARE_NULLPOINTER);
+	}
+}
+
+
+/* emit_exception_check ********************************************************
+
+   Emit an Exception check.
+
+*******************************************************************************/
+
+void emit_exception_check(codegendata *cd, instruction *iptr)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TEST(REG_RESULT);
+		M_BNE(6);
+		M_ALD_MEM(REG_RESULT, EXCEPTION_HARDWARE_EXCEPTION);
 	}
 }
 
@@ -435,9 +542,9 @@ void emit_patcher_stubs(jitdata *jd)
 		/* move pointer to java_objectheader onto stack */
 
 #if defined(ENABLE_THREADS)
-		(void) dseg_addaddress(cd, NULL);                          /* flcword */
-		(void) dseg_addaddress(cd, lock_get_initial_lock_word());
-		disp = dseg_addaddress(cd, NULL);                          /* vftbl   */
+		(void) dseg_add_unique_address(cd, NULL);                  /* flcword */
+		(void) dseg_add_unique_address(cd, lock_get_initial_lock_word());
+		disp = dseg_add_unique_address(cd, NULL);                  /* vftbl   */
 
 		M_MOV_IMM(0, REG_ITMP3);
 		dseg_adddata(cd);
@@ -474,13 +581,15 @@ void emit_patcher_stubs(jitdata *jd)
 
 *******************************************************************************/
 
+#if defined(ENABLE_REPLACEMENT)
 void emit_replacement_stubs(jitdata *jd)
 {
 	codegendata *cd;
 	codeinfo    *code;
 	rplpoint    *rplp;
-	s4           disp;
 	s4           i;
+	s4           branchmpc;
+	s4           outcode;
 
 	/* get required compiler data */
 
@@ -489,20 +598,23 @@ void emit_replacement_stubs(jitdata *jd)
 
 	rplp = code->rplpoints;
 
+	/* store beginning of replacement stubs */
+
+	code->replacementstubs = (u1*) (cd->mcodeptr - cd->mcodebase);
+
 	for (i = 0; i < code->rplpointcount; ++i, ++rplp) {
+		/* do not generate stubs for non-trappable points */
+
+		if (rplp->flags & RPLPOINT_FLAG_NOTRAP)
+			continue;
+
 		/* check code segment size */
 
 		MCODECHECK(512);
 
 		/* note start of stub code */
 
-		rplp->outcode = (u1 *) (ptrint) (cd->mcodeptr - cd->mcodebase);
-
-		/* make machine code for patching */
-
-		disp = (ptrint) (rplp->outcode - rplp->pc) - 5;
-
-		rplp->mcode = 0xe9 | ((u8) disp << 8);
+		outcode = (s4) (cd->mcodeptr - cd->mcodebase);
 
 		/* push address of `rplpoint` struct */
 			
@@ -512,8 +624,20 @@ void emit_replacement_stubs(jitdata *jd)
 
 		M_PUSH_IMM(asm_replacement_out);
 		M_RET;
+
+		/* add jump reference for COUNTDOWN points */
+
+		if (rplp->flags & RPLPOINT_FLAG_COUNTDOWN) {
+
+			branchmpc = (s4)rplp->pc + (7 + 6);
+
+			md_codegen_patch_branch(cd, branchmpc, (s4) outcode);
+		}
+
+		assert(((cd->mcodeptr - cd->mcodebase) - outcode) == REPLACEMENT_STUB_SIZE);
 	}
 }
+#endif /* defined(ENABLE_REPLACEMENT) */
 	
 
 /* emit_verbosecall_enter ******************************************************
@@ -532,6 +656,9 @@ void emit_verbosecall_enter(jitdata *jd)
 	s4            disp;
 	s4            i, t;
 
+	if (!JITDATA_HAS_FLAG_VERBOSECALL(jd))
+		return;
+
 	/* get required compiler data */
 
 	m  = jd->m;
@@ -592,7 +719,7 @@ void emit_verbosecall_enter(jitdata *jd)
 	
 	M_AST_IMM(m, REG_SP, TRACE_ARGS_NUM * 8);
 
-	M_MOV_IMM(builtin_trace_args, REG_ITMP1);
+	M_MOV_IMM(builtin_verbosecall_enter, REG_ITMP1);
 	M_CALL(REG_ITMP1);
 
 	/* restore temporary registers for leaf methods */
@@ -613,6 +740,8 @@ void emit_verbosecall_enter(jitdata *jd)
 
    Generates the code for the call trace.
 
+   void builtin_verbosecall_exit(s8 l, double d, float f, methodinfo *m);
+
 *******************************************************************************/
 
 #if !defined(NDEBUG)
@@ -622,6 +751,9 @@ void emit_verbosecall_exit(jitdata *jd)
 	codegendata  *cd;
 	registerdata *rd;
 
+	if (!JITDATA_HAS_FLAG_VERBOSECALL(jd))
+		return;
+
 	/* get required compiler data */
 
 	m  = jd->m;
@@ -632,21 +764,21 @@ void emit_verbosecall_exit(jitdata *jd)
 
 	M_NOP;
 
-	M_ASUB_IMM(4 + 8 + 8 + 4 + 8, REG_SP);  /* +8: keep stack 16-byte aligned */
+	M_ASUB_IMM(8 + 8 + 4 + 4 + 8, REG_SP);  /* +8: keep stack 16-byte aligned */
 
-	M_AST_IMM(m, REG_SP, 0 * 4);
+	M_LST(REG_RESULT_PACKED, REG_SP, 0 * 8);
 
-	M_LST(REG_RESULT_PACKED, REG_SP, 1 * 4);
+	M_DSTNP(REG_NULL, REG_SP, 1 * 8);
+	M_FSTNP(REG_NULL, REG_SP, 2 * 8);
 
-	M_DSTNP(REG_NULL, REG_SP, 1 * 4 + 1 * 8);
-	M_FSTNP(REG_NULL, REG_SP, 1 * 4 + 2 * 8);
+	M_AST_IMM(m, REG_SP, 2 * 8 + 1 * 4);
 
-	M_MOV_IMM(builtin_displaymethodstop, REG_ITMP1);
+	M_MOV_IMM(builtin_verbosecall_exit, REG_ITMP1);
 	M_CALL(REG_ITMP1);
 
-	M_LLD(REG_RESULT_PACKED, REG_SP, 1 * 4);
+	M_LLD(REG_RESULT_PACKED, REG_SP, 0 * 4);
 
-	M_AADD_IMM(4 + 8 + 8 + 4 + 8, REG_SP);
+	M_AADD_IMM(8 + 8 + 4 + 4 + 8, REG_SP);
 
 	/* mark trace code */
 
@@ -997,6 +1129,20 @@ void emit_alu_imm_membase(codegendata *cd, s4 opc, s4 imm, s4 basereg, s4 disp)
 }
 
 
+void emit_alu_imm_memabs(codegendata *cd, s4 opc, s4 imm, s4 disp)
+{
+	if (IS_IMM8(imm)) { 
+		*(cd->mcodeptr++) = 0x83;
+		emit_mem(opc, disp);
+		emit_imm8((imm));
+	} else { 
+		*(cd->mcodeptr++) = 0x81;
+		emit_mem(opc, disp);
+		emit_imm32((imm));
+	}
+}
+
+
 void emit_test_reg_reg(codegendata *cd, s4 reg, s4 dreg)
 {
 	*(cd->mcodeptr++) = 0x85;
