@@ -10,6 +10,16 @@ atomic_add (volatile int *mem, int val)
 						: "ir" (val), "m" (*mem));
 }
 
+/* compare_and_swap ********************************************************
+
+  Atomically do the following: Check if the location still contains
+  `oldval`. If so, replace it by `newval` and return `oldval`.
+
+  RETURN VALUE:
+      the old value at *p
+
+***************************************************************************/
+
 static inline long
 __attribute__ ((unused))
 compare_and_swap (volatile long *p, long oldval, long newval)
