@@ -1,6 +1,6 @@
-/* jit/inline.h - code inliner
+/* src/vm/jit/inline/inline.h - code inliner
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Edwin Steiner
-
-   Changes:
-
-   $Id: inline.h 4478 2006-02-07 17:22:13Z edwin $
+   $Id: inline.h 7835 2007-04-26 13:43:08Z twisti $
 
 */
 
@@ -36,8 +30,17 @@
 #ifndef _INLINE_H
 #define _INLINE_H
 
+#include "config.h"
+
 #include "vm/global.h"
 
+#include "vm/jit/jit.h"
+
+
+/* function prototypes ********************************************************/
+
+bool inline_inline(jitdata *jd);
+
 #endif /* _INLINE_H */
 
 /*
