@@ -1,6 +1,6 @@
 /* src/vm/jit/mips/codegen.h - code generation macros and definitions for MIPS
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,11 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Andreas Krall
-
-   $Id: codegen.h 5632 2006-10-02 13:43:15Z edwin $
+   $Id: codegen.h 7691 2007-04-12 12:45:10Z twisti $
 
 */
 
@@ -40,37 +36,8 @@
 #include "vm/jit/jit.h"
 
 
-/* some defines ***************************************************************/
-
-#define PATCHER_CALL_SIZE    2 * 4      /* size in bytes of a patcher call    */
-
-
 /* additional functions and macros to generate code ***************************/
 
-#define gen_nullptr_check(objreg) \
-    if (checknull) { \
-        M_BEQZ(objreg, 0); \
-        codegen_add_nullpointerexception_ref(cd); \
-        M_NOP; \
-    }
-
-#define gen_bound_check \
-    if (checkbounds) { \
-        M_ILD(REG_ITMP3, s1, OFFSET(java_arrayheader, size)); \
-        M_CMPULT(s2, REG_ITMP3, REG_ITMP3); \
-        M_BEQZ(REG_ITMP3, 0); \
-        codegen_add_arrayindexoutofboundsexception_ref(cd, s2); \
-        M_NOP; \
-    }
-
-#define gen_div_check(r) \
-    do { \
-        M_BEQZ((r), 0); \
-        codegen_add_arithmeticexception_ref(cd); \
-        M_NOP; \
-    } while (0)
-
-
 /* MCODECHECK(icnt) */
 
 #define MCODECHECK(icnt) \
@@ -85,8 +52,31 @@
         M_NOP; \
     }
 
+#define M_INTMOVE(a,b) \
+    do { \
+        if ((a) != (b)) \
+            M_MOV(a, b); \
+    } while (0)
 
-#define M_INTMOVE(a,b) if (a != b) { M_MOV(a, b); }
+#if SIZEOF_VOID_P == 8
+
+#define M_LNGMOVE(a,b)    M_INTMOVE(a,b)
+
+#else /* SIZEOF_VOID_P == 8 */
+
+#define M_LNGMOVE(a,b) \
+    do { \
+        if (GET_LOW_REG(b) == GET_HIGH_REG(a)) { \
+            assert(GET_HIGH_REG(b) == GET_LOW_REG(a)); \
+            M_INTMOVE(GET_HIGH_REG(a), GET_HIGH_REG(b)); \
+            M_INTMOVE(GET_LOW_REG(a), GET_LOW_REG(b)); \
+        } else { \
+            M_INTMOVE(GET_LOW_REG(a), GET_LOW_REG(b)); \
+            M_INTMOVE(GET_HIGH_REG(a), GET_HIGH_REG(b)); \
+        } \
+    } while (0)
+
+#endif /* SIZEOF_VOID_P == 8 */
 
 #define M_FLTMOVE(a,b) \
     do { \
@@ -104,6 +94,45 @@
 #define LCONST(r,c)                     emit_lconst(cd, (r), (c))
 
 
+/* branch defines *************************************************************/
+
+#define BRANCH_NOPS \
+    do { \
+        if (CODEGENDATA_HAS_FLAG_LONGBRANCHES(cd)) { \
+            M_NOP; \
+            M_NOP; \
+            M_NOP; \
+            M_NOP; \
+            M_NOP; \
+            M_NOP; \
+        } \
+        else { \
+            M_NOP; \
+            M_NOP; \
+        } \
+    } while (0)
+
+
+/* patcher defines ************************************************************/
+
+#define PATCHER_CALL_INSTRUCTIONS    5     /* number of instructions          */
+#define PATCHER_CALL_SIZE            5 * 4 /* size in bytes of a patcher call */
+
+#define PATCHER_NOPS \
+    do { \
+        M_NOP; \
+        M_NOP; \
+        M_NOP; \
+        M_NOP; \
+        M_NOP; \
+    } while (0)
+
+
+/* stub defines ***************************************************************/
+
+#define COMPILERSTUB_CODESIZE    4 * 4
+
+
 /* macros to create code ******************************************************/
 
 /* code generation macros operands:
@@ -123,6 +152,11 @@
         cd->mcodeptr += 4; \
     } while (0)
 
+#define M_ITYPE_GET_RS(x)               (((x) >> 21) & 0x1f  )
+#define M_ITYPE_GET_RT(x)               (((x) >> 16) & 0x1f  )
+#define M_ITYPE_GET_IMM(x)              ( (x)        & 0xffff)
+
+
 #define M_JTYPE(op,imm) \
     do { \
         *((u4 *) cd->mcodeptr) = (((op) << 26) | ((off) & 0x3ffffff)); \
@@ -155,9 +189,9 @@
         if (hi == 0) { \
             M_AADD_IMM(b,lo,a); \
         } else { \
-            M_LUI(REG_ITMP3,hi); \
-            M_AADD_IMM(REG_ITMP3,lo,REG_ITMP3); \
-            M_AADD(REG_ITMP3,b,a); \
+            M_LUI(REG_ITMP3, hi); \
+            M_AADD_IMM(REG_ITMP3, lo, REG_ITMP3); \
+            M_AADD(b, REG_ITMP3, a); \
         } \
     } while (0)
 
@@ -167,7 +201,6 @@
 #define M_SLDU(a,b,disp)        M_ITYPE(0x25,b,a,disp)          /* 16 load    */
 
 #define M_ILD_INTERN(a,b,disp)  M_ITYPE(0x23,b,a,disp)          /* 32 load    */
-#define M_LLD_INTERN(a,b,disp)  M_ITYPE(0x37,b,a,disp)          /* 64 load    */
 
 #define M_ILD(a,b,disp) \
     do { \
@@ -182,16 +215,90 @@
         } \
     } while (0)
 
+#if SIZEOF_VOID_P == 8
+
+#define M_LLD_INTERN(a,b,disp)  M_ITYPE(0x37,b,a,disp)          /* 64 load    */
+
+#else /* SIZEOF_VOID_P == 8 */
+
+# if WORDS_BIGENDIAN == 1
+
+/* ATTENTION: b may be equal to GET_LOW_REG(a) (displacement overflow case ) */
+
+#define M_LLD_INTERN(a,b,disp) \
+    do { \
+        M_ILD_INTERN(GET_HIGH_REG(a), b, disp); \
+        M_ILD_INTERN(GET_LOW_REG(a), b, disp + 4); \
+    } while (0)
+
+/* this macro is specially for ICMD_GETFIELD since the base (source)
+   register may be one of the destination registers */
+
+#define M_LLD_GETFIELD(a,b,disp) \
+    do { \
+        s4 lo = (short) (disp); \
+        s4 hi = (short) (((disp) - lo) >> 16); \
+        if (hi == 0) { \
+            if (GET_HIGH_REG(a) == (b)) { \
+                M_ILD_INTERN(GET_LOW_REG(a), b, disp + 4); \
+                M_ILD_INTERN(GET_HIGH_REG(a), b, disp); \
+            } else { \
+                M_ILD_INTERN(GET_HIGH_REG(a), b, disp); \
+                M_ILD_INTERN(GET_LOW_REG(a), b, disp + 4); \
+            } \
+        } else { \
+            M_LUI(GET_LOW_REG(a), hi); \
+            M_AADD(b, GET_LOW_REG(a), GET_LOW_REG(a)); \
+            M_LLD_INTERN(a, GET_LOW_REG(a), lo); \
+        } \
+    } while (0)
+
+# else /* if WORDS_BIGENDIAN == 1 */
+
+/* ATTENTION: b may be equal to GET_LOW_REG(a) (displacement overflow case ) */
+
+#define M_LLD_INTERN(a,b,disp) \
+    do { \
+        M_ILD_INTERN(GET_HIGH_REG(a), b, disp + 4); \
+        M_ILD_INTERN(GET_LOW_REG(a), b, disp); \
+    } while (0)
+
+/* this macro is specially for ICMD_GETFIELD since the base (source)
+   register may be one of the destination registers */
+
+#define M_LLD_GETFIELD(a,b,disp) \
+    do { \
+        s4 lo = (short) (disp); \
+        s4 hi = (short) (((disp) - lo) >> 16); \
+        if (hi == 0) { \
+            if (GET_HIGH_REG(a) == (b)) { \
+                M_ILD_INTERN(GET_LOW_REG(a), b, disp); \
+                M_ILD_INTERN(GET_HIGH_REG(a), b, disp + 4); \
+            } else { \
+                M_ILD_INTERN(GET_HIGH_REG(a), b, disp + 4); \
+                M_ILD_INTERN(GET_LOW_REG(a), b, disp); \
+            } \
+        } else { \
+            M_LUI(GET_LOW_REG(a), hi); \
+            M_AADD(b, GET_LOW_REG(a), GET_LOW_REG(a)); \
+            M_LLD_INTERN(a, GET_LOW_REG(a), lo); \
+        } \
+    } while (0)
+
+# endif /* if WORDS_BIGENDIAN == 1 */
+
+#endif /* SIZEOF_VOID_P == 8 */
+
 #define M_LLD(a,b,disp) \
     do { \
         s4 lo = (short) (disp); \
         s4 hi = (short) (((disp) - lo) >> 16); \
         if (hi == 0) { \
-            M_LLD_INTERN(a,b,lo); \
+            M_LLD_INTERN(a, b, lo); \
         } else { \
-            M_LUI(a,hi); \
-            M_AADD(b,a,a); \
-            M_LLD_INTERN(a,a,lo); \
+            M_LUI(GET_LOW_REG(a), hi); \
+            M_AADD(b, GET_LOW_REG(a), GET_LOW_REG(a)); \
+            M_LLD_INTERN(a, GET_LOW_REG(a), lo); \
         } \
     } while (0)
 
@@ -199,7 +306,6 @@
 #define M_SST(a,b,disp)         M_ITYPE(0x29,b,a,disp)          /* 16 store   */
 
 #define M_IST_INTERN(a,b,disp)  M_ITYPE(0x2b,b,a,disp)          /* 32 store   */
-#define M_LST_INTERN(a,b,disp)  M_ITYPE(0x3f,b,a,disp)          /* 64 store   */
 
 #define M_IST(a,b,disp) \
     do { \
@@ -210,10 +316,36 @@
         } else { \
             M_LUI(REG_ITMP3, hi); \
             M_AADD(b, REG_ITMP3, REG_ITMP3); \
-            M_IST_INTERN(a,REG_ITMP3,lo); \
+            M_IST_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
+#if SIZEOF_VOID_P == 8
+
+#define M_LST_INTERN(a,b,disp)  M_ITYPE(0x3f,b,a,disp)          /* 64 store   */
+
+#else /* SIZEOF_VOID_P == 8 */
+
+#if WORDS_BIGENDIAN == 1
+
+#define M_LST_INTERN(a,b,disp) \
+    do { \
+        M_IST_INTERN(GET_HIGH_REG(a), b, disp); \
+        M_IST_INTERN(GET_LOW_REG(a), b, disp + 4); \
+    } while (0)
+
+#else /* if WORDS_BIGENDIAN == 1 */
+
+#define M_LST_INTERN(a,b,disp) \
+    do { \
+        M_IST_INTERN(GET_HIGH_REG(a), b, disp + 4); \
+        M_IST_INTERN(GET_LOW_REG(a), b, disp); \
+    } while (0)
+
+#endif /* if WORDS_BIGENDIAN == 1 */
+
+#endif /* SIZEOF_VOID_P == 8 */
+
 #define M_LST(a,b,disp) \
     do { \
         s4 lo = (short) (disp); \
@@ -223,7 +355,7 @@
         } else { \
             M_LUI(REG_ITMP3, hi); \
             M_AADD(b, REG_ITMP3, REG_ITMP3); \
-            M_LST_INTERN(a,REG_ITMP3,lo); \
+            M_LST_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
@@ -237,9 +369,9 @@
         if (hi == 0) { \
             M_FLD_INTERN(a,b,lo); \
         } else { \
-            M_LUI(REG_ITMP3,hi); \
-            M_AADD(b,REG_ITMP3,REG_ITMP3); \
-            M_FLD_INTERN(a,REG_ITMP3,lo); \
+            M_LUI(REG_ITMP3, hi); \
+            M_AADD(b, REG_ITMP3, REG_ITMP3); \
+            M_FLD_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
@@ -250,9 +382,9 @@
         if (hi == 0) { \
             M_DLD_INTERN(a,b,lo); \
         } else { \
-            M_LUI(REG_ITMP3,hi); \
-            M_AADD(b,REG_ITMP3,REG_ITMP3); \
-            M_DLD_INTERN(a,REG_ITMP3,lo); \
+            M_LUI(REG_ITMP3, hi); \
+            M_AADD(b, REG_ITMP3, REG_ITMP3); \
+            M_DLD_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
@@ -268,7 +400,7 @@
         } else { \
             M_LUI(REG_ITMP3, hi); \
             M_AADD(b, REG_ITMP3, REG_ITMP3); \
-            M_FST_INTERN(a,REG_ITMP3,lo); \
+            M_FST_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
@@ -281,7 +413,7 @@
         } else { \
             M_LUI(REG_ITMP3, hi); \
             M_AADD(b, REG_ITMP3, REG_ITMP3); \
-            M_DST_INTERN(a,REG_ITMP3,lo); \
+            M_DST_INTERN(a, REG_ITMP3, lo); \
         } \
     } while (0)
 
@@ -294,6 +426,7 @@
 #define M_BGEZ(a,disp)          M_ITYPE(0x01,a,1,disp)          /* br a >= 0  */
 #define M_BGTZ(a,disp)          M_ITYPE(0x07,a,0,disp)          /* br a >  0  */
 
+#if 0
 #define M_BEQL(a,b,disp)        M_ITYPE(0x14,a,b,disp)          /* br a == b  */
 #define M_BNEL(a,b,disp)        M_ITYPE(0x15,a,b,disp)          /* br a != b  */
 #define M_BEQZL(a,disp)         M_ITYPE(0x14,a,0,disp)          /* br a == 0  */
@@ -302,6 +435,7 @@
 #define M_BNEZL(a,disp)         M_ITYPE(0x15,a,0,disp)          /* br a != 0  */
 #define M_BGEZL(a,disp)         M_ITYPE(0x01,a,3,disp)          /* br a >= 0  */
 #define M_BGTZL(a,disp)         M_ITYPE(0x17,a,0,disp)          /* br a >  0  */
+#endif
 
 #define M_BR(disp)              M_ITYPE(0x04,0,0,disp)          /* branch     */
 #define M_BRS(disp)             M_ITYPE(0x01,0,17,disp)         /* branch sbr */
@@ -341,6 +475,7 @@
 #define M_ISUB(a,b,c)           M_RTYPE(0,a,b,c,0,0x23)         /* 32 sub     */
 #define M_LSUB(a,b,c)           M_RTYPE(0,a,b,c,0,0x2f)         /* 64 sub     */
 #define M_IMUL(a,b)             M_ITYPE(0,a,b,0x18)             /* 32 mul     */
+#define M_IMULU(a,b)            M_ITYPE(0,a,b,0x19)             /* 32 mul     */
 #define M_LMUL(a,b)             M_ITYPE(0,a,b,0x1c)             /* 64 mul     */
 #define M_IDIV(a,b)             M_ITYPE(0,a,b,0x1a)             /* 32 div     */
 #define M_LDIV(a,b)             M_ITYPE(0,a,b,0x1e)             /* 64 div     */
@@ -350,8 +485,8 @@
 
 #define M_IADD_IMM(a,b,c)       M_ITYPE(0x09,a,c,b)             /* 32 add     */
 #define M_LADD_IMM(a,b,c)       M_ITYPE(0x19,a,c,b)             /* 64 add     */
-#define M_ISUB_IMM(a,b,c)       M_ITYPE(0x09,a,c,-(b))          /* 32 sub     */
-#define M_LSUB_IMM(a,b,c)       M_ITYPE(0x19,a,c,-(b))          /* 64 sub     */
+#define M_ISUB_IMM(a,b,c)       M_IADD_IMM(a,-(b),c)            /* 32 sub     */
+#define M_LSUB_IMM(a,b,c)       M_LADD_IMM(a,-(b),c)            /* 64 sub     */
 
 #define M_LUI(a,imm)            M_ITYPE(0x0f,0,a,imm)           /* a = imm<<16*/
 
@@ -372,8 +507,6 @@
 #define M_OR_IMM( a,b,c)        M_ITYPE(0x0d,a,c,b)             /* c = a |  b */
 #define M_XOR_IMM(a,b,c)        M_ITYPE(0x0e,a,c,b)             /* c = a ^  b */
 
-#define M_CZEXT(a,c)            M_AND_IMM(a,0xffff,c)           /* c = zext(a)*/
-
 #define M_ISLL(a,b,c)           M_RTYPE(0,b,a,c,0,0x04)         /* c = a << b */
 #define M_ISRL(a,b,c)           M_RTYPE(0,b,a,c,0,0x06)         /* c = a >>>b */
 #define M_ISRA(a,b,c)           M_RTYPE(0,b,a,c,0,0x07)         /* c = a >> b */
@@ -446,6 +579,9 @@
 #define M_MOVID(i,d)            M_FP3(0,4,d,i,0)                /* d = i      */
 #define M_MOVLD(l,d)            M_FP3(0,5,d,l,0)                /* d = l      */
 
+#define M_MFC1(l,f)             M_FP3(0,0,f,l,0)
+#define M_MTC1(l,f)             M_FP3(0,4,f,l,0)
+
 #define M_DMFC1(l,f)            M_FP3(0,1,f,l,0)
 #define M_DMTC1(l,f)            M_FP3(0,5,f,l,0)
 
@@ -507,42 +643,6 @@
 
 #endif /* SIZEOF_VOID_P == 8 */
 
-
-/* function gen_resolvebranch **************************************************
-
-	backpatches a branch instruction; MIPS branch instructions are very
-	regular, so it is only necessary to overwrite some fixed bits in the
-	instruction.
-
-	parameters: ip ... pointer to instruction after branch (void*)
-	            so ... offset of instruction after branch  (s4)
-	            to ... offset of branch target             (s4)
-
-*******************************************************************************/
-
-#define gen_resolvebranch(ip,so,to) \
-    do { \
-        s4 offset; \
-        \
-        offset = ((s4) (to) - (so)) >> 2; \
-        \
-        /* On the MIPS we can only branch signed 16-bit instruction words */ \
-        /* (signed 18-bit = 32KB = +/- 16KB). Check this!                 */ \
-        \
-        if ((offset < (s4) 0xffff8000) || (offset > (s4) 0x00007fff)) { \
-            throw_cacao_exception_exit(string_java_lang_InternalError, \
-                                       "Jump offset is out of range: %d > +/-%d", \
-                                       offset, 0x00007fff); \
-        } \
-        \
-        ((s4 *) (ip))[-1] |= (offset & 0x0000ffff); \
-    } while (0)
-
-
-/* function prototypes ********************************************************/
-
-void docacheflush(u1 *p, long bytelen);
-
 #endif /* _CODEGEN_H */
 
 
