@@ -1,6 +1,6 @@
 /* src/vm/jit/mips/patcher.c - MIPS code patching functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: patcher.c 5326 2006-09-05 17:21:21Z edwin $
+   $Id: patcher.c 7596 2007-03-28 21:05:53Z twisti $
 
 */
 
@@ -39,19 +33,26 @@
 
 #include "vm/types.h"
 
+#include "vm/jit/mips/codegen.h"
+
 #include "mm/memory.h"
+
 #include "native/native.h"
+
 #include "vm/builtin.h"
-#include "vm/class.h"
 #include "vm/exceptions.h"
-#include "vm/field.h"
 #include "vm/initialize.h"
-#include "vm/options.h"
-#include "vm/resolve.h"
-#include "vm/references.h"
+
 #include "vm/jit/asmpart.h"
+#include "vm/jit/md.h"
 #include "vm/jit/patcher.h"
 
+#include "vmcore/class.h"
+#include "vmcore/field.h"
+#include "vmcore/options.h"
+#include "vm/resolve.h"
+#include "vmcore/references.h"
+
 
 /* patcher_wrapper *************************************************************
 
@@ -69,6 +70,7 @@ java_objectheader *patcher_wrapper(u1 *sp, u1 *pv, u1 *ra)
 	stackframeinfo     sfi;
 	u1                *xpc;
 	java_objectheader *o;
+	u4                 mcode[5];
 	functionptr        f;
 	bool               result;
 	java_objectheader *e;
@@ -81,8 +83,8 @@ java_objectheader *patcher_wrapper(u1 *sp, u1 *pv, u1 *ra)
 
 	/* get stuff from the stack */
 
-	xpc = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	o   = (java_objectheader *) *((ptrint *) (sp + 4 * 8));
+	xpc = (u1 *)                *((ptrint *) (sp + 7 * 8));
+	o   = (java_objectheader *) *((ptrint *) (sp + 6 * 8));
 	f   = (functionptr)         *((ptrint *) (sp + 0 * 8));
 
 	/* store PV into the patcher function position */
@@ -99,7 +101,7 @@ java_objectheader *patcher_wrapper(u1 *sp, u1 *pv, u1 *ra)
 
 	/* create the stackframeinfo */
 
-	stacktrace_create_extern_stackframeinfo(&sfi, pv, sp + 6 * 8, ra, xpc);
+	stacktrace_create_extern_stackframeinfo(&sfi, pv, sp + 8 * 8, ra, xpc);
 
 	/* call the proper patcher function */
 
@@ -119,6 +121,26 @@ java_objectheader *patcher_wrapper(u1 *sp, u1 *pv, u1 *ra)
 		return e;
 	}
 
+	/* patch back original code */
+
+	mcode[0] = *((u4 *) (sp + 3 * 8 + 0));
+	mcode[1] = *((u4 *) (sp + 3 * 8 + 4));
+
+	mcode[2] = *((u4 *) (sp + 4 * 8 + 0));
+	mcode[3] = *((u4 *) (sp + 4 * 8 + 4));
+	mcode[4] = *((u4 *) (sp + 5 * 8 + 0));
+
+	*((u4 *) (xpc + 0 * 4)) = mcode[0];
+	*((u4 *) (xpc + 1 * 4)) = mcode[1];
+
+	*((u4 *) (xpc + 2 * 4)) = mcode[2];
+	*((u4 *) (xpc + 3 * 4)) = mcode[3];
+	*((u4 *) (xpc + 4 * 4)) = mcode[4];
+
+	/* synchronize instruction cache */
+
+	md_icacheflush(xpc, PATCHER_CALL_SIZE);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return NULL;
@@ -137,8 +159,6 @@ java_objectheader *patcher_wrapper(u1 *sp, u1 *pv, u1 *ra)
 
 bool patcher_get_putstatic(u1 *sp)
 {
-	u1               *ra;
-	u4                mcode[2];
 	unresolved_field *uf;
 	s4                disp;
 	u1               *pv;
@@ -146,12 +166,9 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)               *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                      *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                      *((u4 *)     (sp + 3 * 8 + 4));
-	uf       = (unresolved_field *) *((ptrint *) (sp + 2 * 8));
-	disp     =                      *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)               *((ptrint *) (sp + 0 * 8));
+	uf   = (unresolved_field *) *((ptrint *) (sp + 2 * 8));
+	disp =                      *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)               *((ptrint *) (sp + 0 * 8));
 
 	/* get the fieldinfo */
 
@@ -164,15 +181,6 @@ bool patcher_get_putstatic(u1 *sp)
 		if (!initialize_class(fi->class))
 			return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch the field value's address */
 
 	*((ptrint *) (pv + disp)) = (ptrint) &(fi->value);
@@ -197,59 +205,60 @@ bool patcher_get_putstatic(u1 *sp)
 bool patcher_get_putfield(u1 *sp)
 {
 	u1               *ra;
-	u4                mcode[2];
 	unresolved_field *uf;
 	fieldinfo        *fi;
 
-	ra       = (u1 *)               *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                      *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                      *((u4 *)     (sp + 3 * 8 + 4));
-	uf       = (unresolved_field *) *((ptrint *) (sp + 2 * 8));
+	/* get stuff from the stack */
+
+	ra = (u1 *)               *((ptrint *) (sp + 7 * 8));
+	uf = (unresolved_field *) *((ptrint *) (sp + 2 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(fi = resolve_field_eager(uf)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 2 * 4;
-
-	/* patch the field's offset */
+	if (opt_shownops) {
+		ra = ra + PATCHER_CALL_SIZE;
 
 #if SIZEOF_VOID_P == 4
-	if (fi->type == TYPE_LNG) {
+		if (IS_LNG_TYPE(fi->type)) {
 # if WORDS_BIGENDIAN == 1
-		/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
-		*((u4 *) (ra + 0 * 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
-		*((u4 *) (ra + 1 * 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+			/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
+			*((u4 *) (ra + 0 * 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+			*((u4 *) (ra + 1 * 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
 # else
-		/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
-		*((u4 *) (ra + 0 * 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
-		*((u4 *) (ra + 1 * 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+			/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
+			*((u4 *) (ra + 0 * 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+			*((u4 *) (ra + 1 * 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
 # endif
-	} else
-#endif
-		*((u4 *) ra) |= (s2) (fi->offset & 0x0000ffff);
-
-	/* synchronize instruction cache */
-
-	if (opt_showdisassemble) {
-#if SIZEOF_VOID_P == 4
-		if (fi->type == TYPE_LNG)
-			md_icacheflush(ra - 2 * 4, 4 * 4);
+		}
 		else
 #endif
-			md_icacheflush(ra - 2 * 4, 3 * 4);
+			*((u4 *) (ra + 0 * 4)) |= (s2) (fi->offset & 0x0000ffff);
+
+		/* synchronize instruction cache */
+
+		md_icacheflush(ra, 2 * 4);
 	}
 	else {
-		md_icacheflush(ra, 2 * 4);
+#if SIZEOF_VOID_P == 4
+		if (IS_LNG_TYPE(fi->type)) {
+# if WORDS_BIGENDIAN == 1
+			/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
+			*((u4 *) (sp + 3 * 8 + 0)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+			*((u4 *) (sp + 3 * 8 + 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+# else
+			/* ATTENTION: order of these instructions depend on M_LLD_INTERN */
+			*((u4 *) (sp + 3 * 8 + 0)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+			*((u4 *) (sp + 3 * 8 + 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+# endif
+		}
+		else
+#endif
+			*((u4 *) (sp + 3 * 8 + 0)) |= (s2) (fi->offset & 0x0000ffff);
 	}
 
 	return true;
@@ -267,8 +276,6 @@ bool patcher_get_putfield(u1 *sp)
 
 bool patcher_aconst(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	s4                 disp;
 	u1                *pv;
@@ -276,27 +283,15 @@ bool patcher_aconst(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	cr   = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the classinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
@@ -324,8 +319,6 @@ bool patcher_aconst(u1 *sp)
 
 bool patcher_builtin_multianewarray(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	s4                 disp;
 	u1                *pv;
@@ -333,27 +326,15 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	cr   = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the classinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
@@ -380,8 +361,6 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 bool patcher_builtin_arraycheckcast(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	s4                 disp;
 	u1                *pv;
@@ -389,27 +368,15 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	cr   = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the classinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
@@ -435,8 +402,6 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 bool patcher_invokestatic_special(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	unresolved_method *um;
 	s4                 disp;
 	u1                *pv;
@@ -444,27 +409,15 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	um       = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	um   = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(m = resolve_method_eager(um)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch stubroutine */
 
 	*((ptrint *) (pv + disp)) = (ptrint) m->stubroutine;
@@ -492,43 +445,41 @@ bool patcher_invokestatic_special(u1 *sp)
 bool patcher_invokevirtual(u1 *sp)
 {
 	u1                *ra;
-	u4                 mcode[2];
 	unresolved_method *um;
 	methodinfo        *m;
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	um       = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
+	ra = (u1 *)                *((ptrint *) (sp + 7 * 8));
+	um = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(m = resolve_method_eager(um)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 2 * 4;
+	if (opt_shownops) {
+		ra = ra + PATCHER_CALL_SIZE;
 
-	/* patch vftbl index */
+		/* patch vftbl index */
 
-	*((s4 *) (ra + 1 * 4)) |= (s4) ((OFFSET(vftbl_t, table[0]) +
-									 sizeof(methodptr) * m->vftblindex) & 0x0000ffff);
+		*((s4 *) (ra + 1 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, table[0]) +
+				   sizeof(methodptr) * m->vftblindex) & 0x0000ffff);
 
-	/* synchronize instruction cache */
+		/* synchronize instruction cache */
 
-	if (opt_showdisassemble)
-		md_icacheflush(ra - 2 * 4, 4 * 4);
-	else
-		md_icacheflush(ra, 2 * 4);
+		md_icacheflush(ra + 1 * 4, 1 * 4);
+	}
+	else {
+		/* patch vftbl index */
+
+		*((s4 *) (sp + 3 * 8 + 4)) |=
+			(s4) ((OFFSET(vftbl_t, table[0]) +
+				   sizeof(methodptr) * m->vftblindex) & 0x0000ffff);
+	}
 
 	return true;
 }
@@ -550,48 +501,51 @@ bool patcher_invokevirtual(u1 *sp)
 bool patcher_invokeinterface(u1 *sp)
 {
 	u1                *ra;
-	u4                 mcode[2];
 	unresolved_method *um;
 	methodinfo        *m;
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	um       = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
+	ra = (u1 *)                *((ptrint *) (sp + 7 * 8));
+	um = (unresolved_method *) *((ptrint *) (sp + 2 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(m = resolve_method_eager(um)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 2 * 4;
+	if (opt_shownops) {
+		ra = ra + PATCHER_CALL_SIZE;
 
-	/* patch interfacetable index */
+		/* patch interfacetable index */
 
-	*((s4 *) (ra + 1 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) -
-									 sizeof(methodptr*) * m->class->index) & 0x0000ffff);
+		*((s4 *) (ra + 1 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   sizeof(methodptr*) * m->class->index) & 0x0000ffff);
 
-	/* patch method offset */
+		/* patch method offset */
 
-	*((s4 *) (ra + 2 * 4)) |=
-		(s4) ((sizeof(methodptr) * (m - m->class->methods)) & 0x0000ffff);
+		*((s4 *) (ra + 2 * 4)) |=
+			(s4) ((sizeof(methodptr) * (m - m->class->methods)) & 0x0000ffff);
 
-	/* synchronize instruction cache */
+		/* synchronize instruction cache */
 
-	if (opt_showdisassemble)
-		md_icacheflush(ra - 2 * 4, 5 * 4);
-	else
-		md_icacheflush(ra, 3 * 4);
+		md_icacheflush(ra + 1 * 4, 2 * 4);
+	}
+	else {
+		/* patch interfacetable index */
+
+		*((s4 *) (sp + 3 * 8 + 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   sizeof(methodptr*) * m->class->index) & 0x0000ffff);
+
+		/* patch method offset */
+
+		*((s4 *) (sp + 4 * 8 + 0)) |=
+			(s4) ((sizeof(methodptr) * (m - m->class->methods)) & 0x0000ffff);
+	}
 
 	return true;
 }
@@ -611,8 +565,6 @@ bool patcher_invokeinterface(u1 *sp)
 
 bool patcher_checkcast_instanceof_flags(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	s4                 disp;
 	u1                *pv;
@@ -620,27 +572,15 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	cr   = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch class flags */
 
 	*((s4 *) (pv + disp)) = (s4) c->flags;
@@ -653,7 +593,7 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 }
 
 
-/* patcher_checkcast_instanceof_interface **************************************
+/* patcher_checkcast_interface *************************************************
 
    Machine code:
 
@@ -667,48 +607,117 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 *******************************************************************************/
 
-bool patcher_checkcast_instanceof_interface(u1 *sp)
+bool patcher_checkcast_interface(u1 *sp)
 {
 	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	classinfo         *c;
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	ra = (u1 *)                *((ptrint *) (sp + 7 * 8));
+	cr = (constant_classref *) *((ptrint *) (sp + 2 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
+	/* if we show disassembly, we have to skip the nop's */
 
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
+	if (opt_shownops) {
+		ra = ra + PATCHER_CALL_SIZE;
+
+		/* patch super class index */
+
+		*((s4 *) (ra + 2 * 4)) |= (s4) (-(c->index) & 0x0000ffff);
+		/* 	*((s4 *) (ra + 5 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) - */
+		/* 									 c->index * sizeof(methodptr*)) & 0x0000ffff); */
+		*((s4 *) (ra + 6 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   c->index * sizeof(methodptr*)) & 0x0000ffff);
+
+		/* synchronize instruction cache */
+
+		md_icacheflush(ra + 2 * 4, 5 * 4);
+	}
+	else {
+		/* patch super class index */
+
+		*((s4 *) (sp + 2 * 4)) |= (s4) (-(c->index) & 0x0000ffff);
+		/* 	*((s4 *) (ra + 5 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) - */
+		/* 									 c->index * sizeof(methodptr*)) & 0x0000ffff); */
+		*((s4 *) (ra + 6 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   c->index * sizeof(methodptr*)) & 0x0000ffff);
+
+		/* synchronize instruction cache */
+
+		md_icacheflush(ra + 6 * 4, 1 * 4);
+	}
+
+	return true;
+}
+
+
+/* patcher_instanceof_interface ************************************************
+
+   Machine code:
+
+   <patched call position>
+   dd030000    ld       v1,0(a4)
+   8c79001c    lw       t9,28(v1)
+   27390000    addiu    t9,t9,0
+   1b200082    blez     t9,zero,0x000000001051843c
+   00000000    nop
+   dc790000    ld       t9,0(v1)
+
+*******************************************************************************/
+
+bool patcher_instanceof_interface(u1 *sp)
+{
+	u1                *ra;
+	constant_classref *cr;
+	classinfo         *c;
+
+	/* get stuff from the stack */
+
+	ra = (u1 *)                *((ptrint *) (sp + 7 * 8));
+	cr = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+
+	/* get the fieldinfo */
+
+	if (!(c = resolve_classref_eager(cr)))
+		return false;
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 2 * 4;
+	if (opt_shownops) {
+		ra = ra + PATCHER_CALL_SIZE;
 
-	/* patch super class index */
+		/* patch super class index */
 
-	*((s4 *) (ra + 2 * 4)) |= (s4) (-(c->index) & 0x0000ffff);
+		*((s4 *) (ra + 2 * 4)) |= (s4) (-(c->index) & 0x0000ffff);
+		*((s4 *) (ra + 5 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   c->index * sizeof(methodptr*)) & 0x0000ffff);
 
-	*((s4 *) (ra + 5 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) -
-									 c->index * sizeof(methodptr*)) & 0x0000ffff);
+		/* synchronize instruction cache */
 
-	/* synchronize instruction cache */
+		md_icacheflush(ra + 2 * 4, 4 * 4);
+	}
+	else {
+		/* patch super class index */
 
-	if (opt_showdisassemble)
-		md_icacheflush(ra - 2 * 4, 8 * 4);
-	else
-		md_icacheflush(ra, 6 * 4);
+		*((s4 *) (sp + 4 * 8 + 0)) |= (s4) (-(c->index) & 0x0000ffff);
+		*((s4 *) (ra + 5 * 4)) |=
+			(s4) ((OFFSET(vftbl_t, interfacetable[0]) -
+				   c->index * sizeof(methodptr*)) & 0x0000ffff);
+
+		/* synchronize instruction cache */
+
+		md_icacheflush(ra + 5 * 4, 1 * 4);
+	}
 
 	return true;
 }
@@ -726,8 +735,6 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 
 bool patcher_checkcast_instanceof_class(u1 *sp)
 {
-	u1                *ra;
-	u4                 mcode[2];
 	constant_classref *cr;
 	s4                 disp;
 	u1                *pv;
@@ -735,27 +742,15 @@ bool patcher_checkcast_instanceof_class(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)                *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                       *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                       *((u4 *)     (sp + 3 * 8 + 4));
-	cr       = (constant_classref *) *((ptrint *) (sp + 2 * 8));
-	disp     =                       *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)                *((ptrint *) (sp + 0 * 8));
+	cr   = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+	disp =                       *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)                *((ptrint *) (sp + 0 * 8));
 
 	/* get the fieldinfo */
 
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch super class' vftbl */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c->vftbl;
@@ -776,16 +771,11 @@ bool patcher_checkcast_instanceof_class(u1 *sp)
 
 bool patcher_clinit(u1 *sp)
 {
-	u1        *ra;
-	u4         mcode[2];
 	classinfo *c;
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)        *((ptrint *) (sp + 5 * 8));
-	mcode[0] =               *((u4 *)     (sp + 3 * 8));
-	mcode[1] =               *((u4 *)     (sp + 3 * 8 + 4));
-	c        = (classinfo *) *((ptrint *) (sp + 2 * 8));
+	c = (classinfo *) *((ptrint *) (sp + 2 * 8));
 
 	/* check if the class is initialized */
 
@@ -793,15 +783,6 @@ bool patcher_clinit(u1 *sp)
 		if (!initialize_class(c))
 			return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	return true;
 }
 
@@ -817,32 +798,17 @@ bool patcher_clinit(u1 *sp)
 #ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
-	u1               *ra;
-	u4                mcode[2];
 	unresolved_class *uc;
-	classinfo        *c;
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)               *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                      *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                      *((u4 *)     (sp + 3 * 8 + 4));
-	uc       = (unresolved_class *) *((ptrint *) (sp + 2 * 8));
+	uc = (unresolved_class *) *((ptrint *) (sp + 2 * 8));
 
-	/* resolve the class */
+	/* resolve the class and check subtype constraints */
 
-	if (!resolve_class(uc, resolveEager, false, &c))
+	if (!resolve_class_eager_no_access_check(uc))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	return true;
 }
 #endif /* ENABLE_VERIFIER */
@@ -857,8 +823,6 @@ bool patcher_athrow_areturn(u1 *sp)
 #if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
-	u1          *ra;
-	u4           mcode[2];
 	methodinfo  *m;
 	s4           disp;
 	u1          *pv;
@@ -866,27 +830,15 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra       = (u1 *)         *((ptrint *) (sp + 5 * 8));
-	mcode[0] =                *((u4 *)     (sp + 3 * 8));
-	mcode[1] =                *((u4 *)     (sp + 3 * 8 + 4));
-	m        = (methodinfo *) *((ptrint *) (sp + 2 * 8));
-	disp     =                *((s4 *)     (sp + 1 * 8));
-	pv       = (u1 *)         *((ptrint *) (sp + 0 * 8));
+	m    = (methodinfo *) *((ptrint *) (sp + 2 * 8));
+	disp =                *((s4 *)     (sp + 1 * 8));
+	pv   = (u1 *)         *((ptrint *) (sp + 0 * 8));
 
 	/* resolve native function */
 
 	if (!(f = native_resolve_function(m)))
 		return false;
 
-	/* patch back original code */
-
-	*((u4 *) (ra + 0 * 4)) = mcode[0];
-	*((u4 *) (ra + 1 * 4)) = mcode[1];
-
-	/* synchronize instruction cache */
-
-	md_icacheflush(ra, 2 * 4);
-
 	/* patch native function pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) f;
