@@ -1,6 +1,6 @@
 /* src/vm/jit/powerpc/emit.c - PowerPC code emitter functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: emitfuncs.c 4398 2006-01-31 23:43:08Z twisti $
+   $Id: emit.c 7816 2007-04-25 19:38:46Z twisti $
 
 */
 
@@ -43,13 +37,23 @@
 
 #include "vm/jit/powerpc/codegen.h"
 
+#include "mm/memory.h"
+
+#include "threads/lock-common.h"
+
 #include "vm/builtin.h"
+#include "vm/exceptions.h"
+
+#include "vm/jit/abi.h"
 #include "vm/jit/asmpart.h"
+#include "vm/jit/codegen-common.h"
 #include "vm/jit/dseg.h"
 #include "vm/jit/emit-common.h"
 #include "vm/jit/jit.h"
 #include "vm/jit/replace.h"
 
+#include "vmcore/options.h"
+
 
 /* emit_load *******************************************************************
 
@@ -72,17 +76,22 @@ s4 emit_load(jitdata *jd, instruction *iptr, varinfo *src, s4 tempreg)
 
 		disp = src->vv.regoff * 4;
 
-		if (IS_FLT_DBL_TYPE(src->type)) {
-			if (IS_2_WORD_TYPE(src->type))
-				M_DLD(tempreg, REG_SP, disp);
-			else
-				M_FLD(tempreg, REG_SP, disp);
-		}
-		else {
-			if (IS_2_WORD_TYPE(src->type))
-				M_LLD(tempreg, REG_SP, disp);
-			else
-				M_ILD(tempreg, REG_SP, disp);
+		switch (src->type) {
+		case TYPE_INT:
+		case TYPE_ADR:
+			M_ILD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_LNG:
+			M_LLD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_FLT:
+			M_FLD(tempreg, REG_SP, disp);
+			break;
+		case TYPE_DBL:
+			M_DLD(tempreg, REG_SP, disp);
+			break;
+		default:
+			vm_abort("emit_load: unknown type %d", src->type);
 		}
 
 		reg = tempreg;
@@ -164,13 +173,14 @@ s4 emit_load_high(jitdata *jd, instruction *iptr, varinfo *src, s4 tempreg)
 
 /* emit_store ******************************************************************
 
-   XXX
+   Emit a possible store for the given variable.
 
 *******************************************************************************/
 
 void emit_store(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 {
 	codegendata *cd;
+	s4           disp;
 
 	/* get required compiler data */
 
@@ -179,17 +189,24 @@ void emit_store(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 	if (IS_INMEMORY(dst->flags)) {
 		COUNT_SPILLS;
 
-		if (IS_FLT_DBL_TYPE(dst->type)) {
-			if (IS_2_WORD_TYPE(dst->type))
-				M_DST(d, REG_SP, dst->vv.regoff * 4);
-			else
-				M_FST(d, REG_SP, dst->vv.regoff * 4);
-		}
-		else {
-			if (IS_2_WORD_TYPE(dst->type))
-				M_LST(d, REG_SP, dst->vv.regoff * 4);
-			else
-				M_IST(d, REG_SP, dst->vv.regoff * 4);
+		disp = dst->vv.regoff * 4;
+
+		switch (dst->type) {
+		case TYPE_INT:
+		case TYPE_ADR:
+			M_IST(d, REG_SP, disp);
+			break;
+		case TYPE_LNG:
+			M_LST(d, REG_SP, disp);
+			break;
+		case TYPE_FLT:
+			M_FST(d, REG_SP, disp);
+			break;
+		case TYPE_DBL:
+			M_DST(d, REG_SP, disp);
+			break;
+		default:
+			vm_abort("emit_store: unknown type %d", dst->type);
 		}
 	}
 }
@@ -201,18 +218,30 @@ void emit_store(jitdata *jd, instruction *iptr, varinfo *dst, s4 d)
 
 *******************************************************************************/
 
-void emit_copy(jitdata *jd, instruction *iptr, varinfo *src, varinfo *dst)
+void emit_copy(jitdata *jd, instruction *iptr)
 {
-	codegendata  *cd;
-	s4            s1, d;
+	codegendata *cd;
+	varinfo     *src;
+	varinfo     *dst;
+	s4           s1, d;
 
 	/* get required compiler data */
 
 	cd = jd->cd;
 
+	/* get source and destination variables */
+
+	src = VAROP(iptr->s1);
+	dst = VAROP(iptr->dst);
+
 	if ((src->vv.regoff != dst->vv.regoff) ||
 		(IS_INMEMORY(src->flags ^ dst->flags))) {
 
+		if ((src->type == TYPE_RET) || (dst->type == TYPE_RET)) {
+			/* emit nothing, as the value won't be used anyway */
+			return;
+		}
+
 		/* If one of the variables resides in memory, we can eliminate
 		   the register move from/to the temporary register with the
 		   order of getting the destination register and the load. */
@@ -235,15 +264,21 @@ void emit_copy(jitdata *jd, instruction *iptr, varinfo *src, varinfo *dst)
 		}
 
 		if (s1 != d) {
-			if (IS_FLT_DBL_TYPE(src->type))
+			switch (src->type) {
+			case TYPE_INT:
+			case TYPE_ADR:
+				M_MOV(s1, d);
+				break;
+			case TYPE_LNG:
+				M_MOV(GET_LOW_REG(s1), GET_LOW_REG(d));
+				M_MOV(GET_HIGH_REG(s1), GET_HIGH_REG(d));
+				break;
+			case TYPE_FLT:
+			case TYPE_DBL:
 				M_FMOV(s1, d);
-			else {
-				if (IS_2_WORD_TYPE(src->type)) {
-					M_MOV(GET_LOW_REG(s1), GET_LOW_REG(d));
-					M_MOV(GET_HIGH_REG(s1), GET_HIGH_REG(d));
-                }
-				else
-                    M_MOV(s1, d);
+				break;
+			default:
+				vm_abort("emit_copy: unknown type %d", src->type);
 			}
 		}
 
@@ -271,99 +306,202 @@ void emit_iconst(codegendata *cd, s4 d, s4 value)
 }
 
 
-/* emit_exception_stubs ********************************************************
+/* emit_branch *****************************************************************
 
-   Generates the code for the exception stubs.
+   Emits the code for conditional and unconditional branchs.
 
 *******************************************************************************/
 
-void emit_exception_stubs(jitdata *jd)
+void emit_branch(codegendata *cd, s4 disp, s4 condition, s4 reg, u4 opt)
 {
-	codegendata  *cd;
-	registerdata *rd;
-	exceptionref *eref;
-	s4            targetdisp;
-	s4            disp;
+	s4 checkdisp;
+	s4 branchdisp;
 
-	/* get required compiler data */
+	/* calculate the different displacements */
 
-	cd = jd->cd;
-	rd = jd->rd;
+	checkdisp  =  disp + 4;
+	branchdisp = (disp - 4) >> 2;
 
-	/* generate exception stubs */
+	/* check which branch to generate */
 
-	targetdisp = 0;
+	if (condition == BRANCH_UNCONDITIONAL) {
+		/* check displacement for overflow */
 
-	for (eref = cd->exceptionrefs; eref != NULL; eref = eref->next) {
-		gen_resolvebranch(cd->mcodebase + eref->branchpos, 
-						  eref->branchpos, cd->mcodeptr - cd->mcodebase);
+		if ((checkdisp < (s4) 0xfe000000) || (checkdisp > (s4) 0x01fffffc)) {
+			/* if the long-branches flag isn't set yet, do it */
 
-		MCODECHECK(100);
-
-		/* Move the value register to a temporary register, if
-		   there is the need for it. */
-
-		if (eref->reg != -1)
-			M_MOV(eref->reg, REG_ITMP1);
-
-		/* calcuate exception address */
-
-		M_LDA(REG_ITMP2_XPC, REG_PV, eref->branchpos - 4);
-
-		/* move function to call into REG_ITMP3 */
-
-		disp = dseg_add_functionptr(cd, eref->function);
-		M_ALD(REG_ITMP3, REG_PV, disp);
-
-		if (targetdisp == 0) {
-		    targetdisp = ((u4 *) cd->mcodeptr) - ((u4 *) cd->mcodebase);
-
-			if (jd->isleafmethod) {
-				M_MFLR(REG_ZERO);
-				M_AST(REG_ZERO, REG_SP, cd->stackframesize * 4 + LA_LR_OFFSET);
+			if (!CODEGENDATA_HAS_FLAG_LONGBRANCHES(cd)) {
+				cd->flags |= (CODEGENDATA_FLAG_ERROR |
+							  CODEGENDATA_FLAG_LONGBRANCHES);
 			}
 
-			M_MOV(REG_PV, rd->argintregs[0]);
-			M_MOV(REG_SP, rd->argintregs[1]);
-
-			if (jd->isleafmethod)
-				M_MOV(REG_ZERO, rd->argintregs[2]);
-			else
-				M_ALD(rd->argintregs[2],
-					  REG_SP, cd->stackframesize * 4 + LA_LR_OFFSET);
-
-			M_MOV(REG_ITMP2_XPC, rd->argintregs[3]);
-			M_MOV(REG_ITMP1, rd->argintregs[4]);
-
-			M_STWU(REG_SP, REG_SP, -(LA_SIZE + 6 * 4));
-			M_AST(REG_ITMP2_XPC, REG_SP, LA_SIZE + 5 * 4);
-
-			M_MTCTR(REG_ITMP3);
-			M_JSR;
-			M_MOV(REG_RESULT, REG_ITMP1_XPTR);
-
-			M_ALD(REG_ITMP2_XPC, REG_SP, LA_SIZE + 5 * 4);
-			M_IADD_IMM(REG_SP, LA_SIZE + 6 * 4, REG_SP);
-
-			if (jd->isleafmethod) {
-				/* XXX FIXME: REG_ZERO can cause problems here! */
-				assert(cd->stackframesize * 4 <= 32767);
-
-				M_ALD(REG_ZERO, REG_SP, cd->stackframesize * 4 + LA_LR_OFFSET);
-				M_MTLR(REG_ZERO);
-			}
-
-			disp = dseg_add_functionptr(cd, asm_handle_exception);
-			M_ALD(REG_ITMP3, REG_PV, disp);
-			M_MTCTR(REG_ITMP3);
-			M_RTS;
+			vm_abort("emit_branch: emit unconditional long-branch code");
 		}
 		else {
-			disp = (((u4 *) cd->mcodebase) + targetdisp) -
-				(((u4 *) cd->mcodeptr) + 1);
-			M_BR(disp);
+			M_BR(branchdisp);
 		}
 	}
+	else {
+		/* and displacement for overflow */
+
+		if ((checkdisp < (s4) 0xffff8000) || (checkdisp > (s4) 0x00007fff)) {
+			/* if the long-branches flag isn't set yet, do it */
+
+			if (!CODEGENDATA_HAS_FLAG_LONGBRANCHES(cd)) {
+				cd->flags |= (CODEGENDATA_FLAG_ERROR |
+							  CODEGENDATA_FLAG_LONGBRANCHES);
+			}
+
+			switch (condition) {
+			case BRANCH_EQ:
+				M_BNE(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_NE:
+				M_BEQ(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_LT:
+				M_BGE(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_GE:
+				M_BLT(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_GT:
+				M_BLE(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_LE:
+				M_BGT(1);
+				M_BR(branchdisp);
+				break;
+			case BRANCH_NAN:
+				vm_abort("emit_branch: long BRANCH_NAN");
+				break;
+			default:
+				vm_abort("emit_branch: unknown condition %d", condition);
+			}
+		}
+		else {
+			switch (condition) {
+			case BRANCH_EQ:
+				M_BEQ(branchdisp);
+				break;
+			case BRANCH_NE:
+				M_BNE(branchdisp);
+				break;
+			case BRANCH_LT:
+				M_BLT(branchdisp);
+				break;
+			case BRANCH_GE:
+				M_BGE(branchdisp);
+				break;
+			case BRANCH_GT:
+				M_BGT(branchdisp);
+				break;
+			case BRANCH_LE:
+				M_BLE(branchdisp);
+				break;
+			case BRANCH_NAN:
+				M_BNAN(branchdisp);
+				break;
+			default:
+				vm_abort("emit_branch: unknown condition %d", condition);
+			}
+		}
+	}
+}
+
+
+/* emit_arithmetic_check *******************************************************
+
+   Emit an ArithmeticException check.
+
+*******************************************************************************/
+
+void emit_arithmetic_check(codegendata *cd, instruction *iptr, s4 reg)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TST(reg);
+		M_BNE(1);
+		M_ALD_INTERN(REG_ZERO, REG_ZERO, EXCEPTION_HARDWARE_ARITHMETIC);
+	}
+}
+
+
+/* emit_arrayindexoutofbounds_check ********************************************
+
+   Emit a ArrayIndexOutOfBoundsException check.
+
+*******************************************************************************/
+
+void emit_arrayindexoutofbounds_check(codegendata *cd, instruction *iptr, s4 s1, s4 s2)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_ILD(REG_ITMP3, s1, OFFSET(java_arrayheader, size));
+		M_TRAPGEU(s2, REG_ITMP3);
+	}
+}
+
+
+/* emit_classcast_check ********************************************************
+
+   Emit a ClassCastException check.
+
+*******************************************************************************/
+
+void emit_classcast_check(codegendata *cd, instruction *iptr, s4 condition, s4 reg, s4 s1)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		switch (condition) {
+		case BRANCH_LE:
+			M_BGT(1);
+			break;
+		case BRANCH_EQ:
+			M_BNE(1);
+			break;
+		case BRANCH_GT:
+			M_BLE(1);
+			break;
+		default:
+			vm_abort("emit_classcast_check: unknown condition %d", condition);
+		}
+		M_ALD_INTERN(s1, REG_ZERO, EXCEPTION_HARDWARE_CLASSCAST);
+	}
+}
+
+
+/* emit_nullpointer_check ******************************************************
+
+   Emit a NullPointerException check.
+
+*******************************************************************************/
+
+void emit_nullpointer_check(codegendata *cd, instruction *iptr, s4 reg)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TST(reg);
+		M_BNE(1);
+		M_ALD_INTERN(REG_ZERO, REG_ZERO, EXCEPTION_HARDWARE_NULLPOINTER);
+	}
+}
+
+
+/* emit_exception_check ********************************************************
+
+   Emit an Exception check.
+
+*******************************************************************************/
+
+void emit_exception_check(codegendata *cd, instruction *iptr)
+{
+	if (INSTRUCTION_MUST_CHECK(iptr)) {
+		M_TST(REG_RESULT);
+		M_BNE(1);
+		M_ALD_INTERN(REG_ZERO, REG_ZERO, EXCEPTION_HARDWARE_EXCEPTION);
+	}
 }
 
 
@@ -485,14 +623,17 @@ void emit_patcher_stubs(jitdata *jd)
 
 *******************************************************************************/
 
+#if defined(ENABLE_REPLACEMENT)
 void emit_replacement_stubs(jitdata *jd)
 {
 	codegendata *cd;
 	codeinfo    *code;
 	rplpoint    *rplp;
-	u1          *savedmcodeptr;
 	s4           disp;
 	s4           i;
+#if !defined(NDEBUG)
+	u1          *savedmcodeptr;
+#endif
 
 	/* get required compiler data */
 
@@ -501,24 +642,23 @@ void emit_replacement_stubs(jitdata *jd)
 
 	rplp = code->rplpoints;
 
+	/* store beginning of replacement stubs */
+
+	code->replacementstubs = (u1*) (cd->mcodeptr - cd->mcodebase);
+
 	for (i = 0; i < code->rplpointcount; ++i, ++rplp) {
+		/* do not generate stubs for non-trappable points */
+
+		if (rplp->flags & RPLPOINT_FLAG_NOTRAP)
+			continue;
+
 		/* check code segment size */
 
 		MCODECHECK(100);
 
-		/* note start of stub code */
-
-		rplp->outcode = (u1 *) (cd->mcodeptr - cd->mcodebase);
-
-		/* make machine code for patching */
-
+#if !defined(NDEBUG)
 		savedmcodeptr = cd->mcodeptr;
-		cd->mcodeptr  = (u1 *) &(rplp->mcode) + 1;              /* big-endian */
-
-		disp = (ptrint) ((s4 *) rplp->outcode - (s4 *) rplp->pc) - 1;
-		M_BR(disp);
-
-		cd->mcodeptr = savedmcodeptr;
+#endif
 
 		/* create stack frame - keep 16-byte aligned */
 
@@ -536,8 +676,11 @@ void emit_replacement_stubs(jitdata *jd)
 		M_ALD(REG_ITMP3, REG_PV, disp);
 		M_MTCTR(REG_ITMP3);
 		M_RTS;
+
+		assert((cd->mcodeptr - savedmcodeptr) == 4*REPLACEMENT_STUB_SIZE);
 	}
 }
+#endif /* defined(ENABLE_REPLACEMENT) */
 
 
 /* emit_verbosecall_enter ******************************************************
@@ -546,9 +689,9 @@ void emit_replacement_stubs(jitdata *jd)
 
 *******************************************************************************/
 
-#if !defined(NDEBUG)
 void emit_verbosecall_enter(jitdata *jd)
 {
+#if !defined(NDEBUG)
 	methodinfo   *m;
 	codegendata  *cd;
 	registerdata *rd;
@@ -557,6 +700,9 @@ void emit_verbosecall_enter(jitdata *jd)
 	int stack_size;
 	methoddesc *md;
 
+	if (!JITDATA_HAS_FLAG_VERBOSECALL(jd))
+		return;
+
 	/* get required compiler data */
 
 	m  = jd->m;
@@ -607,28 +753,31 @@ void emit_verbosecall_enter(jitdata *jd)
 	/* integer argument regs                                             */
 	/* all integer argument registers have to be saved                   */
 	for (p = 0; p < 8; p++) {
-		d = rd->argintregs[p];
+		d = abi_registers_integer_argument[p];
 		/* save integer argument registers */
 		M_IST(d, REG_SP, LA_SIZE + 4 * 8 + 4 + p * 4);
 	}
 	p = 4;
 #endif
 	stack_off = LA_SIZE;
+
 	for (; p < md->paramcount && p < TRACE_ARGS_NUM; p++, stack_off += 8) {
 		t = md->paramtypes[p].type;
+
 		if (IS_INT_LNG_TYPE(t)) {
-			if (!md->params[p].inmemory) { /* Param in Arg Reg */
+			if (!md->params[p].inmemory) {
+				s1 = md->params[p].regoff;
+
 				if (IS_2_WORD_TYPE(t)) {
-					M_IST(rd->argintregs[GET_HIGH_REG(md->params[p].regoff)]
-						  , REG_SP, stack_off);
-					M_IST(rd->argintregs[GET_LOW_REG(md->params[p].regoff)]
-						  , REG_SP, stack_off + 4);
-				} else {
-					M_IST(REG_ITMP1, REG_SP, stack_off);
-					M_IST(rd->argintregs[md->params[p].regoff]
-						  , REG_SP, stack_off + 4);
+					M_IST(GET_HIGH_REG(s1), REG_SP, stack_off);
+					M_IST(GET_LOW_REG(s1), REG_SP, stack_off + 4);
 				}
-			} else { /* Param on Stack */
+				else {
+					M_IST(REG_ITMP1, REG_SP, stack_off);
+					M_IST(s1, REG_SP, stack_off + 4);
+				}
+			}
+			else {
 				s1 = (md->params[p].regoff + cd->stackframesize) * 4 
 					+ stack_size;
 				if (IS_2_WORD_TYPE(t)) {
@@ -636,22 +785,26 @@ void emit_verbosecall_enter(jitdata *jd)
 					M_IST(REG_ITMP2, REG_SP, stack_off);
 					M_ILD(REG_ITMP2, REG_SP, s1 + 4);
 					M_IST(REG_ITMP2, REG_SP, stack_off + 4);
-				} else {
+				}
+				else {
 					M_IST(REG_ITMP1, REG_SP, stack_off);
 					M_ILD(REG_ITMP2, REG_SP, s1);
 					M_IST(REG_ITMP2, REG_SP, stack_off + 4);
 				}
 			}
-		} else { /* IS_FLT_DBL_TYPE(t) */
-			if (!md->params[p].inmemory) { /* in Arg Reg */
-				s1 = rd->argfltregs[md->params[p].regoff];
+		}
+		else {
+			if (!md->params[p].inmemory) {
+				s1 = md->params[p].regoff;
+
 				if (!IS_2_WORD_TYPE(t)) {
 					M_IST(REG_ITMP1, REG_SP, stack_off);
 					M_FST(s1, REG_SP, stack_off + 4);
-				} else {
-					M_DST(s1, REG_SP, stack_off);
 				}
-			} else { /* on Stack */
+				else
+					M_DST(s1, REG_SP, stack_off);
+			}
+			else {
 				/* this should not happen */
 			}
 		}
@@ -660,40 +813,48 @@ void emit_verbosecall_enter(jitdata *jd)
 	/* load first 4 (==INT_ARG_CNT/2) arguments into integer registers */
 #if defined(__DARWIN__)
 	for (p = 0; p < 8; p++) {
-		d = rd->argintregs[p];
+		d = abi_registers_integer_argument[p];
 		M_ILD(d, REG_SP, LA_SIZE + p * 4);
 	}
 #else
 	/* LINUX */
 	/* Set integer and float argument registers vor trace_args call */
 	/* offset to saved integer argument registers                   */
+
 	stack_off = LA_SIZE + 4 * 8 + 4;
+
 	for (p = 0; (p < 4) && (p < md->paramcount); p++) {
 		t = md->paramtypes[p].type;
+
 		if (IS_INT_LNG_TYPE(t)) {
 			/* "stretch" int types */
 			if (!IS_2_WORD_TYPE(t)) {
-				M_CLR(rd->argintregs[2 * p]);
-				M_ILD(rd->argintregs[2 * p + 1], REG_SP,stack_off);
+				M_CLR(abi_registers_integer_argument[2 * p]);
+				M_ILD(abi_registers_integer_argument[2 * p + 1], REG_SP,stack_off);
 				stack_off += 4;
-			} else {
-				M_ILD(rd->argintregs[2 * p + 1], REG_SP,stack_off + 4);
-				M_ILD(rd->argintregs[2 * p], REG_SP,stack_off);
+			}
+			else {
+				M_ILD(abi_registers_integer_argument[2 * p + 1], REG_SP,stack_off + 4);
+				M_ILD(abi_registers_integer_argument[2 * p], REG_SP,stack_off);
 				stack_off += 8;
 			}
-		} else { /* Float/Dbl */
-			if (!md->params[p].inmemory) { /* Param in Arg Reg */
+		}
+		else {
+			if (!md->params[p].inmemory) {
 				/* use reserved Place on Stack (sp + 5 * 16) to copy  */
 				/* float/double arg reg to int reg                    */
-				s1 = rd->argfltregs[md->params[p].regoff];
+
+				s1 = md->params[p].regoff;
+
 				if (!IS_2_WORD_TYPE(t)) {
 					M_FST(s1, REG_SP, 5 * 16);
-					M_ILD(rd->argintregs[2 * p + 1], REG_SP, 5 * 16);
-					M_CLR(rd->argintregs[2 * p]);
-				} else {
+					M_ILD(abi_registers_integer_argument[2 * p + 1], REG_SP, 5 * 16);
+					M_CLR(abi_registers_integer_argument[2 * p]);
+				}
+				else {
 					M_DST(s1, REG_SP, 5 * 16);
-					M_ILD(rd->argintregs[2 * p + 1], REG_SP,  5 * 16 + 4);
-					M_ILD(rd->argintregs[2 * p], REG_SP, 5 * 16);
+					M_ILD(abi_registers_integer_argument[2 * p + 1], REG_SP,  5 * 16 + 4);
+					M_ILD(abi_registers_integer_argument[2 * p], REG_SP, 5 * 16);
 				}
 			}
 		}
@@ -708,7 +869,7 @@ void emit_verbosecall_enter(jitdata *jd)
 #else
 	M_AST(REG_ITMP1, REG_SP, LA_SIZE + 4 * 8);
 #endif
-	p = dseg_add_functionptr(cd, builtin_trace_args);
+	p = dseg_add_functionptr(cd, builtin_verbosecall_enter);
 	M_ALD(REG_ITMP2, REG_PV, p);
 	M_MTCTR(REG_ITMP2);
 	M_JSR;
@@ -717,28 +878,27 @@ void emit_verbosecall_enter(jitdata *jd)
 	/* restore integer argument registers from the reserved stack space */
 
 	stack_off = LA_SIZE;
-	for (p = 0; p < md->paramcount && p < TRACE_ARGS_NUM; 
-		 p++, stack_off += 8) {
+
+	for (p = 0; p < md->paramcount && p < TRACE_ARGS_NUM; p++, stack_off += 8) {
 		t = md->paramtypes[p].type;
 
 		if (IS_INT_LNG_TYPE(t)) {
 			if (!md->params[p].inmemory) {
+				s1 = md->params[p].regoff;
+
 				if (IS_2_WORD_TYPE(t)) {
-					M_ILD(rd->argintregs[GET_HIGH_REG(md->params[p].regoff)]
-						  , REG_SP, stack_off);
-					M_ILD(rd->argintregs[GET_LOW_REG(md->params[p].regoff)]
-						  , REG_SP, stack_off + 4);
-				} else {
-					M_ILD(rd->argintregs[md->params[p].regoff]
-						  , REG_SP, stack_off + 4);
+					M_ILD(GET_HIGH_REG(s1), REG_SP, stack_off);
+					M_ILD(GET_LOW_REG(s1), REG_SP, stack_off + 4);
 				}
+				else
+					M_ILD(s1, REG_SP, stack_off + 4);
 			}
 		}
 	}
 #else
 	/* LINUX */
 	for (p = 0; p < 8; p++) {
-		d = rd->argintregs[p];
+		d = abi_registers_integer_argument[p];
 		/* save integer argument registers */
 		M_ILD(d, REG_SP, LA_SIZE + 4 * 8 + 4 + p * 4);
 	}
@@ -751,25 +911,30 @@ void emit_verbosecall_enter(jitdata *jd)
 	/* mark trace code */
 
 	M_NOP;
-}
 #endif /* !defined(NDEBUG) */
+}
 
 
 /* emit_verbosecall_exit *******************************************************
 
    Generates the code for the call trace.
 
+   void builtin_verbosecall_exit(s8 l, double d, float f, methodinfo *m);
+
 *******************************************************************************/
 
-#if !defined(NDEBUG)
 void emit_verbosecall_exit(jitdata *jd)
 {
+#if !defined(NDEBUG)
 	methodinfo   *m;
 	codegendata  *cd;
 	registerdata *rd;
 	methoddesc   *md;
 	s4            disp;
 
+	if (!JITDATA_HAS_FLAG_VERBOSECALL(jd))
+		return;
+
 	/* get required compiler data */
 
 	m  = jd->m;
@@ -795,33 +960,22 @@ void emit_verbosecall_exit(jitdata *jd)
 	switch (md->returntype.type) {
 	case TYPE_INT:
 	case TYPE_ADR:
-#if defined(__DARWIN__)
-		M_MOV(REG_RESULT, rd->argintregs[2]);
-		M_CLR(rd->argintregs[1]);
-#else
-		M_MOV(REG_RESULT, rd->argintregs[3]);
-		M_CLR(rd->argintregs[2]);
-#endif
+		M_INTMOVE(REG_RESULT, REG_A1);
+		M_CLR(REG_A0);
 		break;
 
 	case TYPE_LNG:
-#if defined(__DARWIN__)
-		M_MOV(REG_RESULT2, rd->argintregs[2]);
-		M_MOV(REG_RESULT, rd->argintregs[1]);
-#else
-		M_MOV(REG_RESULT2, rd->argintregs[3]);
-		M_MOV(REG_RESULT, rd->argintregs[2]);
-#endif
+		M_LNGMOVE(REG_RESULT_PACKED, REG_A0_A1_PACKED);
 		break;
 	}
 
-	M_FLTMOVE(REG_FRESULT, rd->argfltregs[0]);
-	M_FLTMOVE(REG_FRESULT, rd->argfltregs[1]);
+	M_FLTMOVE(REG_FRESULT, REG_FA0);
+	M_FLTMOVE(REG_FRESULT, REG_FA1);
 
 	disp = dseg_add_address(cd, m);
-	M_ALD(rd->argintregs[0], REG_PV, disp);
+	M_ALD(REG_A2, REG_PV, disp);
 
-	disp = dseg_add_functionptr(cd, builtin_displaymethodstop);
+	disp = dseg_add_functionptr(cd, builtin_verbosecall_exit);
 	M_ALD(REG_ITMP2, REG_PV, disp);
 	M_MTCTR(REG_ITMP2);
 	M_JSR;
@@ -838,8 +992,8 @@ void emit_verbosecall_exit(jitdata *jd)
 	/* mark trace code */
 
 	M_NOP;
-}
 #endif /* !defined(NDEBUG) */
+}
 
 
 /*
