@@ -28,7 +28,7 @@
 
    Changes:
 
-   $Id: md-asm.h 4357 2006-01-22 23:33:38Z twisti $
+   $Id: md-asm.h 5179 2006-07-26 10:16:50Z twisti $
 
 */
 
@@ -36,7 +36,73 @@
 #ifndef _MD_ASM_H
 #define _MD_ASM_H
 
-#include <asm/ppc_asm.h>
+/* register defines ***********************************************************/
+
+#define r0      0
+#define r1      1
+#define r2      2
+#define r3      3
+#define r4      4
+#define r5      5
+#define r6      6
+#define r7      7
+#define r8      8
+#define r9      9
+#define r10     10
+#define r11     11
+#define r12     12
+#define r13     13
+#define r14     14
+#define r15     15
+#define r16     16
+#define r17     17
+#define r18     18
+#define r19     19
+#define r20     20
+#define r21     21
+#define r22     22
+#define r23     23
+#define r24     24
+#define r25     25
+#define r26     26
+#define r27     27
+#define r28     28
+#define r29     29
+#define r30     30
+#define r31     31
+
+#define fr0     0
+#define fr1     1
+#define fr2     2
+#define fr3     3
+#define fr4     4
+#define fr5     5
+#define fr6     6
+#define fr7     7
+#define fr8     8
+#define fr9     9
+#define fr10    10
+#define fr11    11
+#define fr12    12
+#define fr13    13
+#define fr14    14
+#define fr15    15
+#define fr16    16
+#define fr17    17
+#define fr18    18
+#define fr19    19
+#define fr20    20
+#define fr21    21
+#define fr22    22
+#define fr23    23
+#define fr24    24
+#define fr25    25
+#define fr26    26
+#define fr27    27
+#define fr28    28
+#define fr29    29
+#define fr30    30
+#define fr31    31
 
 
 /* register defines ***********************************************************/
@@ -44,7 +110,7 @@
 #define zero  r0
 #define sp    r1
 
-/* #define XXX   r2  -  system reserved register */
+/* #define foo   r2  -  system reserved register */
 
 #define a0    r3
 #define a1    r4
@@ -86,48 +152,48 @@
 #define xptr  itmp1
 #define xpc   itmp2
 
-#define mptr  r12
-#define mptrn 12
+#define mptr  itmp2
+#define mptrn itmp2
 
 
-#define ftmp3 fr0
+#define ftmp3 r0
 
-#define fa0   fr1
-#define fa1   fr2
-#define fa2   fr3
-#define fa3   fr4
-#define fa4   fr5
-#define fa5   fr6
-#define fa6   fr7
-#define fa7   fr8
+#define fa0   r1
+#define fa1   r2
+#define fa2   r3
+#define fa3   r4
+#define fa4   r5
+#define fa5   r6
+#define fa6   r7
+#define fa7   r8
 
-#define fa8   fr9
-#define fa9   fr10
-#define fa10  fr11
-#define fa11  fr12
-#define fa12  fr13
+#define fa8   r9
+#define fa9   r10
+#define fa10  r11
+#define fa11  r12
+#define fa12  r13
 
-#define fs0   fr14
-#define fs1   fr15
+#define fs0   r14
+#define fs1   r15
 
-#define ftmp1 fr16
-#define ftmp2 fr17
+#define ftmp1 r16
+#define ftmp2 r17
 
-#define ft0   fr18
-#define ft1   fr19
-#define ft2   fr20
-#define ft3   fr21
-#define ft4   fr22
-#define ft5   fr23
+#define ft0   r18
+#define ft1   r19
+#define ft2   r20
+#define ft3   r21
+#define ft4   r22
+#define ft5   r23
 
-#define fs2   fr24
-#define fs3   fr25
-#define fs4   fr26
-#define fs5   fr27
-#define fs6   fr28
-#define fs7   fr29
-#define fs8   fr30
-#define fs9   fr31
+#define fs2   r24
+#define fs3   r25
+#define fs4   r26
+#define fs5   r27
+#define fs6   r28
+#define fs7   r29
+#define fs8   r30
+#define fs9   r31
 
 #define fv0   fa0
 
