@@ -1,6 +1,6 @@
 /* src/vm/jit/powerpc/patcher.c - PowerPC code patching functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: patcher.c 5281 2006-08-28 15:18:54Z twisti $
+   $Id: patcher.c 7596 2007-03-28 21:05:53Z twisti $
 
 */
 
@@ -41,17 +35,22 @@
 
 #include "mm/memory.h"
 #include "native/native.h"
+
 #include "vm/builtin.h"
-#include "vm/class.h"
 #include "vm/exceptions.h"
-#include "vm/field.h"
 #include "vm/initialize.h"
-#include "vm/options.h"
-#include "vm/resolve.h"
-#include "vm/references.h"
+
 #include "vm/jit/asmpart.h"
 #include "vm/jit/patcher.h"
+#include "vm/jit/md.h"
 #include "vm/jit/methodheader.h"
+#include "vm/jit/stacktrace.h"
+
+#include "vmcore/class.h"
+#include "vmcore/field.h"
+#include "vmcore/options.h"
+#include "vm/resolve.h"
+#include "vmcore/references.h"
 
 
 /* patcher_wrapper *************************************************************
@@ -176,15 +175,14 @@ bool patcher_initialize_class(u1 *sp)
 bool patcher_resolve_class(u1 *sp)
 {
 	unresolved_class *uc;
-	classinfo        *c;
 
 	/* get stuff from the stack */
 
 	uc = (unresolved_class *) *((ptrint *) (sp + 2 * 4));
 
-	/* resolve the class */
+	/* resolve the class and check subtype constraints */
 
-	if (!resolve_class(uc, resolveEager, false, &c))
+	if (!resolve_class_eager_no_access_check(uc))
 		return false;
 
 	return true;
@@ -454,9 +452,9 @@ bool patcher_get_putfield(u1 *sp)
 	if (!(fi = resolve_field_eager(uf)))
 		return false;
 
-	/* if we show disassembly, we have to skip the nop */
+	/* if we show NOPs, we have to skip them */
 
-	if (opt_showdisassemble) {
+	if (opt_shownops) {
 		/* patch the field's offset */
 
 		if (IS_LNG_TYPE(fi->type)) {
@@ -580,9 +578,9 @@ bool patcher_invokevirtual(u1 *sp)
 	if (!(m = resolve_method_eager(um)))
 		return false;
 
-	/* if we show disassembly, we have to skip the nop */
+	/* if we show NOPs, we have to skip them */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 1 * 4;
 
 	/* patch vftbl index */
@@ -629,9 +627,9 @@ bool patcher_invokeinterface(u1 *sp)
 	if (!(m = resolve_method_eager(um)))
 		return false;
 
-	/* if we show disassembly, we have to skip the nop */
+	/* if we show NOPs, we have to skip them */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 1 * 4;
 
 	/* patch interfacetable index */
@@ -659,20 +657,21 @@ bool patcher_invokeinterface(u1 *sp)
 }
 
 
-/* patcher_checkcast_instanceof_interface **************************************
+/* patcher_checkcast_interface *************************************************
 
    Machine code:
 
    <patched call position>
-   81870000    lwz   r12,0(r7)
-   800c0010    lwz   r0,16(r12)
-   34000000    addic.        r0,r0,0
-   408101fc    ble-  0x3002e518
-   800c0000    lwz   r0,0(r12)
+   81870000    lwz     r12,0(r7)
+   800c0010    lwz     r0,16(r12)
+   34000000    addic.  r0,r0,0
+   41810008    bgt-    0x014135d8
+   83c00003    lwz     r30,3(0)
+   800c0000    lwz     r0,0(r12)
 
 *******************************************************************************/
 
-bool patcher_checkcast_instanceof_interface(u1 *sp)
+bool patcher_checkcast_interface(u1 *sp)
 {
 	u1                *ra;
 	constant_classref *cr;
@@ -689,9 +688,63 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 	if (!(c = resolve_classref_eager(cr)))
 		return false;
 
-	/* if we show disassembly, we have to skip the nop */
+	/* if we show NOPs, we have to skip them */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
+		ra = ra + 1 * 4;
+
+	/* patch super class index */
+
+	disp = -(c->index);
+
+	*((s4 *) (ra + 2 * 4)) |= (disp & 0x0000ffff);
+
+	disp = OFFSET(vftbl_t, interfacetable[0]) - c->index * sizeof(methodptr*);
+
+	*((s4 *) (ra + 5 * 4)) |= (disp & 0x0000ffff);
+
+	/* synchronize instruction cache */
+
+	md_icacheflush(ra + 2 * 4, 4 * 4);
+
+	return true;
+}
+
+
+/* patcher_instanceof_interface ************************************************
+
+   Machine code:
+
+   <patched call position>
+   81870000    lwz     r12,0(r7)
+   800c0010    lwz     r0,16(r12)
+   34000000    addic.  r0,r0,0
+   41810008    bgt-    0x014135d8
+   83c00003    lwz     r30,3(0)
+   800c0000    lwz     r0,0(r12)
+
+*******************************************************************************/
+
+bool patcher_instanceof_interface(u1 *sp)
+{
+	u1                *ra;
+	constant_classref *cr;
+	classinfo         *c;
+	s4                 disp;
+
+	/* get stuff from the stack */
+
+	ra = (u1 *)                *((ptrint *) (sp + 5 * 4));
+	cr = (constant_classref *) *((ptrint *) (sp + 2 * 4));
+
+	/* get the fieldinfo */
+
+	if (!(c = resolve_classref_eager(cr)))
+		return false;
+
+	/* if we show NOPs, we have to skip them */
+
+	if (opt_shownops)
 		ra = ra + 1 * 4;
 
 	/* patch super class index */
