@@ -1,6 +1,6 @@
 /* src/vm/jit/verify/typecheck-stackbased.c - stack-based verifier
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,27 +22,27 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Edwin Steiner
-
-   Changes: 
-
    $Id$
 
 */
 
 
 #include "config.h"
-#include "vm/types.h"
-#include "vm/global.h"
 
 #include <assert.h>
 
-#include <vm/jit/stack.h>
-#include <vm/jit/parse.h>
-#include <vm/jit/show.h>
-#include <typecheck-common.h>
+#include "vm/types.h"
+
+#include "vm/builtin.h"
+#include "mm/memory.h"
+
+#include "vm/global.h"
+
+#include "vm/jit/parse.h"
+#include "vm/jit/show.h"
+#include "vm/jit/stack.h"
+#include "vm/jit/verify/typecheck-common.h"
+
 
 /* this #if runs over the whole file: */
 #if defined(ENABLE_VERIFIER)
@@ -141,8 +141,7 @@ static void typecheck_stackbased_show_state(verifier_state *state,
 
 #define REACH(target)                                                \
     do {                                                             \
-        tbptr = BLOCK_OF((target).insindex);                         \
-        REACH_BLOCK(tbptr);                                          \
+        REACH_BLOCK((target).block);                                 \
     } while (0)
 
 #undef TYPECHECK_INT
@@ -544,7 +543,7 @@ static typedescriptor *typecheck_stackbased_jsr(verifier_state *state,
 
 	jd = state->jd;
 
-	tbptr = BLOCK_OF(state->iptr->sx.s23.s3.jsrtarget.insindex);
+	tbptr = state->iptr->sx.s23.s3.jsrtarget.block;
 	jsr = state->jsrinfos[tbptr->nr];
 
 	if (jsr && tbptr->flags == BBFINISHED) {
@@ -599,15 +598,18 @@ static typedescriptor *typecheck_stackbased_jsr(verifier_state *state,
 		jsr->next = state->topjsr;
 		state->topjsr = jsr;
 
-		/* XXX ugly */
-		assert(BLOCK_OF(state->iptr->sx.s23.s3.jsrtarget.insindex)->flags == BBTYPECHECK_REACHED);
+		assert(state->iptr->sx.s23.s3.jsrtarget.block->flags == BBTYPECHECK_REACHED);
+
 		tbptr->flags = BBFINISHED;
+
 		for (tbptr = state->basicblocks; tbptr != NULL; tbptr = tbptr->next) {
 			jsr->blockflags[tbptr->nr] = tbptr->flags;
+
 			if (tbptr->flags == BBTYPECHECK_REACHED)
 				tbptr->flags = BBFINISHED;
 		}
-		BLOCK_OF(state->iptr->sx.s23.s3.jsrtarget.insindex)->flags = BBTYPECHECK_REACHED;
+
+		state->iptr->sx.s23.s3.jsrtarget.block->flags = BBTYPECHECK_REACHED;
 	}
 
 	/* register this block as a caller, if not already done */
