@@ -1,6 +1,6 @@
 /* src/vm/jit/x86_64/patcher.c - x86_64 code patching functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: patcher.c 5233 2006-08-14 10:59:39Z twisti $
+   $Id: patcher.c 7596 2007-03-28 21:05:53Z twisti $
 
 */
 
@@ -39,16 +33,21 @@
 #include "vm/jit/x86_64/codegen.h"
 
 #include "mm/memory.h"
+
 #include "native/native.h"
+
 #include "vm/builtin.h"
-#include "vm/class.h"
 #include "vm/exceptions.h"
-#include "vm/field.h"
 #include "vm/initialize.h"
-#include "vm/options.h"
-#include "vm/references.h"
-#include "vm/resolve.h"
+
 #include "vm/jit/patcher.h"
+#include "vm/jit/stacktrace.h"
+
+#include "vmcore/class.h"
+#include "vmcore/field.h"
+#include "vmcore/options.h"
+#include "vmcore/references.h"
+#include "vm/resolve.h"
 
 
 /* patcher_wrapper *************************************************************
@@ -169,7 +168,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the field value's address */
@@ -215,7 +214,7 @@ bool patcher_get_putfield(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the field's offset: we check for the field type, because the     */
@@ -279,7 +278,7 @@ bool patcher_putfieldconst(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the field's offset */
@@ -342,7 +341,7 @@ bool patcher_aconst(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the classinfo pointer */
@@ -390,7 +389,7 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the classinfo pointer */
@@ -436,7 +435,7 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch the classinfo pointer */
@@ -483,8 +482,8 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 5;
+	if (opt_shownops)
+		ra = ra + PATCHER_CALL_SIZE;
 
 	/* patch stubroutine */
 
@@ -530,7 +529,7 @@ bool patcher_invokevirtual(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch vftbl index */
@@ -578,7 +577,7 @@ bool patcher_invokeinterface(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch interfacetable index */
@@ -630,7 +629,7 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch class flags */
@@ -641,20 +640,20 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 }
 
 
-/* patcher_checkcast_instanceof_interface **************************************
+/* patcher_checkcast_interface *************************************************
 
    Machine code:
 
    <patched call position>
    45 8b 9a 1c 00 00 00             mov    0x1c(%r10),%r11d
-   49 81 eb 00 00 00 00             sub    $0x0,%r11
-   4d 85 db                         test   %r11,%r11
-   0f 8e 94 04 00 00                jle    0x00002aaaaab018f8
+   41 81 fb 00 00 00 00             cmp    $0x0,%r11d
+   0f 8f 08 00 00 00                jg     0x00002aaaaae511d5
+   48 8b 0c 25 03 00 00 00          mov    0x3,%rcx
    4d 8b 9a 00 00 00 00             mov    0x0(%r10),%r11
 
 *******************************************************************************/
 
-bool patcher_checkcast_instanceof_interface(u1 *sp)
+bool patcher_checkcast_interface(u1 *sp)
 {
 	u1                *ra;
 	u8                 mcode;
@@ -678,14 +677,14 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
-		ra = ra + 5;
+	if (opt_shownops)
+		ra = ra + PATCHER_CALL_SIZE;
 
 	/* patch super class index */
 
 	*((s4 *) (ra + 7 + 3)) = (s4) c->index;
 
-	*((s4 *) (ra + 7 + 7 + 3 + 6 + 3)) =
+	*((s4 *) (ra + 7 + 7 + 6 + 8 + 3)) =
 		(s4) (OFFSET(vftbl_t, interfacetable[0]) -
 			  c->index * sizeof(methodptr*));
 
@@ -730,7 +729,7 @@ bool patcher_checkcast_class(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch super class' vftbl */
@@ -742,6 +741,57 @@ bool patcher_checkcast_class(u1 *sp)
 }
 
 
+/* patcher_instanceof_interface ************************************************
+
+   Machine code:
+
+   <patched call position>
+   45 8b 9a 1c 00 00 00             mov    0x1c(%r10),%r11d
+   41 81 fb 00 00 00 00             cmp    $0x0,%r11d
+   0f 8e 94 04 00 00                jle    0x00002aaaaab018f8
+   4d 8b 9a 00 00 00 00             mov    0x0(%r10),%r11
+
+*******************************************************************************/
+
+bool patcher_instanceof_interface(u1 *sp)
+{
+	u1                *ra;
+	u8                 mcode;
+	constant_classref *cr;
+	classinfo         *c;
+
+	/* get stuff from the stack */
+
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 8));
+	mcode =                       *((u8 *)     (sp + 3 * 8));
+	cr    = (constant_classref *) *((ptrint *) (sp + 2 * 8));
+
+	/* get the fieldinfo */
+
+	if (!(c = resolve_classref_eager(cr)))
+		return false;
+
+	/* patch back original code */
+
+	*((u8 *) ra) = mcode;
+
+	/* if we show disassembly, we have to skip the nop's */
+
+	if (opt_shownops)
+		ra = ra + PATCHER_CALL_SIZE;
+
+	/* patch super class index */
+
+	*((s4 *) (ra + 7 + 3)) = (s4) c->index;
+
+	*((s4 *) (ra + 7 + 7 + 6 + 3)) =
+		(s4) (OFFSET(vftbl_t, interfacetable[0]) -
+			  c->index * sizeof(methodptr*));
+
+	return true;
+}
+
+
 /* patcher_instanceof_class ****************************************************
 
    Machine code:
@@ -775,7 +825,7 @@ bool patcher_instanceof_class(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch super class' vftbl */
@@ -838,7 +888,6 @@ bool patcher_athrow_areturn(u1 *sp)
 	u1               *ra;
 	u8                mcode;
 	unresolved_class *uc;
-	classinfo        *c;
 
 	/* get stuff from the stack */
 
@@ -846,9 +895,9 @@ bool patcher_athrow_areturn(u1 *sp)
 	mcode =                      *((u8 *)     (sp + 3 * 8));
 	uc    = (unresolved_class *) *((ptrint *) (sp + 2 * 8));
 
-	/* resolve the class */
+	/* resolve the class and check subtype constraints */
 
-	if (!resolve_class(uc, resolveEager, false, &c))
+	if (!resolve_class_eager_no_access_check(uc))
 		return false;
 
 	/* patch back original code */
@@ -895,7 +944,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* if we show disassembly, we have to skip the nop's */
 
-	if (opt_showdisassemble)
+	if (opt_shownops)
 		ra = ra + 5;
 
 	/* patch native function pointer */
