@@ -1,6 +1,6 @@
 /* src/vm/resolve.c - resolving classes/interfaces/fields/methods
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Edwin Steiner
-
-   Changes: Christan Thalinger
-
-   $Id: resolve.c 5768 2006-10-13 12:47:03Z edwin $
+   $Id: resolve.c 7983 2007-05-30 20:04:42Z twisti $
 
 */
 
@@ -37,20 +31,24 @@
 
 #include <assert.h>
 
+#include "vm/types.h"
+
 #include "mm/memory.h"
-#include "vm/resolve.h"
+
 #include "vm/access.h"
-#include "vm/classcache.h"
-#include "vm/descriptor.h"
 #include "vm/exceptions.h"
 #include "vm/global.h"
-#include "vm/linker.h"
-#include "vm/loader.h"
-#include "vm/options.h"
-#include "vm/stringlocal.h"
+
 #include "vm/jit/jit.h"
 #include "vm/jit/verify/typeinfo.h"
 
+#include "vmcore/classcache.h"
+#include "vmcore/descriptor.h"
+#include "vmcore/linker.h"
+#include "vmcore/loader.h"
+#include "vmcore/options.h"
+#include "vm/resolve.h"
+
 
 /******************************************************************************/
 /* DEBUG HELPERS                                                              */
@@ -104,10 +102,13 @@ bool resolve_class_from_name(classinfo *referer,
 							 bool link,
 							 classinfo **result)
 {
-	classinfo *cls = NULL;
-	char *utf_ptr;
-	int len;
-	
+	classinfo *cls;
+	char      *utf_ptr;
+	int        len;
+	char      *msg;
+	s4         msglen;
+	utf       *u;
+
 	assert(result);
 	assert(referer);
 	assert(classname);
@@ -189,19 +190,27 @@ bool resolve_class_from_name(classinfo *referer,
 #endif
 	
 	/* check access rights of referer to refered class */
-	if (checkaccess && !access_is_accessible_class(referer,cls)) {
-		int msglen;
-		char *message;
 
-		msglen = utf_bytes(cls->name) + utf_bytes(referer->name) + 100;
-		message = MNEW(char, msglen);
-		strcpy(message, "class is not accessible (");
-		utf_cat_classname(message, cls->name);
-		strcat(message, " from ");
-		utf_cat_classname(message, referer->name);
-		strcat(message, ")");
-		*exceptionptr = new_exception_message(string_java_lang_IllegalAccessException, message);
-		MFREE(message,char,msglen);
+	if (checkaccess && !access_is_accessible_class(referer,cls)) {
+		msglen =
+			utf_bytes(cls->name) +
+			utf_bytes(referer->name) +
+			100;
+
+		msg = MNEW(char, msglen);
+
+		strcpy(msg, "class is not accessible (");
+		utf_cat_classname(msg, cls->name);
+		strcat(msg, " from ");
+		utf_cat_classname(msg, referer->name);
+		strcat(msg, ")");
+
+		u = utf_new_char(msg);
+
+		MFREE(msg, char, msglen);
+
+		exceptions_throw_illegalaccessexception(u);
+
 		return false; /* exception */
 	}
 
@@ -264,7 +273,10 @@ bool resolve_classref(methodinfo *refmethod,
 /* resolve_classref_or_classinfo ***********************************************
  
    Resolve a symbolic class reference if necessary
-  
+
+   NOTE: If given, refmethod->class is used as the referring class.
+         Otherwise, cls.ref->referer is used.
+
    IN:
        refmethod........the method from which resolution was triggered
                         (may be NULL if not applicable)
@@ -298,6 +310,7 @@ bool resolve_classref_or_classinfo(methodinfo *refmethod,
 								   classinfo **result)
 {
 	classinfo         *c;
+	classinfo         *referer;
 	
 	assert(cls.any);
 	assert(mode == resolveEager || mode == resolveLazy);
@@ -314,7 +327,18 @@ bool resolve_classref_or_classinfo(methodinfo *refmethod,
 	if (IS_CLASSREF(cls)) {
 		/* we must resolve this reference */
 
-		if (!resolve_class_from_name(cls.ref->referer, refmethod, cls.ref->name,
+		/* determine which class to use as the referer */
+
+		/* Common cases are refmethod == NULL or both referring classes */
+		/* being the same, so the referer usually is cls.ref->referer.    */
+		/* There is one important case where it is not: When we do a      */
+		/* deferred assignability check to a formal argument of a method, */
+		/* we must use refmethod->class (the caller's class) to resolve   */
+		/* the type of the formal argument.                               */
+
+		referer = (refmethod) ? refmethod->class : cls.ref->referer;
+
+		if (!resolve_class_from_name(referer, refmethod, cls.ref->name,
 									 mode, checkaccess, link, &c))
 			goto return_exception;
 
@@ -349,6 +373,33 @@ bool resolve_classref_or_classinfo(methodinfo *refmethod,
 }
 
 
+/* resolve_classref_or_classinfo_eager *****************************************
+ 
+   Resolve a symbolic class reference eagerly if necessary.
+   No attempt is made to link the class.
+
+   IN:
+       cls..............class reference or classinfo
+       checkaccess......if true, access rights to the class are checked
+  
+   RETURN VALUE:
+       classinfo *......the resolved class
+       NULL.............an exception has been thrown
+   
+*******************************************************************************/
+
+classinfo *resolve_classref_or_classinfo_eager(classref_or_classinfo cls,
+											   bool checkaccess)
+{
+	classinfo *c;
+
+	if (!resolve_classref_or_classinfo(NULL, cls, resolveEager, checkaccess, false, &c))
+		return NULL;
+
+	return c;
+}
+
+
 /* resolve_class_from_typedesc *************************************************
  
    Return a classinfo * for the given type descriptor
@@ -459,9 +510,12 @@ static resolve_result_t resolve_subtype_check(methodinfo *refmethod,
 											  resolve_mode_t mode,
 											  resolve_err_t error)
 {
-	classinfo *subclass;
-	typeinfo subti;
-	typecheck_result r;
+	classinfo        *subclass;
+	typeinfo          subti;
+	typecheck_result  r;
+	char             *msg;
+	s4                msglen;
+	utf              *u;
 
 	assert(refmethod);
 	assert(subtype.any);
@@ -475,7 +529,7 @@ static resolve_result_t resolve_subtype_check(methodinfo *refmethod,
 		/* the subclass could not be resolved. therefore we are sure that  */
 		/* no instances of this subclass will ever exist -> skip this test */
 		/* XXX this assumes that class loading has invariant results (as in JVM spec) */
-		*exceptionptr = NULL;
+		exceptions_clear_exception();
 		return resolveSucceeded;
 	}
 	if (!subclass)
@@ -516,27 +570,38 @@ check_again:
 	if (!r) {
 		/* sub class relationship is false */
 
-		char *message;
-		int msglen;
-
 #if defined(RESOLVE_VERBOSE)
 		printf("SUBTYPE CHECK FAILED!\n");
 #endif
 
-		msglen = utf_bytes(subclass->name) + utf_bytes(CLASSREF_OR_CLASSINFO_NAME(supertype)) + 200;
-		message = MNEW(char, msglen);
-		strcpy(message, (error == resolveIllegalAccessError) ?
-				"illegal access to protected member ("
-				: "subtype constraint violated (");
-		utf_cat_classname(message, subclass->name);
-		strcat(message, " is not a subclass of ");
-		utf_cat_classname(message, CLASSREF_OR_CLASSINFO_NAME(supertype));
-		strcat(message, ")");
+		msglen =
+			utf_bytes(subclass->name) +
+			utf_bytes(CLASSREF_OR_CLASSINFO_NAME(supertype))
+			+ 200;
+
+		msg = MNEW(char, msglen);
+
+		strcpy(msg, (error == resolveIllegalAccessError) ?
+			   "illegal access to protected member (" :
+			   "subtype constraint violated (");
+
+		utf_cat_classname(msg, subclass->name);
+		strcat(msg, " is not a subclass of ");
+		utf_cat_classname(msg, CLASSREF_OR_CLASSINFO_NAME(supertype));
+		strcat(msg, ")");
+
+		u = utf_new_char(msg);
+
 		if (error == resolveIllegalAccessError)
-			*exceptionptr = new_exception_message(string_java_lang_IllegalAccessException, message);
+			exceptions_throw_illegalaccessexception(u);
 		else
-			*exceptionptr = exceptions_new_linkageerror(message, NULL);
-		MFREE(message, char, msglen);
+			exceptions_throw_linkageerror(msg, NULL);
+
+		/* ATTENTION: We probably need msg for
+		   exceptions_throw_linkageerror. */
+
+		MFREE(msg, char, msglen);
+
 		return resolveFailed; /* exception */
 	}
 
@@ -926,6 +991,32 @@ classinfo * resolve_class_eager(unresolved_class *ref)
 }
 #endif /* ENABLE_VERIFIER */
 
+/* resolve_class_eager_no_access_check *****************************************
+ 
+   Resolve an unresolved class reference eagerly. The class is also linked.
+   Access rights are _not_ checked.
+  
+   IN:
+       ref..............struct containing the reference
+   
+   RETURN VALUE:
+       classinfo * to the class, or
+	   NULL if an exception has been thrown
+   
+*******************************************************************************/
+
+#ifdef ENABLE_VERIFIER
+classinfo * resolve_class_eager_no_access_check(unresolved_class *ref)
+{
+	classinfo *c;
+
+	if (!resolve_class(ref, resolveEager, false, &c))
+		return NULL;
+
+	return c;
+}
+#endif /* ENABLE_VERIFIER */
+
 /******************************************************************************/
 /* FIELD RESOLUTION                                                           */
 /******************************************************************************/
@@ -961,10 +1052,13 @@ resolve_result_t resolve_field_verifier_checks(methodinfo *refmethod,
 											   bool isstatic,
 											   bool isput)
 {
-	classinfo *declarer;
-	classinfo *referer;
-	resolve_result_t result;
+	classinfo         *declarer;
+	classinfo         *referer;
+	resolve_result_t   result;
 	constant_classref *fieldtyperef;
+	char              *msg;
+	s4                 msglen;
+	utf               *u;
 
 	assert(refmethod);
 	assert(fieldref);
@@ -990,30 +1084,39 @@ resolve_result_t resolve_field_verifier_checks(methodinfo *refmethod,
 
 	if (((fi->flags & ACC_STATIC) != 0) != isstatic) {
 		/* a static field is accessed via an instance, or vice versa */
-		*exceptionptr =
-			new_exception_message(string_java_lang_IncompatibleClassChangeError,
-				(fi->flags & ACC_STATIC) ? "static field accessed via instance"
-				                         : "instance field  accessed without instance");
+		exceptions_throw_incompatibleclasschangeerror(declarer,
+													  (fi->flags & ACC_STATIC)
+													  ? "static field accessed via instance"
+													  : "instance field  accessed without instance");
+
 		return resolveFailed;
 	}
 
 	/* check access rights */
 
 	if (!access_is_accessible_member(referer,declarer,fi->flags)) {
-		int msglen;
-		char *message;
+		msglen =
+			utf_bytes(declarer->name) +
+			utf_bytes(fi->name) +
+			utf_bytes(referer->name) +
+			100;
+
+		msg = MNEW(char, msglen);
+
+		strcpy(msg, "field is not accessible (");
+		utf_cat_classname(msg, declarer->name);
+		strcat(msg, ".");
+		utf_cat(msg, fi->name);
+		strcat(msg, " from ");
+		utf_cat_classname(msg, referer->name);
+		strcat(msg, ")");
+
+		u = utf_new_char(msg);
+
+		MFREE(msg, char, msglen);
+
+		exceptions_throw_illegalaccessexception(u);
 
-		msglen = utf_bytes(declarer->name) + utf_bytes(fi->name) + utf_bytes(referer->name) + 100;
-		message = MNEW(char, msglen);
-		strcpy(message, "field is not accessible (");
-		utf_cat_classname(message, declarer->name);
-		strcat(message, ".");
-		utf_cat(message, fi->name);
-		strcat(message, " from ");
-		utf_cat_classname(message, referer->name);
-		strcat(message, ")");
-		*exceptionptr = new_exception_message(string_java_lang_IllegalAccessException, message);
-		MFREE(message,char,msglen);
 		return resolveFailed; /* exception */
 	}
 
@@ -1177,7 +1280,7 @@ resolve_result_t resolve_field_lazy(methodinfo *refmethod,
 		/* this error must not be reported now. (It will be reported   */
 		/* if eager resolving of this field is ever tried.)           */
 
-		*exceptionptr = NULL;
+		exceptions_clear_exception();
 		return resolveDeferred; /* be lazy */
 	}
 
@@ -1276,7 +1379,7 @@ bool resolve_field(unresolved_field *ref,
 			/* this error must not be reported now. (It will be reported   */
 			/* if eager resolving of this field is ever tried.)           */
 
-			*exceptionptr = NULL;
+			exceptions_clear_exception();
 			return true; /* be lazy */
 		}
 
@@ -1441,7 +1544,9 @@ methodinfo * resolve_method_invokespecial_lookup(methodinfo *refmethod,
 
 			if (mi == NULL) {
 				/* the spec calls for an AbstractMethodError in this case */
+
 				exceptions_throw_abstractmethoderror();
+
 				return NULL;
 			}
 		}
@@ -1476,6 +1581,9 @@ resolve_result_t resolve_method_verifier_checks(methodinfo *refmethod,
 {
 	classinfo *declarer;
 	classinfo *referer;
+	char      *msg;
+	s4         msglen;
+	utf       *u;
 
 	assert(refmethod);
 	assert(methodref);
@@ -1498,33 +1606,43 @@ resolve_result_t resolve_method_verifier_checks(methodinfo *refmethod,
 
 	if (((mi->flags & ACC_STATIC) != 0) != (invokestatic != false)) {
 		/* a static method is accessed via an instance, or vice versa */
-		*exceptionptr =
-			new_exception_message(string_java_lang_IncompatibleClassChangeError,
-				(mi->flags & ACC_STATIC) ? "static method called via instance"
-				                         : "instance method called without instance");
+		exceptions_throw_incompatibleclasschangeerror(declarer,
+													  (mi->flags & ACC_STATIC)
+													  ? "static method called via instance"
+													  : "instance method called without instance");
+
 		return resolveFailed;
 	}
 
 	/* check access rights */
 
 	if (!access_is_accessible_member(referer,declarer,mi->flags)) {
-		int msglen;
-		char *message;
-
 		/* XXX clean this up. this should be in exceptions.c */
-		msglen = utf_bytes(declarer->name) + utf_bytes(mi->name) +
-			utf_bytes(mi->descriptor) + utf_bytes(referer->name) + 100;
-		message = MNEW(char, msglen);
-		strcpy(message, "method is not accessible (");
-		utf_cat_classname(message, declarer->name);
-		strcat(message, ".");
-		utf_cat(message, mi->name);
-		utf_cat(message, mi->descriptor);
-		strcat(message," from ");
-		utf_cat_classname(message, referer->name);
-		strcat(message,")");
-		*exceptionptr = new_exception_message(string_java_lang_IllegalAccessException, message);
-		MFREE(message, char, msglen);
+
+		msglen =
+			utf_bytes(declarer->name) +
+			utf_bytes(mi->name) +
+			utf_bytes(mi->descriptor) +
+			utf_bytes(referer->name) +
+			100;
+
+		msg = MNEW(char, msglen);
+
+		strcpy(msg, "method is not accessible (");
+		utf_cat_classname(msg, declarer->name);
+		strcat(msg, ".");
+		utf_cat(msg, mi->name);
+		utf_cat(msg, mi->descriptor);
+		strcat(msg, " from ");
+		utf_cat_classname(msg, referer->name);
+		strcat(msg, ")");
+
+		u = utf_new_char(msg);
+
+		MFREE(msg, char, msglen);
+
+		exceptions_throw_illegalaccessexception(u);
+
 		return resolveFailed; /* exception */
 	}
 
@@ -1882,7 +2000,7 @@ resolve_result_t resolve_method_lazy(methodinfo *refmethod,
 		/* this error must not be reported now. (It will be reported   */
 		/* if eager resolving of this method is ever tried.)           */
 
-		*exceptionptr = NULL;
+		exceptions_clear_exception();
 		return resolveDeferred; /* be lazy */
 	}
 
@@ -2001,7 +2119,7 @@ bool resolve_method(unresolved_method *ref, resolve_mode_t mode, methodinfo **re
 			/* this error must not be reported now. (It will be reported   */
 			/* if eager resolving of this method is ever tried.)           */
 
-			*exceptionptr = NULL;
+			exceptions_clear_exception();
 			return true; /* be lazy */
 		}
 
@@ -2975,6 +3093,7 @@ void unresolved_method_debug_dump(unresolved_method *ref,FILE *file)
 }
 #endif /* !defined(NDEBUG) */
 
+
 /*
  * These are local overrides for various environment variables in Emacs.
  * Please do not remove this and leave it at the end of the file, where
@@ -2988,4 +3107,3 @@ void unresolved_method_debug_dump(unresolved_method *ref,FILE *file)
  * End:
  * vim:noexpandtab:sw=4:ts=4:
  */
-
