@@ -1,6 +1,6 @@
 /* src/vm/vm.c - VM startup and shutdown functions
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
    C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
    E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
    J. Wenninger, Institut f. Computersprachen - TU Wien
@@ -22,13 +22,7 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes: Martin Platter
-
-   $Id: vm.c 4357 2006-01-22 23:33:38Z twisti $
+   $Id: vm.c 8008 2007-06-05 07:44:38Z twisti $
 
 */
 
@@ -46,36 +40,49 @@
 
 #include "vm/types.h"
 
-#include "mm/boehm.h"
+#include "mm/gc-common.h"
 #include "mm/memory.h"
+
 #include "native/jni.h"
 #include "native/native.h"
+#include "native/include/java_lang_String.h" /* required by java_lang_Class.h */
+#include "native/include/java_lang_Class.h"
 
-#if defined(ENABLE_THREADS)
-# include "threads/native/threads.h"
-#endif
+#include "threads/threads-common.h"
 
-#include "vm/classcache.h"
+#include "toolbox/logging.h"
+
+#include "vm/builtin.h"
 #include "vm/exceptions.h"
 #include "vm/finalizer.h"
 #include "vm/global.h"
 #include "vm/initialize.h"
-#include "vm/options.h"
 #include "vm/properties.h"
 #include "vm/signallocal.h"
 #include "vm/stringlocal.h"
-#include "vm/suck.h"
 #include "vm/vm.h"
+
 #include "vm/jit/jit.h"
+#include "vm/jit/md.h"
 #include "vm/jit/asmpart.h"
 
-#include "vm/jit/recompile.h"
+#if defined(ENABLE_PROFILING)
+# include "vm/jit/optimizing/profile.h"
+#endif
 
-#include "vm/jit/profile/profile.h"
-#include "vm/rt-timing.h"
+#include "vm/jit/optimizing/recompile.h"
+
+#include "vmcore/classcache.h"
+#include "vmcore/options.h"
+#include "vmcore/statistics.h"
+#include "vmcore/suck.h"
 
 #if defined(ENABLE_JVMTI)
-#include "native/jvmti/cacaodbg.h"
+# include "native/jvmti/cacaodbg.h"
+#endif
+
+#if defined(ENABLE_VMLOG)
+#include <vmlog_cacao.h>
 #endif
 
 
@@ -143,6 +150,7 @@ enum {
 
 	OPT_HELP,
 	OPT_X,
+	OPT_XX,
 
 	OPT_ESA,
 	OPT_DSA,
@@ -170,7 +178,6 @@ enum {
 
 	OPT_VERBOSE1,
 	OPT_NOIEEE,
-	OPT_SOFTNULL,
 
 #if defined(ENABLE_STATISTICS)
 	OPT_TIME,
@@ -213,7 +220,18 @@ enum {
 
 #if defined(ENABLE_INLINING)
 	OPT_INLINING,
+#if !defined(NDEBUG)
+	OPT_INLINE_LOG,
 #endif
+#if defined(ENABLE_INLINING_DEBUG)
+	OPT_INLINE_DEBUG_ALL,
+	OPT_INLINE_DEBUG_END,
+	OPT_INLINE_DEBUG_MIN,
+	OPT_INLINE_DEBUG_MAX,
+	OPT_INLINE_REPLACE_VERBOSE,
+	OPT_INLINE_REPLACE_VERBOSE2,
+#endif /* defined(ENABLE_INLINING_DEBUG) */
+#endif /* defined(ENABLE_INLINING) */
 
 #if defined(ENABLE_INTRP)
 	/* interpreter options */
@@ -235,6 +253,12 @@ enum {
 	OPT_AGENTPATH,
 #endif
 
+#if defined(ENABLE_DEBUG_FILTER)
+	OPT_FILTER_VERBOSECALL_INCLUDE,
+	OPT_FILTER_VERBOSECALL_EXCLUDE,
+	OPT_FILTER_SHOW_METHOD,
+#endif
+
 	DUMMY
 };
 
@@ -262,6 +286,7 @@ opt_struct opts[] = {
 	{ "help",              false, OPT_HELP },
 	{ "?",                 false, OPT_HELP },
 	{ "X",                 false, OPT_X },
+	{ "XX",                false, OPT_XX },
 
 	{ "esa",                     false, OPT_ESA },
 	{ "enablesystemassertions",  false, OPT_ESA },
@@ -281,7 +306,6 @@ opt_struct opts[] = {
 #if defined(__ALPHA__)
 	{ "noieee",            false, OPT_NOIEEE },
 #endif
-	{ "softnull",          false, OPT_SOFTNULL },
 #if defined(ENABLE_STATISTICS)
 	{ "time",              false, OPT_TIME },
 	{ "stat",              false, OPT_STAT },
@@ -349,11 +373,24 @@ opt_struct opts[] = {
 	{ "Xprof",             false, OPT_PROF },
 #endif
 
-	/* keep these at the end of the list */
+	/* inlining options */
 
 #if defined(ENABLE_INLINING)
-	{ "i",                 true,  OPT_INLINING },
+#if defined(ENABLE_INLINING_DEBUG)
+	{ "ia",                false, OPT_INLINE_DEBUG_ALL },
+	{ "ii",                true,  OPT_INLINE_DEBUG_MIN },
+	{ "im",                true,  OPT_INLINE_DEBUG_MAX },
+	{ "ie",                true,  OPT_INLINE_DEBUG_END },
+	{ "ir",                false, OPT_INLINE_REPLACE_VERBOSE },
+	{ "iR",                false, OPT_INLINE_REPLACE_VERBOSE2 },
+#endif /* defined(ENABLE_INLINING_DEBUG) */
+#if !defined(NDEBUG)
+	{ "il",                false, OPT_INLINE_LOG },
 #endif
+	{ "i",                 false, OPT_INLINING },
+#endif /* defined(ENABLE_INLINING) */
+
+	/* keep these at the end of the list */
 
 #if !defined(NDEBUG)
 	{ "m",                 true,  OPT_METHOD },
@@ -362,6 +399,12 @@ opt_struct opts[] = {
 	{ "s",                 true,  OPT_SHOW },
 	{ "debug-color",      false,  OPT_DEBUGCOLOR },
 
+#if defined(ENABLE_DEBUG_FILTER)
+	{ "XXfi",              true,  OPT_FILTER_VERBOSECALL_INCLUDE },
+	{ "XXfx",              true,  OPT_FILTER_VERBOSECALL_EXCLUDE },
+	{ "XXfm",              true,  OPT_FILTER_SHOW_METHOD },
+#endif
+
 	{ NULL,                false, 0 }
 };
 
@@ -379,7 +422,7 @@ void usage(void)
 	puts("   or  cacao [-options] -jar jarfile [arguments]");
 	puts("               (to run a standalone jar file)\n");
 
-	puts("Java options:");
+	puts("where options include:");
 	puts("    -d32                     use 32-bit data model if available");
 	puts("    -d64                     use 64-bit data model if available");
 	puts("    -client                  compatibility (currently ignored)");
@@ -396,11 +439,11 @@ void usage(void)
 	puts("    -showversion             print product version and continue");
 	puts("    -help, -?                print this help message");
 	puts("    -X                       print help on non-standard Java options");
+	puts("    -XX                      print help on CACAO options");
 	puts("    -esa | -enablesystemassertions");
 	puts("                             enable system assertions");
 	puts("    -dsa | -disablesystemassertions");
 	puts("                             disable system assertions");
-	puts("");
 
 #ifdef ENABLE_JVMTI
 	puts("    -agentlib:<agent-lib-name>=<options>  library to load containg JVMTI agent");
@@ -408,67 +451,6 @@ void usage(void)
 	puts("    -agentpath:<path-to-agent>=<options>  path to library containg JVMTI agent");
 #endif
 
-	puts("CACAO options:");
-	puts("    -v                       write state-information");
-	puts("    -verbose[:call|exception|jit]");
-	puts("                             enable specific verbose output");
-	puts("    -debug-color             colored output for ANSI terms");
-#ifdef TYPECHECK_VERBOSE
-	puts("    -verbosetc               write debug messages while typechecking");
-#endif
-#if defined(__ALPHA__)
-	puts("    -noieee                  don't use ieee compliant arithmetic");
-#endif
-#if defined(ENABLE_VERIFIER)
-	puts("    -noverify                don't verify classfiles");
-#endif
-	puts("    -softnull                use software nullpointer check");
-#if defined(ENABLE_STATISTICS)
-	puts("    -time                    measure the runtime");
-	puts("    -stat                    detailed compiler statistics");
-#endif
-	puts("    -log logfile             specify a name for the logfile");
-	puts("    -c(heck)b(ounds)         don't check array bounds");
-	puts("            s(ync)           don't check for synchronization");
-#if defined(ENABLE_LOOP)
-	puts("    -oloop                   optimize array accesses in loops");
-#endif
-	puts("    -l                       don't start the class after loading");
-	puts("    -eager                   perform eager class loading and linking");
-#if !defined(NDEBUG)
-	puts("    -all                     compile all methods, no execution");
-	puts("    -m                       compile only a specific method");
-	puts("    -sig                     specify signature for a specific method");
-#endif
-
-	puts("    -s(how)...               show...");
-	puts("           c(onstants)       the constant pool");
-	puts("           m(ethods)         class fields and methods");
-	puts("           u(tf)             the utf - hash");
-	puts("           i(ntermediate)    intermediate representation");
-#if defined(ENABLE_DISASSEMBLER)
-	puts("           a(ssembler)       disassembled listing");
-	puts("           e(xceptionstubs)  disassembled exception stubs (only with -sa)");
-	puts("           n(ative)          disassembled native stubs");
-#endif
-	puts("           d(atasegment)     data segment listing");
-#if defined(ENABLE_INLINING)
-	puts("    -i     n(line)           activate inlining");
-	puts("           v(irtual)         inline virtual methods (uses/turns rt option on)");
-	puts("           e(exception)      inline methods with exceptions");
-	puts("           p(aramopt)        optimize argument renaming");
-	puts("           o(utsiders)       inline methods of foreign classes");
-#endif /* defined(ENABLE_INLINING) */
-#if defined(ENABLE_IFCONV)
-	puts("    -ifconv                  use if-conversion");
-#endif
-#if defined(ENABLE_LSRA)
-	puts("    -lsra                    use linear scan register allocation");
-#endif
-#if defined(ENABLE_SSA)
-	puts("    -lsra                    use linear scan register allocation (with SSA)");
-#endif
-
 	/* exit with error code */
 
 	exit(1);
@@ -514,6 +496,89 @@ static void Xusage(void)
 }   
 
 
+static void XXusage(void)
+{
+	puts("    -v                       write state-information");
+#if !defined(NDEBUG)
+	puts("    -verbose[:call|exception|jit|memory|threads]");
+	puts("                             enable specific verbose output");
+	puts("    -debug-color             colored output for ANSI terms");
+#endif
+#ifdef TYPECHECK_VERBOSE
+	puts("    -verbosetc               write debug messages while typechecking");
+#endif
+#if defined(__ALPHA__)
+	puts("    -noieee                  don't use ieee compliant arithmetic");
+#endif
+#if defined(ENABLE_VERIFIER)
+	puts("    -noverify                don't verify classfiles");
+#endif
+#if defined(ENABLE_STATISTICS)
+	puts("    -time                    measure the runtime");
+	puts("    -stat                    detailed compiler statistics");
+#endif
+	puts("    -log logfile             specify a name for the logfile");
+	puts("    -c(heck)b(ounds)         don't check array bounds");
+	puts("            s(ync)           don't check for synchronization");
+#if defined(ENABLE_LOOP)
+	puts("    -oloop                   optimize array accesses in loops");
+#endif
+	puts("    -l                       don't start the class after loading");
+	puts("    -eager                   perform eager class loading and linking");
+#if !defined(NDEBUG)
+	puts("    -all                     compile all methods, no execution");
+	puts("    -m                       compile only a specific method");
+	puts("    -sig                     specify signature for a specific method");
+#endif
+
+	puts("    -s...                    show...");
+	puts("      (c)onstants            the constant pool");
+	puts("      (m)ethods              class fields and methods");
+	puts("      (u)tf                  the utf - hash");
+	puts("      (i)ntermediate         intermediate representation");
+#if defined(ENABLE_DISASSEMBLER)
+	puts("      (a)ssembler            disassembled listing");
+	puts("      n(o)ps                 show NOPs in disassembler output");
+	puts("      (e)xceptionstubs       disassembled exception stubs (only with -sa)");
+	puts("      (n)ative               disassembled native stubs");
+#endif
+	puts("      (d)atasegment          data segment listing");
+
+#if defined(ENABLE_INLINING)
+	puts("    -i                       activate inlining");
+#if !defined(NDEBUG)
+	puts("    -il                      log inlining");
+#endif
+#if defined(ENABLE_INLINING_DEBUG)
+	puts("    -ia                      use inlining for all methods");
+	puts("    -ii <size>               set minimum size for inlined result");
+	puts("    -im <size>               set maximum size for inlined result");
+	puts("    -ie <number>             stop inlining after the given number of roots");
+	puts("    -ir                      log on-stack replacement");
+	puts("    -iR                      log on-stack replacement, more verbose");
+#endif /* defined(ENABLE_INLINING_DEBUG) */
+#endif /* defined(ENABLE_INLINING) */
+
+#if defined(ENABLE_IFCONV)
+	puts("    -ifconv                  use if-conversion");
+#endif
+#if defined(ENABLE_LSRA)
+	puts("    -lsra                    use linear scan register allocation");
+#endif
+#if defined(ENABLE_SSA)
+	puts("    -lsra                    use linear scan register allocation (with SSA)");
+#endif
+#if defined(ENABLE_DEBUG_FILTER)
+	puts("    -XXfi <regex>            begin of dynamic scope for verbosecall filter");
+	puts("    -XXfx <regex>            end of dynamic scope for verbosecall filter");
+	puts("    -XXfm <regex>            filter for show options");
+#endif
+	/* exit with error code */
+
+	exit(1);
+}
+
+
 /* version *********************************************************************
 
    Only prints cacao version information.
@@ -525,7 +590,7 @@ static void version(bool opt_exit)
 	puts("java version \""JAVA_VERSION"\"");
 	puts("CACAO version "VERSION"");
 
-	puts("Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,");
+	puts("Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,");
 	puts("C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,");
 	puts("E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,");
 	puts("J. Wenninger, Institut f. Computersprachen - TU Wien\n");
@@ -553,15 +618,21 @@ static void version(bool opt_exit)
 	printf("  maximum heap size              : %d\n", HEAP_MAXSIZE);
 	printf("  initial heap size              : %d\n", HEAP_STARTSIZE);
 	printf("  stack size                     : %d\n", STACK_SIZE);
-	puts("  java.boot.class.path           : "CACAO_VM_ZIP":"CLASSPATH_GLIBJ_ZIP"");
+#if defined(WITH_CLASSPATH_GNU)
+	puts("  java.boot.class.path           : "CACAO_VM_ZIP":"CLASSPATH_CLASSES"");
+#else
+	puts("  java.boot.class.path           : "CLASSPATH_CLASSES"");
+#endif
 	puts("  gnu.classpath.boot.library.path: "CLASSPATH_LIBDIR"/classpath\n");
 
 	puts("Runtime variables:\n");
 	printf("  maximum heap size              : %d\n", opt_heapmaxsize);
 	printf("  initial heap size              : %d\n", opt_heapstartsize);
 	printf("  stack size                     : %d\n", opt_stacksize);
+	printf("  libjvm.so                      : %s\n", cacao_libjvm);
 	printf("  java.boot.class.path           : %s\n", _Jv_bootclasspath);
 	printf("  gnu.classpath.boot.library.path: %s\n", classpath_libdir);
+	printf("  java.class.path                : %s\n", _Jv_classpath);
 
 	/* exit normally, if requested */
 
@@ -596,9 +667,79 @@ static void  vm_compile_method(void);
 #endif
 
 
+/* vm_createjvm ****************************************************************
+
+   Implementation for JNI_CreateJavaVM.
+
+*******************************************************************************/
+
+bool vm_createjvm(JavaVM **p_vm, void **p_env, void *vm_args)
+{
+	JavaVMInitArgs *_vm_args;
+	_Jv_JNIEnv     *env;
+	_Jv_JavaVM     *vm;
+
+	/* get the arguments for the new JVM */
+
+	_vm_args = (JavaVMInitArgs *) vm_args;
+
+	/* get the VM and Env tables (must be set before vm_create) */
+
+	env = NEW(_Jv_JNIEnv);
+
+#if defined(ENABLE_JNI)
+	env->env = &_Jv_JNINativeInterface;
+#endif
+
+	/* XXX Set the global variable.  Maybe we should do that differently. */
+
+	_Jv_env = env;
+
+	/* create and fill a JavaVM structure */
+
+	vm = NEW(_Jv_JavaVM);
+
+#if defined(ENABLE_JNI)
+	vm->functions = &_Jv_JNIInvokeInterface;
+#endif
+
+	/* XXX Set the global variable.  Maybe we should do that differently. */
+	/* XXX JVMTI Agents needs a JavaVM  */
+
+	_Jv_jvm = vm;
+
+	/* actually create the JVM */
+
+	if (!vm_create(_vm_args))
+		goto error;
+
+#if defined(ENABLE_JNI)
+	/* setup the local ref table (must be created after vm_create) */
+
+	if (!jni_init_localref_table())
+		goto error;
+#endif
+
+	/* now return the values */
+
+	*p_vm  = (JavaVM *) vm;
+	*p_env = (void *) env;
+
+	return true;
+
+ error:
+	/* release allocated memory */
+
+	FREE(env, _Jv_JNIEnv);
+	FREE(vm, _Jv_JavaVM);
+
+	return false;
+}
+
+
 /* vm_create *******************************************************************
 
-   Creates a JVM.  Called by JNI_CreateJavaVM.
+   Creates a JVM.  Called by vm_createjvm.
 
 *******************************************************************************/
 
@@ -618,6 +759,10 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	jdwp = agentbypath = false;
 #endif
 
+#if defined(ENABLE_VMLOG)
+	vmlog_cacao_init(vm_args);
+#endif
+
 	/* check the JNI version requested */
 
 	switch (vm_args->version) {
@@ -641,16 +786,19 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	if (opt_verbose)
 		log_text("CACAO started -------------------------------------------------------");
 
+	/* We need to check if the actual size of a java.lang.Class object
+	   is smaller or equal than the assumption made in
+	   src/vmcore/class.h. */
+
+	if (sizeof(java_lang_Class) > sizeof(dummy_java_lang_Class))
+		vm_abort("vm_create: java_lang_Class structure is bigger than classinfo.object (%d > %d)", sizeof(java_lang_Class), sizeof(dummy_java_lang_Class));
+
 	/* set the VM starttime */
 
 	_Jv_jvm->starttime = builtin_currenttimemillis();
 
 	/* get stuff from the environment *****************************************/
 
-#if defined(DISABLE_GC)
-	nogc_init(HEAP_MAXSIZE, HEAP_STARTSIZE);
-#endif
-
 #if defined(WITH_JRE_LAYOUT)
 	/* SUN also uses a buffer of 4096-bytes (strace is your friend). */
 
@@ -702,30 +850,38 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	else {
 #if defined(WITH_JRE_LAYOUT)
 		len =
+# if defined(WITH_CLASSPATH_GNU)
 			strlen(cacao_prefix) +
 			strlen("/share/cacao/vm.zip") +
 			strlen(":") +
+# endif
 			strlen(cacao_prefix) +
 			strlen("/share/classpath/glibj.zip") +
 			strlen("0");
 
 		_Jv_bootclasspath = MNEW(char, len);
+# if defined(WITH_CLASSPATH_GNU)
 		strcat(_Jv_bootclasspath, cacao_prefix);
 		strcat(_Jv_bootclasspath, "/share/cacao/vm.zip");
 		strcat(_Jv_bootclasspath, ":");
+# endif
 		strcat(_Jv_bootclasspath, cacao_prefix);
 		strcat(_Jv_bootclasspath, "/share/classpath/glibj.zip");
 #else
 		len =
+# if defined(WITH_CLASSPATH_GNU)
 			strlen(CACAO_VM_ZIP) +
 			strlen(":") +
-			strlen(CLASSPATH_GLIBJ_ZIP) +
+# endif
+			strlen(CLASSPATH_CLASSES) +
 			strlen("0");
 
 		_Jv_bootclasspath = MNEW(char, len);
+# if defined(WITH_CLASSPATH_GNU)
 		strcat(_Jv_bootclasspath, CACAO_VM_ZIP);
 		strcat(_Jv_bootclasspath, ":");
-		strcat(_Jv_bootclasspath, CLASSPATH_GLIBJ_ZIP);
+# endif
+		strcat(_Jv_bootclasspath, CLASSPATH_CLASSES);
 #endif
 	}
 
@@ -754,7 +910,6 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	opt_version       = false;
 	opt_exit          = false;
 
-	checknull         = false;
 	opt_noieee        = false;
 
 	opt_heapmaxsize   = HEAP_MAXSIZE;
@@ -767,15 +922,10 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	jvmti = false;
 #endif
 
-	/* initialize properties before commandline handling */
+	/* initialize and fill properties before command-line handling */
 
 	if (!properties_init())
-		throw_cacao_exception_exit(string_java_lang_InternalError,
-								   "Unable to init properties");
-
-	/* add some default properties */
-
-	properties_add("java.endorsed.dirs", ""CACAO_PREFIX"/jre/lib/endorsed");
+		vm_abort("properties_init failed");
 
 	/* iterate over all passed options */
 
@@ -847,7 +997,7 @@ bool vm_create(JavaVMInitArgs *vm_args)
 
 			_Jv_bootclasspath = MREALLOC(_Jv_bootclasspath,
 										 char,
-										 len,
+										 len + strlen("0"),
 										 len + strlen(":") +
 										 strlen(opt_arg) + strlen("0"));
 
@@ -955,18 +1105,22 @@ bool vm_create(JavaVMInitArgs *vm_args)
 			break;
 
 		case OPT_VERBOSE:
-			if (strcmp("class", opt_arg) == 0)
+			if (strcmp("class", opt_arg) == 0) {
 				opt_verboseclass = true;
-
-			else if (strcmp("gc", opt_arg) == 0)
+			}
+			else if (strcmp("gc", opt_arg) == 0) {
 				opt_verbosegc = true;
-
-			else if (strcmp("jni", opt_arg) == 0)
+			}
+			else if (strcmp("jni", opt_arg) == 0) {
 				opt_verbosejni = true;
-
-			else if (strcmp("call", opt_arg) == 0)
+			}
+#if !defined(NDEBUG)
+			else if (strcmp("call", opt_arg) == 0) {
 				opt_verbosecall = true;
-
+			}
+			else if (strcmp("exception", opt_arg) == 0) {
+				opt_verboseexception = true;
+			}
 			else if (strcmp("jit", opt_arg) == 0) {
 				opt_verbose = true;
 				loadverbose = true;
@@ -974,9 +1128,25 @@ bool vm_create(JavaVMInitArgs *vm_args)
 				initverbose = true;
 				compileverbose = true;
 			}
-			else if (strcmp("exception", opt_arg) == 0)
-				opt_verboseexception = true;
+			else if (strcmp("memory", opt_arg) == 0) {
+				opt_verbosememory = true;
+
+# if defined(ENABLE_STATISTICS)
+				/* we also need statistics */
+
+				opt_stat = true;
+# endif
+			}
+			else if (strcmp("threads", opt_arg) == 0) {
+				opt_verbosethreads = true;
+			}
+#endif
+			else {
+				printf("Unknown -verbose option: %s\n", opt_arg);
+				usage();
+			}
 			break;
+
 		case OPT_DEBUGCOLOR:
 			opt_debugcolor = true;
 			break;
@@ -1010,10 +1180,6 @@ bool vm_create(JavaVMInitArgs *vm_args)
 			break;
 #endif
 
-		case OPT_SOFTNULL:
-			checknull = true;
-			break;
-
 #if defined(ENABLE_STATISTICS)
 		case OPT_TIME:
 			opt_getcompilingtime = true;
@@ -1097,6 +1263,10 @@ bool vm_create(JavaVMInitArgs *vm_args)
 					compileverbose = true;
 					break;
 
+				case 'o':
+					opt_shownops = true;
+					break;
+
 				case 'e':
 					opt_showexceptionstubs = true;
 					break;
@@ -1123,30 +1293,34 @@ bool vm_create(JavaVMInitArgs *vm_args)
 #endif
 
 #if defined(ENABLE_INLINING)
+#if defined(ENABLE_INLINING_DEBUG)
+		case OPT_INLINE_DEBUG_ALL:
+			opt_inline_debug_all = true;
+			break;
+		case OPT_INLINE_DEBUG_END:
+			opt_inline_debug_end_counter = atoi(opt_arg);
+			break;
+		case OPT_INLINE_DEBUG_MIN:
+			opt_inline_debug_min_size = atoi(opt_arg);
+			break;
+		case OPT_INLINE_DEBUG_MAX:
+			opt_inline_debug_max_size = atoi(opt_arg);
+			break;
+		case OPT_INLINE_REPLACE_VERBOSE:
+			opt_replace_verbose = 1;
+			break;
+		case OPT_INLINE_REPLACE_VERBOSE2:
+			opt_replace_verbose = 2;
+			break;
+#endif /* defined(ENABLE_INLINING_DEBUG) */
+#if !defined(NDEBUG)
+		case OPT_INLINE_LOG:
+			opt_inline_debug_log = true;
+			break;
+#endif /* !defined(NDEBUG) */
+
 		case OPT_INLINING:
-			for (i = 0; i < strlen(opt_arg); i++) {		
-				switch (opt_arg[i]) {
-				case 'n':
-					/* define in options.h; Used in main.c, jit.c
-					   & inline.c inlining is currently
-					   deactivated */
-					break;
-				case 'v':
-					inlinevirtuals = true;
-					break;
-				case 'e':
-					inlineexceptions = true;
-					break;
-				case 'p':
-					inlineparamopt = true;
-					break;
-				case 'o':
-					inlineoutsiders = true;
-					break;
-				default:
-					usage();
-				}
-			}
+			opt_inlining = true;
 			break;
 #endif /* defined(ENABLE_INLINING) */
 
@@ -1170,6 +1344,10 @@ bool vm_create(JavaVMInitArgs *vm_args)
 			Xusage();
 			break;
 
+		case OPT_XX:
+			XXusage();
+			break;
+
 		case OPT_ESA:
 			_Jv_jvm->Java_java_lang_VMClassLoader_defaultAssertionStatus = true;
 			break;
@@ -1247,6 +1425,20 @@ bool vm_create(JavaVMInitArgs *vm_args)
 			break;
 #endif
 
+#if defined(ENABLE_DEBUG_FILTER)
+		case OPT_FILTER_VERBOSECALL_INCLUDE:
+			opt_filter_verbosecall_include = opt_arg;
+			break;
+
+		case OPT_FILTER_VERBOSECALL_EXCLUDE:
+			opt_filter_verbosecall_exclude = opt_arg;
+			break;
+
+		case OPT_FILTER_SHOW_METHOD:
+			opt_filter_show_method = opt_arg;
+			break;
+
+#endif
 		default:
 			printf("Unknown option: %s\n",
 				   vm_args->options[opt_index].optionString);
@@ -1254,14 +1446,6 @@ bool vm_create(JavaVMInitArgs *vm_args)
 		}
 	}
 
-
-	/* Now we have all options handled and we can print the version
-	   information. */
-
-	if (opt_version)
-		version(opt_exit);
-
-
 	/* get the main class *****************************************************/
 
 	if (opt_index < vm_args->nOptions) {
@@ -1299,10 +1483,8 @@ bool vm_create(JavaVMInitArgs *vm_args)
 
 		jvmti_set_phase(JVMTI_PHASE_PRIMORDIAL);
 	}
-
 #endif
 
-
 	/* initialize this JVM ****************************************************/
 
 	vm_initializing = true;
@@ -1318,6 +1500,11 @@ bool vm_create(JavaVMInitArgs *vm_args)
 
 	gc_init(opt_heapmaxsize, opt_heapstartsize);
 
+	/* install architecture dependent signal handlers */
+
+	if (!signal_init())
+		vm_abort("vm_create: signal_init failed");
+
 #if defined(ENABLE_INTRP)
 	/* Allocate main thread stack on the Java heap. */
 
@@ -1327,48 +1514,62 @@ bool vm_create(JavaVMInitArgs *vm_args)
 	}
 #endif
 
-	/* initialize the string hashtable stuff: lock (must be done
-	   _after_ threads_preinit) */
+	/* AFTER: threads_preinit */
 
 	if (!string_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: string_init failed");
 
-	/* initialize the utf8 hashtable stuff: lock, often used utf8
-	   strings (must be done _after_ threads_preinit) */
+	/* AFTER: threads_preinit */
 
 	if (!utf8_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: utf8_init failed");
+
+	/* AFTER: thread_preinit */
+
+	if (!suck_init())
+		vm_abort("vm_create: suck_init failed");
+
+	suck_add_from_property("java.endorsed.dirs");
+
+	/* Now we have all options handled and we can print the version
+	   information.
+
+	   AFTER: suck_add_from_property("java.endorsed.dirs"); */
+
+	if (opt_version)
+		version(opt_exit);
+
+	/* AFTER: utf8_init */
+
+	suck_add(_Jv_bootclasspath);
+
+	/* Now re-set some of the properties that may have changed. This
+	   must be done after _all_ environment variables have been
+	   processes (e.g. -jar handling).
+
+	   AFTER: suck_add_from_property, since it may change the
+	   _Jv_bootclasspath pointer. */
+
+	if (!properties_postinit())
+		vm_abort("vm_create: properties_postinit failed");
 
 	/* initialize the classcache hashtable stuff: lock, hashtable
 	   (must be done _after_ threads_preinit) */
 
 	if (!classcache_init())
-		throw_main_exception_exit();
-
-	/* initialize the loader with bootclasspath (must be done _after_
-	   thread_preinit) */
-
-	if (!suck_init())
-		throw_main_exception_exit();
-
-	suck_add_from_property("java.endorsed.dirs");
-	suck_add(_Jv_bootclasspath);
+		vm_abort("vm_create: classcache_init failed");
 
 	/* initialize the memory subsystem (must be done _after_
 	   threads_preinit) */
 
 	if (!memory_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: memory_init failed");
 
 	/* initialize the finalizer stuff (must be done _after_
 	   threads_preinit) */
 
 	if (!finalizer_init())
-		throw_main_exception_exit();
-
-	/* install architecture dependent signal handler used for exceptions */
-
-	signal_init();
+		vm_abort("vm_create: finalizer_init failed");
 
 	/* initialize the codegen subsystems */
 
@@ -1395,70 +1596,83 @@ bool vm_create(JavaVMInitArgs *vm_args)
        classcache_init) */
 
 	if (!loader_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: loader_init failed");
 
 	if (!linker_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: linker_init failed");
+
+	/* Initialize the native subsystem. */
 
 	if (!native_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: native_init failed");
 
 	if (!exceptions_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: exceptions_init failed");
 
 	if (!builtin_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: builtin_init failed");
 
+#if defined(ENABLE_JNI)
 	/* Initialize the JNI subsystem (must be done _before_
 	   threads_init, as threads_init can call JNI methods
 	   (e.g. NewGlobalRef). */
 
 	if (!jni_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: jni_init failed");
+#endif
 
 #if defined(ENABLE_THREADS)
   	if (!threads_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: threads_init failed");
 #endif
 
-	/* That's important, otherwise we get into trouble, if the Runtime
-	   static initializer is called before (circular dependency. This
-	   is with classpath 0.09. Another important thing is, that this
-	   has to happen after initThreads!!! */
-
-	if (!initialize_class(class_java_lang_System))
-		throw_main_exception_exit();
-
 #if defined(ENABLE_PROFILING)
 	/* initialize profiling */
 
 	if (!profile_init())
-		throw_main_exception_exit();
+		vm_abort("vm_create: profile_init failed");
 #endif
 
 #if defined(ENABLE_THREADS)
 	/* initialize recompilation */
 
 	if (!recompile_init())
-		throw_main_exception_exit();
-		
+		vm_abort("vm_create: recompile_init failed");
+
+	/* start the signal handler thread */
+
+#if defined(__LINUX__)
+	/* XXX Remove for exact-GC. */
+	if (threads_pthreads_implementation_nptl)
+#endif
+		if (!signal_start_thread())
+			vm_abort("vm_create: signal_start_thread failed");
+
 	/* finally, start the finalizer thread */
 
 	if (!finalizer_start_thread())
-		throw_main_exception_exit();
+		vm_abort("vm_create: finalizer_start_thread failed");
+
+# if !defined(NDEBUG)
+	/* start the memory profiling thread */
+
+	if (opt_verbosememory)
+		if (!memory_start_thread())
+			vm_abort("vm_create: memory_start_thread failed");
+# endif
 
 	/* start the recompilation thread (must be done before the
 	   profiling thread) */
 
 	if (!recompile_start_thread())
-		throw_main_exception_exit();
+		vm_abort("vm_create: recompile_start_thread failed");
 
 # if defined(ENABLE_PROFILING)
 	/* start the profile sampling thread */
 
 /* 	if (opt_prof) */
 /* 		if (!profile_start_thread()) */
-/* 			throw_main_exception_exit(); */
+/* 			vm_abort("vm_create: profile_start_thread failed"); */
 # endif
 #endif
 
@@ -1491,15 +1705,15 @@ bool vm_create(JavaVMInitArgs *vm_args)
 
 void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 {
-	utf              *mainutf;
-	classinfo        *mainclass;
-	methodinfo       *m;
-	java_objectarray *oa; 
-	s4                oalength;
-	utf              *u;
-	java_lang_String *s;
-	s4                status;
-	s4                i;
+	utf               *mainutf;
+	classinfo         *mainclass;
+	methodinfo        *m;
+	java_objectarray  *oa; 
+	s4                 oalength;
+	utf               *u;
+	java_objectheader *s;
+	s4                 status;
+	s4                 i;
 
 #if !defined(NDEBUG)
 	if (compileall) {
@@ -1522,24 +1736,36 @@ void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 
 	status = 0;
 
-	if (opt_jar == true)
+	if (opt_jar == true) {
 		/* open jar file with java.util.jar.JarFile */
+
 		mainstring = vm_get_mainclass_from_jar(mainstring);
 
+		if (mainstring == NULL)
+			vm_exit(1);
+	}
+
 	/* load the main class */
 
 	mainutf = utf_new_char(mainstring);
 
-	if (!(mainclass = load_class_from_sysloader(mainutf)))
-		throw_main_exception_exit();
+#if defined(ENABLE_JAVAME_CLDC1_1)
+	mainclass = load_class_bootstrap(mainutf);
+#else
+	mainclass = load_class_from_sysloader(mainutf);
+#endif
 
 	/* error loading class */
 
-	if ((*exceptionptr != NULL) || (mainclass == NULL))
-		throw_main_exception_exit();
+	if ((exceptions_get_exception() != NULL) || (mainclass == NULL)) {
+		exceptions_print_stacktrace(); 
+		vm_exit(1);
+	}
 
-	if (!link_class(mainclass))
-		throw_main_exception_exit();
+	if (!link_class(mainclass)) {
+		exceptions_print_stacktrace();
+		vm_exit(1);
+	}
 			
 	/* find the `main' method of the main class */
 
@@ -1549,18 +1775,21 @@ void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 								 class_java_lang_Object,
 								 false);
 
-	if (*exceptionptr) {
-		throw_main_exception_exit();
+	if (exceptions_get_exception()) {
+		exceptions_print_stacktrace();
+		vm_exit(1);
 	}
 
 	/* there is no main method or it isn't static */
 
 	if ((m == NULL) || !(m->flags & ACC_STATIC)) {
-		*exceptionptr = NULL;
+		exceptions_clear_exception();
+		exceptions_throw_nosuchmethoderror(mainclass,
+										   utf_new_char("main"), 
+										   utf_new_char("([Ljava/lang/String;)V"));
 
-		*exceptionptr =
-			new_exception_message(string_java_lang_NoSuchMethodError, "main");
-		throw_main_exception_exit();
+		exceptions_print_stacktrace();
+		vm_exit(1);
 	}
 
 	/* build argument array */
@@ -1573,7 +1802,7 @@ void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 		u = utf_new_char(vm_args->options[opt_index + i].optionString);
 		s = javastring_new(u);
 
-		oa->data[i] = (java_objectheader *) s;
+		oa->data[i] = s;
 	}
 
 #ifdef TYPEINFO_DEBUG_TEST
@@ -1586,9 +1815,15 @@ void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 	jvmti_set_phase(JVMTI_PHASE_LIVE);
 #endif
 
-	/* increase total started thread count */
+	/* set ThreadMXBean variables */
 
-	_Jv_jvm->total_started_thread_count++;
+	_Jv_jvm->java_lang_management_ThreadMXBean_ThreadCount++;
+	_Jv_jvm->java_lang_management_ThreadMXBean_TotalStartedThreadCount++;
+
+	if (_Jv_jvm->java_lang_management_ThreadMXBean_ThreadCount >
+		_Jv_jvm->java_lang_management_ThreadMXBean_PeakThreadCount)
+		_Jv_jvm->java_lang_management_ThreadMXBean_PeakThreadCount =
+			_Jv_jvm->java_lang_management_ThreadMXBean_ThreadCount;
 
 	/* start the main thread */
 
@@ -1596,14 +1831,14 @@ void vm_run(JavaVM *vm, JavaVMInitArgs *vm_args)
 
 	/* exception occurred? */
 
-	if (*exceptionptr) {
-		throw_main_exception();
+	if (exceptions_get_exception()) {
+		exceptions_print_stacktrace();
 		status = 1;
 	}
 
 	/* unload the JavaVM */
 
-	vm_destroy(vm);
+	(void) vm_destroy(vm);
 
 	/* and exit */
 
@@ -1653,8 +1888,10 @@ void vm_exit(s4 status)
 	}
 #endif
 
-	if (!link_class(class_java_lang_System))
-		throw_main_exception_exit();
+	if (!link_class(class_java_lang_System)) {
+		exceptions_print_stacktrace();
+		exit(1);
+	}
 
 	/* call java.lang.System.exit(I)V */
 
@@ -1664,8 +1901,10 @@ void vm_exit(s4 status)
 								 class_java_lang_Object,
 								 true);
 	
-	if (m == NULL)
-		throw_main_exception_exit();
+	if (m == NULL) {
+		exceptions_print_stacktrace();
+		exit(1);
+	}
 
 	/* call the exit function with passed exit status */
 
@@ -1765,8 +2004,6 @@ void vm_exit_handler(void)
 #endif
 		}
 
-		mem_usagelog(1);
-
 		if (opt_getcompilingtime)
 			print_times();
 #endif /* defined(ENABLE_STATISTICS) */
@@ -1812,20 +2049,23 @@ static char *vm_get_mainclass_from_jar(char *mainstring)
 	classinfo         *c;
 	java_objectheader *o;
 	methodinfo        *m;
-	java_lang_String  *s;
+	java_objectheader *s;
 
 	c = load_class_from_sysloader(utf_new_char("java/util/jar/JarFile"));
 
-	if (c == NULL)
-		throw_main_exception_exit();
-	
+	if (c == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
+
 	/* create JarFile object */
 
 	o = builtin_new(c);
 
-	if (o == NULL)
-		throw_main_exception_exit();
-
+	if (o == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	m = class_resolveclassmethod(c,
 								 utf_init, 
@@ -1833,15 +2073,19 @@ static char *vm_get_mainclass_from_jar(char *mainstring)
 								 class_java_lang_Object,
 								 true);
 
-	if (m == NULL)
-		throw_main_exception_exit();
+	if (m == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	s = javastring_new_from_ascii(mainstring);
 
 	(void) vm_call_method(m, o, s);
 
-	if (*exceptionptr)
-		throw_main_exception_exit();
+	if (exceptions_get_exception()) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	/* get manifest object */
 
@@ -1851,14 +2095,16 @@ static char *vm_get_mainclass_from_jar(char *mainstring)
 								 class_java_lang_Object,
 								 true);
 
-	if (m == NULL)
-		throw_main_exception_exit();
+	if (m == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	o = vm_call_method(m, o);
 
 	if (o == NULL) {
 		fprintf(stderr, "Could not get manifest from %s (invalid or corrupt jarfile?)\n", mainstring);
-		vm_exit(1);
+		return NULL;
 	}
 
 
@@ -1870,14 +2116,16 @@ static char *vm_get_mainclass_from_jar(char *mainstring)
 								 class_java_lang_Object,
 								 true);
 
-	if (m == NULL)
-		throw_main_exception_exit();
+	if (m == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	o = vm_call_method(m, o);
 
 	if (o == NULL) {
 		fprintf(stderr, "Could not get main attributes from %s (invalid or corrupt jarfile?)\n", mainstring);
-		vm_exit(1);
+		return NULL;
 	}
 
 
@@ -1889,15 +2137,19 @@ static char *vm_get_mainclass_from_jar(char *mainstring)
 								 class_java_lang_Object,
 								 true);
 
-	if (m == NULL)
-		throw_main_exception_exit();
+	if (m == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	s = javastring_new_from_ascii("Main-Class");
 
 	o = vm_call_method(m, o, s);
 
-	if (o == NULL)
-		throw_main_exception_exit();
+	if (o == NULL) {
+		exceptions_print_stacktrace();
+		return NULL;
+	}
 
 	return javastring_tochar(o);
 }
@@ -1946,7 +2198,7 @@ static void vm_compile_all(void)
 
 						/* print out exception and cause */
 
-						exceptions_print_exception(*exceptionptr);
+						exceptions_print_current_exception();
 
 						/* goto next class */
 
@@ -1970,7 +2222,7 @@ static void vm_compile_all(void)
 
 							/* print out exception and cause */
 
-							exceptions_print_exception(*exceptionptr);
+							exceptions_print_current_exception();
 						}
 					}
 				}
@@ -1994,11 +2246,13 @@ static void vm_compile_method(void)
 
 	/* create, load and link the main class */
 
-	if (!(mainclass = load_class_bootstrap(utf_new_char(mainstring))))
-		throw_main_exception_exit();
+	mainclass = load_class_bootstrap(utf_new_char(mainstring));
+
+	if (mainclass == NULL)
+		exceptions_print_stacktrace();
 
 	if (!link_class(mainclass))
-		throw_main_exception_exit();
+		exceptions_print_stacktrace();
 
 	if (opt_signature != NULL) {
 		m = class_resolveclassmethod(mainclass,
@@ -2015,17 +2269,9 @@ static void vm_compile_method(void)
 									 false);
 	}
 
-	if (m == NULL) {
-		char message[MAXLOGTEXT];
-		sprintf(message, "%s%s", opt_method,
-				opt_signature ? opt_signature : "");
-
-		*exceptionptr =
-			new_exception_message(string_java_lang_NoSuchMethodException,
-								  message);
-										 
-		throw_main_exception_exit();
-	}
+	if (m == NULL)
+		vm_abort("vm_compile_method: java.lang.NoSuchMethodException: %s.%s",
+				 opt_method, opt_signature ? opt_signature : "");
 		
 	jit_compile(m);
 }
@@ -2289,6 +2535,8 @@ java_objectheader *vm_call_method_vmarg(methodinfo *m, s4 vmargscount,
 {
 	java_objectheader *o;
 
+	STATISTICS(count_calls_native_to_java++);
+
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (opt_intrp)
@@ -2419,6 +2667,8 @@ s4 vm_call_method_int_vmarg(methodinfo *m, s4 vmargscount, vm_arg *vmargs)
 {
 	s4 i;
 
+	STATISTICS(count_calls_native_to_java++);
+
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (opt_intrp)
@@ -2549,6 +2799,8 @@ s8 vm_call_method_long_vmarg(methodinfo *m, s4 vmargscount, vm_arg *vmargs)
 {
 	s8 l;
 
+	STATISTICS(count_calls_native_to_java++);
+
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (opt_intrp)
@@ -2681,6 +2933,8 @@ float vm_call_method_float_vmarg(methodinfo *m, s4 vmargscount, vm_arg *vmargs)
 {
 	float f;
 
+	STATISTICS(count_calls_native_to_java++);
+
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (opt_intrp)
@@ -2814,6 +3068,8 @@ double vm_call_method_double_vmarg(methodinfo *m, s4 vmargscount,
 {
 	double d;
 
+	STATISTICS(count_calls_native_to_java++);
+
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (opt_intrp)
